/*
 * Yield stuff, including interrupt hook, and child-hook dispatch.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: blocker.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __KERN_YIELD_H
#define __KERN_YIELD_H(x) __tkern_ ## x

/* ---------------------------------------------------------------------- */

#define blocker_get_indos_flag       __KERN_YIELD_H(blocker_get_indos_flag)
extern int blocker_get_indos_flag(void);
#define blocker_get_inisr_flag       __KERN_YIELD_H(blocker_get_inisr_flag)
extern int blocker_get_inisr_flag(void);

/* ---------------------------------------------------------------------- */

#define blocker_hook_op              __KERN_YIELD_H(blocker_hook_op)
extern int blocker_hook_op(int hooktype, void (*fxn)(void), int mode);

#define BLOCKER_HOOKTYPE_ISR      0x01
#define BLOCKER_HOOKTYPE_IDLE     0x02
#define BLOCKER_HOOKTYPE_PERIODIC 0x04

#define blocker_tisrhook_add(__f) blocker_hook_op(BLOCKER_HOOKTYPE_ISR,__f,'+')
#define blocker_tisrhook_del(__f) blocker_hook_op(BLOCKER_HOOKTYPE_ISR,__f,'-')
#define blocker_idlehook_add(__f) blocker_hook_op(BLOCKER_HOOKTYPE_IDLE,__f,'+')
#define blocker_idlehook_del(__f) blocker_hook_op(BLOCKER_HOOKTYPE_IDLE,__f,'-')
#define blocker_perdhook_add(__f) blocker_hook_op(BLOCKER_HOOKTYPE_PERIODIC,__f,'+')
#define blocker_perdhook_del(__f) blocker_hook_op(BLOCKER_HOOKTYPE_PERIODIC,__f,'-')

#define blocker_critsect_begin    __KERN_YIELD_H(blocker_critsect_begin)
#define blocker_critsect_end      __KERN_YIELD_H(blocker_critsect_end)

extern void blocker_critsect_begin(void);
extern void blocker_critsect_end(void);

/* ---------------------------------------------------------------------- */

#define blocker_inc_hwork   __KERN_YIELD_H(blocker_inc_hwork)
extern void blocker_inc_hwork(void);

/* ---------------------------------------------------------------------- */

#define YIELDFLAG_HARD   0x01 /* Do yield to OS */
#define YIELDFLAG_CHKBRK 0x02 /* Do ^C checking */
#define YIELDFLAG_NOIHKS 0x80 /* Do NOT run idle hooks */

#define blocker_cyield       __KERN_YIELD_H(blocker_cyield)
#define blocker_cdelay       __KERN_YIELD_H(blocker_cdelay)

extern int blocker_cyield (int yield_flags,const char *caller,int line);
extern int blocker_cdelay (unsigned int msecs,const char *caller,int line);

#if defined(NDEBUG)
#define blocker_yield(__x) blocker_cyield((__x),0,0)
#define blocker_delay(__x) blocker_cdelay((__x),0,0)
#elif defined(__GNUC__) || defined(__WATCOMC__)
#define blocker_yield(__x) blocker_cyield((__x),__FUNCTION__,__LINE__)
#define blocker_delay(__x) blocker_cdelay((__x),__FUNCTION__,__LINE__)
#else
#define blocker_yield(__x) blocker_cyield((__x),__FILE__,__LINE__)
#define blocker_delay(__x) blocker_cdelay((__x),__FILE__,__LINE__)
#endif

#endif /* __KERN_YIELD_H */

