/*
 * - _my_atexit(): a replacement for atexit() so as to not 
 *   consume libc atexit() slots.
 * - _my_atterm(): a means to implement shutdown handlers that
 *   are called from _exit()/_terminate()
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: ctordtor.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#include "target.h"   /* exit(), _exit(), atexit() etc */
#include "dpmi.h"     /* __my_dpmi_{get_csalias|lock_code}() */
#include "printk.h"   /* for debugging */
#include "ctordtor.h" /* ourselves */

/* ---------------------------------------------------------------------- */

#undef  TRACE_ATx_CALLS /* define to see atterm()/atexit() funcs being called */

#if defined(TRACE_ATx_CALLS)
#undef TRACE_ATx_CALLS
#define TRACE_ATx_CALLS(x) _printk x
#else
#define TRACE_ATx_CALLS(x) NO_OPERATION()
#endif

/* ===================================================================== */
/*                     atterm() to match libc's atexit()                 */
/* ===================================================================== */

/*
 * atterm() is to _exit() what atexit() is to exit(), ie it is a means 
 * of installing destructors that need to be called when the application
 * is terminating, and even when doing due to a direct call to _exit() 
 * or __exit() [fatal error, abort(), SIGxxx, etc]. 
 *
 * atterm() (clever name, eh?:) returns non-zero (like atexit) if 
 * registration fails.
 *
 * Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * __WATCOMC__ doesn't really need the function ptr redirection hackery 
 * since it has (a class of) destructors that can be called at _exit() time.
 * I haven't used this capability in order to be uniform with other compilers.
*/
static struct
{
  char isnbuf[8];
  /* Since atterm() is not going to be used by anything other than 
   * _my_atterm(), it only needs one slot in the function table.
  */
  void (*fxn[1])(void); /* only one function needed */
} FAR atterm_data;

/* ------------------------------------------------------------- */

static void __dtor__exit(void)
{
  unsigned n = (sizeof(atterm_data.fxn)/sizeof(atterm_data.fxn[0]));
  while (n)
  {
    n--;
    if (atterm_data.fxn[n])
    {
      _my_athook_t fxn = atterm_data.fxn[n];
      TRACE_ATx_CALLS(("calling atterm hook %p\n", fxn ));
      atterm_data.fxn[n] = 0;
      (*fxn)(); 
    }
  }
  return;
}

/*
 * __WATCOMC__ doesn't really need the function ptr redirection hackery 
 * since it has (a class of) destructors that can be called at _exit() time.
 * I haven't used this capability in order to be uniform with other compilers.
*/
#if 1
#undef HAVE_ATTERM_STYLE_DESTRUCTOR
#else
#define __SETUP_DESTRUCTOR_FOR__ __dtor__exit
#include "ctordtor.h"
#endif

/* ------------------------------------------------------------- */

#if !defined(HAVE_ATTERM_STYLE_DESTRUCTOR)
#if defined(__DJGPP__) /* sighandlers must be shutdown after us */
#define __exit _exit
#else                  /* no deinitialization order grief. */
#  if defined(__TURBOC__)
#  define __exit _terminate
#  endif
#  ifdef __cplusplus
   extern "C" void __exit(int);
#  else
   extern void __exit(int);
#  endif
#endif
static void __stub__exit(int ret_code)
{
  #if defined(__DJGPP__)
  /* __asm__ __volatile__("movw %w0,%%ds\n\t" :: "rm"(__djgpp_ds_alias) ); */
  #endif

  TRACE_ATx_CALLS(("begin __stub_exit() 1\n"));
  #if defined(__DPMI32__)
  movedata( _my_ds(), (unsigned)&(atterm_data.isnbuf[0]), 
            __my_dpmi_get_csalias(), (unsigned)__exit, 
            sizeof(atterm_data.isnbuf) );
  __my_dpmi_unlock_code( (const void *)((unsigned)__exit), 
                         sizeof(atterm_data.isnbuf) );
  __my_dpmi_unlock_data(&atterm_data, sizeof(atterm_data));
  #else
  _fmemcpy( __exit, &(atterm_data.isnbuf[0]), sizeof(atterm_data.isnbuf) );
  #endif

  TRACE_ATx_CALLS(("begin __stub_exit() 2\n"));
  __dtor__exit();
  TRACE_ATx_CALLS(("begin __stub_exit() 3\n"));
  __exit(ret_code);
  TRACE_ATx_CALLS(("begin __stub_exit() 4 (huh?)\n"));
}  
#endif /* !defined(HAVE_ATTERM_STYLE_DESTRUCTOR) */

/* ------------------------------------------------------------- */

static int atterm(_my_athook_t runner)
{
#if !defined(HAVE_ATTERM_STYLE_DESTRUCTOR)
  static int done_init = -1;
  if (done_init == -1)
  {
    unsigned long offset;
    char buf[sizeof(atterm_data.isnbuf)];

    #if defined(__DPMI32__)
    if (__my_dpmi_get_csalias() == -1)
      done_init = 0;
    else if (__my_dpmi_lock_code( 
      (const void *)((unsigned)__exit), sizeof(atterm_data.isnbuf) ) != 0)
      done_init = 0;
    else if (__my_dpmi_lock_data(&atterm_data, sizeof(atterm_data))!=0)
      done_init = 0;
    else
    {
      movedata( _my_cs(), (unsigned)__exit, 
                _my_ds(), (unsigned)&(atterm_data.isnbuf[0]), 
                sizeof(atterm_data.isnbuf) );
      offset = ((unsigned long)__stub__exit) - ((unsigned long)__exit);
      buf[0] = 0xe9;    /* "jmp" (relative) */
      *((unsigned long *)(&buf[1])) = offset - 5; /* 5==sizeof('jmp xyz') */
      movedata( _my_ds(), (unsigned)(&buf[0]), 
               __my_dpmi_get_csalias(), (unsigned)__exit, 5 );    
      done_init = +1;
    }
    #else 
    {
      _fmemcpy( atterm_data.isnbuf, __exit, sizeof(atterm_data.isnbuf) );
      offset = (unsigned long)((void _far *)__stub__exit);
      buf[0] = 0xEA;    /* jmp absolute */
      *((unsigned long *)(&buf[1])) = offset;
      _fmemcpy( __exit, buf, 5 );
      done_init = +1;
    }
    #endif
  }
  if (done_init == +1)
#endif /* !!defined(HAVE_ATTERM_STYLE_DESTRUCTOR) */
  { 
    if (runner)
    {
      unsigned int n;
      for (n=0; n<(sizeof(atterm_data.fxn)/sizeof(atterm_data.fxn[0])); n++)
      {
        if (!atterm_data.fxn[n])
        {
          atterm_data.fxn[n] = runner;
          return 0;
        }
      }
    }
  }
  return -1;
}

/* ===================================================================== */
/*                       _my_atexit() and _my_atterm()                   */
/* ===================================================================== */

struct __stop_hooks_info
{
  unsigned int hook_added;
  unsigned int hook_count;
  void (*hook_list[8])(void);
};

static struct
{
  struct __stop_hooks_info atexit_hooks;
  struct __stop_hooks_info atterm_hooks;
  int is_exiting;
} FAR stop_hooks_tabs;

/* ------------------------------------------------------------------- */

static void ___run_hooks(struct __stop_hooks_info FAR *shi)
{
  int hk = shi->hook_count - 1;
  stop_hooks_tabs.is_exiting++;
  for (; hk >= 0; hk--)
  {
    if (shi->hook_list[hk])
    {
      auto void (*fxn)(void) = shi->hook_list[hk];
      shi->hook_count = hk;
      shi->hook_list[hk] = 0;
      TRACE_ATx_CALLS(("doing at%s hook #%u %p\n",
        (shi == &stop_hooks_tabs.atexit_hooks ? "exit" : "term" ), hk, fxn ));
      (*fxn)();
    }
  }
  return;
}

static void ___run_atterm_hooks(void)
{
  ___run_hooks( &(stop_hooks_tabs.atterm_hooks) );
}

static void ___run_atexit_hooks(void)
{
  ___run_hooks( &(stop_hooks_tabs.atexit_hooks) );
}

static int __add_hooks(_my_athook_t fxn, 
                       struct __stop_hooks_info FAR *shi,
                       const char *at_fxn_name,
                       int (*at_fxn_vect)(void (*)(void)), 
                       void (*runner)(void) )
{
  int rc = 0; /* atexit() compatible */
  if (!shi->hook_added)
  {
    rc = 0;
    if (at_fxn_vect)
      rc = (*at_fxn_vect)(runner);
    if (rc == 0)
      shi->hook_added = 1;
    else
    {
      __nonfatal_runtime_error("PANIC: libc ");
      __nonfatal_runtime_error(at_fxn_name);
      __nonfatal_runtime_error(" returned non-zero\n");
      rc = -1;
    }
  }
  if (rc == 0 && fxn && shi->hook_count >= 
         (sizeof(shi->hook_list)/sizeof(shi->hook_list[0]))-1)
  {
    __nonfatal_runtime_error("PANIC: kernel ");
    __nonfatal_runtime_error(at_fxn_name);
    __nonfatal_runtime_error(" table is too small!\n");
    rc = -1;
  }
  if (rc != 0)
  {
    if (fxn && shi == &(stop_hooks_tabs.atexit_hooks))
      (*fxn)();
    ___run_atexit_hooks();
    if (fxn && shi == &(stop_hooks_tabs.atterm_hooks))
      (*fxn)();
    ___run_atterm_hooks();
    exit(255);
  }
  else if (fxn)
  {
    TRACE_ATx_CALLS(("added %s hook #%u (%p)\n", 
                      at_fxn_name, shi->hook_count, fxn ));
    shi->hook_list[shi->hook_count++]=fxn;
  }
  return rc; /* compatibility with atexit() */
}

/* ------------------------------------------------------------------- */

/* _my_atexit() == atexit() but does not consume libc atexit() 
** table space and guarantees that hook placement will succeed.
*/
int _my_atexit(_my_athook_t fxn)
{
  return __add_hooks( fxn, &(stop_hooks_tabs.atexit_hooks),
                      "atexit", atexit, ___run_atexit_hooks );
}

int _my_atterm(_my_athook_t fxn)
{
  return __add_hooks( fxn, &(stop_hooks_tabs.atterm_hooks),
                      "atterm", atterm, ___run_atterm_hooks );
}

int __kern_is_exiting(void)
{
  return stop_hooks_tabs.is_exiting;
}
