/*
** Debugging message helpers
** May 2001 <cyp@fb14.uni-mainz.de>
*/

#ifndef __KERN_DEBUG_H
#define __KERN_DEBUG_H(x) __tkern_ ## x

#define __wdl_msg     __KERN_DEBUG_H(wdl_msg)
#define __dbg_msg     __KERN_DEBUG_H(dbg_msg)
#define __dbg_dmp     __KERN_DEBUG_H(dbg_dmp)
#define __dbg_t2a     __KERN_DEBUG_H(dbg_t2a)
#define __ima_debugee __KERN_DEBUG_H(ima_debugee)

/* --------------------------------------------------------------------- */

/* debug_on levels - all KMSG()s must fall into one of these categories */
#define WDLEV_UCFAIL    1 /* 1: something a user might be able to do to fix */
#define WDLEV_NCFAIL    2 /* 2: non-config reasons why a function might fail */
#define WDLEV_TRACE     3 /* 3: debugging */

extern void (*__wdl_msg(const char *file,int line,const char *fxn)) \
                              (int msglev, const char *fmt, ...);

#if defined(__GNUC__) || defined(__WATCOMC__)
#define KMSG(x) (*(__wdl_msg(__FILE__,__LINE__,__FUNCTION__))) x
#else
#define KMSG(x) (*(__wdl_msg(__FILE__,__LINE__,((const char *)0) ))) x
#endif

/* --------------------------------------------------------------------- */

extern void (*__dbg_msg(const char *file,int line,const char *fxn)) \
                              (const char *fmt, ...);
extern void (*__dbg_dmp(const char *file,int line,const char *fxn)) \
                              (const char *label, unsigned short pktid, 
                               const void FAR *pkt, unsigned int len );
extern const char *__dbg_t2a( unsigned long t, int scale, 
                              char *buffer, unsigned int buflen );

#if defined(NDEBUG)
#define DBGPRINTF(x)      NO_OPERATION() /* nothing */
#define DBGDUMP(x)        NO_OPERATION() /* nothing */
#define DBG_T2A(t,s,b,bl) NO_OPERATION() /* nothing */
#elif defined(__GNUC__) || defined(__WATCOMC__)
#define DBGPRINTF(x)   (*(__dbg_msg(__FILE__,__LINE__,__FUNCTION__))) x
#define DBGDUMP(x)     (*(__dbg_dmp(__FILE__,__LINE__,__FUNCTION__))) x
#define DBG_T2A(t,s,b,bl) __dbg_t2a(t,s,b,bl)
#else
#define DBGPRINTF(x)   (*(__dbg_msg(__FILE__,__LINE__,((const char *)0) ))) x
#define DBGDUMP(x)     (*(__dbg_dmp(__FILE__,__LINE__,((const char *)0) ))) x
#define DBG_T2A(t,s,b,bl) __dbg_t2a(t,s,b,bl)
#endif

extern int __ima_debugee(void);

#endif /* __KERN_DEBUG_H */

