/*
** "uniform" C-callable interface to the DPMI API
**
** Created May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
**
** For simplicity, the C-callable interface to DPMI functions as
** used/prototyped here is the same as the one prototyped in DJGPP.
** The DGPP GO32 extensions to DPMI are _not_ emulated.
**
** Note: __DPMI32__ will always be defined by dpmi.h for 32bit targets.
** - if a 386 extender is not DPMI compliant (for example PMODE
**   without its Int31h interface), then the application that links this 
**   library will never go anywhere. :)
** - __DPMI32__ is defined even when pharlap.h has been included.
**   (Some people think that an app running under a pharlap extender 
**   cannot make DPMI calls. Let those poor, huddled, unwashed masses 
**   be herewith corrected).
*/
/*
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: dpmi.c,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#define PRINTK_ENABLED
#include <stdlib.h> /* malloc() */
#include <string.h> /* memset()/memcpy() */
#include <dos.h>    /* int386x() */
#include "printk.h" /* printk() */
#include "dpmi.h"   /* ourselves */

#undef DPMI_INIT_TRACE /* define this to show init trace */

/* -------------------------------------------------------------------- */

#if defined(__WATCOMC__)
  #if defined(__cplusplus)
  #pragma aux ___wcpp_4_data_init_fs_root_ "__argc";
  #pragma aux ___wcpp_4_data_init_longjmp_ "__argc";
  #pragma aux ___wcpp_4_data_init_fs_root  "__argc";
  #pragma aux ___wcpp_4_data_init_longjmp  "__argc";
  extern "C" {
  #endif
    extern unsigned _STACKLOW;
    extern unsigned _STACKTOP;
    extern unsigned char _Extender;
    extern unsigned short __X32VM;
  #if defined(__cplusplus)
  }
  #endif
#elif defined(__DJGPP__)
  #if defined(__cplusplus)
  extern "C" {
  #endif
  extern unsigned __djgpp_stack_limit;
  #if defined(__cplusplus)
  }
  #endif
#endif

/* ==================================================================== */
/*                            helper routines                           */
/* ==================================================================== */

#if defined(__DPMI32__) && !defined(__DJGPP__)
static int __IsPharlap(void)
{
#if 0
  static int is = -1;
  if (is == -1)
  {
    union REGS r;
    r.x.ebx = 0x50484152; /* "PHAR" */
    r.x.eax = 0x3000; /* ah=30h */
    int386 (0x21, &r, &r);
    r.x.eax >>= 16;
    is = ((r.x.eax == 0x4458 /* 'DX' */) ? (+1) : (0));
    /* we don't care about Code Builder since CB is a DPMI host */
    /* with pharlap extensions (ie, not pharlap with DPMI extensions) */
  }
  return is;
#else
  return 0;
#endif
}
#endif /* __DPMI32__ */

/* -------------------------------------------------------------------- */

#if 0 /* defined(__DPMI32__) */ /* unused */
/* warning: What dpmi returns for cputype is only as good 
** as the info that was available when the host was written
** The DPMI spec itself only qualifies 2=286,3=386,4=486.
** Many hosts return 0xff for a P4.
** NTVDM always returns 0x03
*/
int __my_dpmi_cpu_type (void)
{
  union REGS r;
  r.x.eax = 0x400;              /* Get DPMI Version */
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
     return (-1);
  return (r.h.cl);
} 
#endif

/* -------------------------------------------------------------------- */

#if 0 /* defined(__DPMI32__) */ /* unused */
int __my_dpmi_get_extender_type(void)
{
  static short xtype = -1;
  if (xtype == -1)
  {
#if defined(__WATCOMC__)
    #ifndef __cplusplus
    extern unsigned char _Extender;
    #endif
    if (_Extender == 0)       /* Ergo DPMI */
      xtype = DX_ERGO;
    else if (_Extender == 1)  /* DOS4G/DOS4GW */
      xtype = DX_RATIONAL;
    else if (_Extender >= 2 && _Extender <= 8)
    {
      #ifndef __cplusplus
      extern unsigned short __X32VM;
      #endif
      xtype = DX_PHARLAP;
      if (__X32VM)
        xtype = DX_FLASHTEK;
    }
    else if (_Extender == 9)
      xtype = DX_INTELCB;
#elif defined(__DJGPP__)
    unsigned long base;
    if (__dpmi_get_segment_base_address(__my_dpmi_get_ds(), &base ) != 0)
      base = (unsigned long)-1;
    xtype = DX_CWSDPMI;
    if (base == 0)
      xtype = DX_PMODEDJ;
#else
    union REGS r;
    r.x.ebx = 0x50484152; /* "PHAR" */
    r.x.eax = 0x3000; /* ah=30h */
    int386 (0x21, &r, &r);
    r.x.eax >>= 16;
    if (r.x.eax == 0x4243) /* 'BC' */
      xtype = DX_INTELCB;
    else if (r.x.eax == 0x4458) /* 'DX' */
    {
      xtype = DX_PHARLAP;
      /* flashtek detection needed here */ 
    }
    if (xtype == -1)
    {
      r.x.edx = 0x78;
      r.x.eax = 0x0FF00h;
      int386 (0x21, &r, &r);
      if (r.x.al != 0)
        xtype = DX_RATIONAL;
    }
    if (xtype == -1)
    {
      unsigned int buf[16/4];
      struct SREGS s;
      segread (&s);
      s.es = s.ss; 
      r.x.edi = (unsigned)&buf[0];
      r.x.eax = 0xA1FE;
      r.x.ebx = 0x81;
      int386x (0x2F, &r, &r, &s);
      if (buf[0] != 0x48424149 /*"IABH"/'HBAI'*/)
      {
        r.x.edi = (unsigned int)&buf[0];
        r.x.eax = 0xA1FF;
        r.x.ebx = 0x81;
        int386x (0x2F, &r, &r, &s);
      }
      if (buf[0] == 0x48424149 /*"IABH"/'HBAI'*/)
        xtype = DX_ERGO;
    }
#endif
    if (xtype == -1)
      xtype = 0;
  }
  return xtype;
}
#endif

/* ==================================================================== */
/*                            get DOSmem selector                       */
/* ==================================================================== */

#if defined(__DPMI32__) && !defined(__DPMI32_DOSDS_INTRINSIC__) 
unsigned int __my_dpmi_get_dosds(void) /* inline for DJGPP/WATCOM/PHARLAP_H */
{
#if defined(__DJGPP__)
  return _dos_ds;             /* inline macro returns this */
#elif defined(__WATCOMC__)
  return _ExtenderRealModeSelector; /* inline macro returns this */
#elif defined(PHARLAP_H)
  return 0x34 | (__my_dpmi_get_ds() & 0x03); /* inline macro returns this */
#else
  static int my_dos_ds = -1;
  int sel = my_dos_ds;

  if (sel == -1)
  {
    unsigned long base, limit;
    int allocated = 0;

    /* first see if my DS == DOSmem DS. This is true for ...
     * - (ZeroBase) Rational-DOS4GW and compatibles (DOS32, PMODE/W, Causeway)
     * - IntelCodeBuilder 
    */
    sel = __my_dpmi_get_ds();
    if (__dpmi_get_segment_base_address(sel, &base ) != 0)
      sel = -1;
    else if (base != 0)
      sel = -1;
    else if ((limit = __dpmi_get_segment_limit(sel)) == 0)
      sel = -1;
    else if (limit < 0xfffff)
    {
      if (__dpmi_set_segment_limit(sel, 0xfffff))
        sel = -1;
    }

    /* Now check for Pharlap (but not Flashtek) and Ergo DPMI (OS386).
     * These have selector 0x34 mapped to the first meg of memory (RO)
    */
    if (sel == -1)
    {
      if (__dpmi_get_segment_base_address(0x34, &base ) == 0)
      {
        if (base == 0)
        {
          /* use same priv level as DS register */
          sel = 0x34 | (__my_dpmi_get_ds() & 0x03);
        }
      }
    }

    /* Will only get here for BorlandC/MSC with
    ** FlashTek or other rarely seen extender.
    */
    if (sel == -1)
    {
      sel = __dpmi_allocate_ldt_descriptors(1);
      if (sel != -1)
      {
        allocated = 1;
        if (__dpmi_set_segment_limit(sel, 0x10ffff))
        {
          __dpmi_free_ldt_descriptor(sel);
          sel = -1;
          allocated = 0;
        } 
        /* stays open forever - extender will free it on exit */
      } 
    }

    /* if we have succeeded in obtaining a selector,
     * then lock code and data since this function may be
     * called from a hardware interrupt handler.
    */
    if (sel != -1)
    {
      if (__my_dpmi_lock_code(__my_dpmi_get_dosds,1024) != 0 ||
          __my_dpmi_lock_data(&my_dos_ds, sizeof(my_dos_ds)) != 0)
      {
        if (allocated)
          __dpmi_free_ldt_descriptor(sel);
        sel = -1;
      }
    }

    if (sel != -1)
      my_dos_ds = sel;
  }
  return sel;
#endif
}  
#endif

/* ==================================================================== */
/*                         get writable code selector                   */
/* ==================================================================== */

/* __my_dpmi_get_csalias() is self initializing.
*/
#if defined(__DPMI32__)
int __my_dpmi_get_csalias(void)
{
  static int cs_alias = -2;
  if (cs_alias == -2)
  {
    int sel = __dpmi_create_alias_descriptor(_my_cs()); /* -1 on error */
    if (sel != -1)
    {
      if (__my_dpmi_lock_code((const void *)
                    ((unsigned)__my_dpmi_get_csalias), 1024) != 0 ||
          __my_dpmi_lock_data(&cs_alias, sizeof(cs_alias)) != 0)
      {
        __dpmi_free_ldt_descriptor(sel);
        sel = -1;
      }
    }
    cs_alias = sel;
  }
  return cs_alias;
}
#endif

/* ==================================================================== */
/*                   DPMI API compatible with DJGPP                     */
/* ==================================================================== */

#if defined(__DPMI32__) && !defined(__DJGPP__)
void __dpmi_yield (void)
{
  __dpmi_regs r;
  r.x.ax = 0x1680;
  __dpmi_int (0x2F, &r);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
unsigned long __dpmi_get_segment_limit( unsigned int _sel )
{              /* not really a DPMI API call - returns zero if not legal */
#if defined(__GNUC__)
  /* GCC as of 2.8 doesn't understand 'lsll'
   * unsigned int lim; 
   * __asm__("lsll %1,%0": "=r" (lim) : "r" (seg));
   * return(lim+1);
  */
  return __dpmi_get_segment_limit(_sel);
#elif defined(__HIGHC__)
  return _mwlsl(__pmseg)
#else /* if defined(__WATCOMC__) || defined(__BORLANDC__) */
  unsigned long result = 0;
  __asm xor eax,eax
  __asm mov eax, _sel
  __asm lsl eax, ax
  __asm setnz cl      /* set cl = 0 on error, non-0 on success */
  __asm cmp cl,1      /* set carry on success */
  __asm sbb ecx, ecx  /* set ecx = -1 on success, ecx = 0 on error */
  __asm and eax,ecx
  __asm mov result,eax
  return result;
#endif
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_allocate_ldt_descriptors(int count)
{                                    /* Int 31H Function 0000H [0.9] */
  union REGS r;
  r.x.eax = 0x0000;
  r.x.ecx = count;
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
    return -1;
  r.x.eax &= 0xffff;
  return r.x.eax;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_free_ldt_descriptor(int sel)
{                                    /* Int 31H Function 0001H [0.9] */
  union REGS r;
  r.x.eax = 0x0001;
  r.x.ebx = sel;
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
    return -1;
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_segment_base_address( int selector, unsigned long *base )
{                                         /* Int 31H Function 0006H [0.9] */
  union REGS r;
  *base = 0;
  if (__IsPharlap())
  {
    r.x.eax = 0x2508;
    r.x.ebx = selector;
    int386 (0x21, &r, &r);
    if (r.w.cflag & 1)
      return -1;
    *base = r.x.ecx;
    return 0;
  }
  r.x.eax = 0x0006;
  r.x.ebx = selector;
  r.x.ecx = r.x.edx = 0xfffffffful;
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1) 
    return -1;
  r.x.ecx <<= 16;
  r.w.cx = r.w.dx;
  *base = r.x.ecx;
  return 0;  
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_set_segment_base_address(int sel, unsigned long base)
{                                    /* Int 31H Function 0007H [0.9] */
  union REGS r;
  r.x.eax = 0x0007;
  r.x.ebx = sel;
  r.x.edx = base;
  r.x.ecx = base;
  r.x.ecx >>= 16;
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
    return -1;
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_set_segment_limit(int sel, unsigned long limit)
{                                    /* Int 31H Function 0008H [0.9] */
  union REGS r;
  r.x.eax = 0x0008;
  r.x.ebx = sel;
  r.x.edx = limit;
  r.x.ecx = limit;
  r.x.ecx >>= 16;
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
    return -1;
  return 0; 
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_create_alias_descriptor(int sel) /* Int 31H Function 000AH [0.9] */  
{                                  
  union REGS r;
  if (__IsPharlap())
  {
    struct SREGS s;
    segread (&s);
    r.x.eax = 0x2515;
    r.x.ebx = s.ds;
    int386 (0x21, &r, &r);
    if (r.w.cflag & 1)
      return 0;
    r.x.eax = 0x2513;
    r.x.ebx = s.cs;        
    /* inherit cx */ 
    int386 (0x21, &r, &r);
  }
  else
  {
    r.x.eax = 0x000A;               
    r.x.ebx = sel;                  
    int386 (0x31, &r, &r);
  }
  if (r.w.cflag & 1)
    return -1;
  r.x.eax &= 0xffff;
  return r.x.eax;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_allocate_dos_memory (int paras, int *sel_or_max)
{                               /* Int 31H Function 0100H [0.9] */
  union REGS r;
  if (((unsigned int)paras) > 0xffffu)
    paras = 0xffff; /* 0xffff means get max size */

  r.x.eax = 0x0100;             /* DPMI allocate DOS memory */
  r.x.ebx = paras;              /* Number of paragraphs requested */
  int386 (0x31, &r, &r);
  if ((r.w.cflag & 1) == 0) 
  { 
    r.x.eax &= 0xffff;
    r.x.edx &= 0xffff;
    *sel_or_max = r.x.edx;
    return r.x.eax; /* Return segment address */
  }
  /* error - bx has max size */
  r.x.ebx &= 0xffff;
  *sel_or_max = r.x.ebx;
  return -1;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_free_dos_memory (int selector) /* Int 31H Function 0101H [0.9] */
{
  union REGS r;
  r.x.eax = 0x101;              /* DPMI free DOS memory */
  r.x.ebx = selector;           /* Selector to free */
  int386 (0x31, &r, &r);
  return (0); /* this function always succeeds */
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_resize_dos_memory (int sel, int newparas, int *ret_max)
{                               /* Int 31H Function 0102H [0.9] */
  union REGS r;
  if (((unsigned int)newparas) > 0xffffu)
    newparas = 0xffff; /* 0xffff means get max size */

  r.x.edx = sel;
  r.x.ebx = newparas;           /* Number of paragraphs requested */
  r.x.eax = 0x0100;             /* DPMI resize DOS memory */
  int386 (0x31, &r, &r);

  if ((r.w.cflag & 1) == 0)
    return 0;

  /* error - bx has max size */
  r.x.ebx &= 0xffff;
  *ret_max = r.x.ebx;
  return -1;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_real_mode_interrupt_vector(int intno, __dpmi_raddr *rmaddr) 
{                                         /* Int 31H Function 0200H [0.9] */
  union REGS r;
  if (__IsPharlap())
  {
    r.x.eax = 0x2503;
    r.x.ecx = intno;
    int386 (0x21, &r, &r);
    *((unsigned long *)rmaddr) = r.x.ebx;
    return 0;
  }
  r.x.eax = 0x200;
  r.x.ebx = intno;
  int386 (0x31, &r, &r);
  rmaddr->segment = r.w.cx;
  rmaddr->offset16 = r.w.dx;
  return (0); /* this function always succeeds */
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_set_real_mode_interrupt_vector(int intno, __dpmi_raddr *rmaddr)
{                                         /* Int 31H Function 0201H [0.9] */
  union REGS r;
  if (__IsPharlap())
  {
    r.x.eax = 0x2505;
    r.x.ecx = intno;
    r.x.ebx = *((unsigned long *)rmaddr);
    int386 (0x21, &r, &r);
    return 0;
  }
  r.x.eax = 0x201;
  r.x.ebx = intno;
  r.w.cx = rmaddr->segment;
  r.w.dx = rmaddr->offset16;
  int386 (0x31, &r, &r); 
  return (0); /* this function always succeeds */
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_protected_mode_interrupt_vector(int intno, __dpmi_paddr *pmaddr)
{                                          /* Int 31H Function 0204H [0.9] */
  union REGS r;
  if (__IsPharlap())
  {
    struct SREGS s;
    segread( &s); 
    r.x.eax = 0x2502;
    r.x.ecx = intno;
    int386x (0x21, &r, &r, &s);
    pmaddr->offset32 = r.x.ebx;
    pmaddr->selector = s.es;
    return 0;
  }
  r.x.eax = 0x204;
  r.x.ebx = intno;
  int386 (0x31, &r, &r); 
  pmaddr->selector = r.w.cx;
  pmaddr->offset32 = r.x.edx;
  return (0); /* this function always succeeds */
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_set_protected_mode_interrupt_vector(int intno, __dpmi_paddr *pmaddr)
{                                           /* Int 31H Function 0205H [0.9] */
  union REGS r;
  if (__IsPharlap())
  {
    struct SREGS s;
    segread( &s); 
    r.x.eax = 0x2504;
    r.x.ecx = intno;
    r.x.edx = pmaddr->offset32;
    s.ds = pmaddr->selector;
    int386x (0x21, &r, &r, &s);
    return 0;
  }
  r.x.eax = 0x205;
  r.x.ebx = intno;
  r.w.cx = pmaddr->selector;
  r.x.edx = pmaddr->offset32;
  int386 (0x31, &r, &r); 
  if (r.w.cflag & 1)
    return -1; /* 0x8022 = invalid selector */
  return (0);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_simulate_real_mode_interrupt (int intno, __dpmi_regs *reg)
{
  union REGS  r;
  struct SREGS s;
  memset (&r, 0, sizeof(r));
  segread (&s);
  if (__IsPharlap())
  {
    unsigned char rpack[0x12];
    *((unsigned short *)&rpack[0x00]) = (unsigned short)intno;
    *((unsigned short *)&rpack[0x02]) = reg->x.ds;
    *((unsigned short *)&rpack[0x04]) = reg->x.es;
    *((unsigned short *)&rpack[0x06]) = reg->x.fs;
    *((unsigned short *)&rpack[0x08]) = reg->x.gs;
    *((unsigned long  *)&rpack[0x0a]) = reg->d.eax;
    *((unsigned long  *)&rpack[0x0e]) = reg->d.edx;
    r.x.ecx = reg->d.ecx; 
    r.x.ebx = reg->d.ebx; 
    r.x.esi = reg->d.esi; 
    r.x.edi = reg->d.edi; 
    s.ds = s.ss;
    r.x.edx = (unsigned long)(&rpack[0]);
    r.x.eax = 0x2511;
    int386x(0x21, &r, &r, &s);
    reg->x.flags = (short)r.w.cflag;
    reg->d.ecx = r.x.ecx; 
    reg->d.ebx = r.x.ebx; 
    reg->d.esi = r.x.esi; 
    reg->d.edi = r.x.edi; 
    reg->x.ds  = *((unsigned short *)&rpack[0x02]);
    reg->x.es  = *((unsigned short *)&rpack[0x04]);
    reg->x.fs  = *((unsigned short *)&rpack[0x06]);
    reg->x.gs  = *((unsigned short *)&rpack[0x08]);
    reg->d.eax = *((unsigned long  *)&rpack[0x0a]);
    reg->d.edx = *((unsigned long  *)&rpack[0x0e]);
    return 0;
  }
  r.x.eax = 0x300;
  r.x.ebx = intno;
  r.x.ecx = 0;
  s.es    = s.ds;
  r.x.edi = (unsigned)reg;
  int386x (0x31, &r, &r, &s);
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_int( int intr, __dpmi_regs *reg)
{
  reg->x.flags = 0; reg->x.ss = reg->x.sp = 0; /* DPMI host provides stack */
  return __dpmi_simulate_real_mode_interrupt (intr, reg);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__)
static int __lk_unlk_linear_region(int do_lk,unsigned long linaddr,unsigned len)
{
#if defined(__DJGPP__)
  __dpmi_meminfo info;
  info.handle = 0;
  info.size = len;
  info.address = linaddr;
  if (do_lk)
    return __dpmi_lock_linear_region( &info );
  return __dpmi_unlock_linear_region( &info );
#else
  union REGS r;
  if (__IsPharlap())
  {
    r.x.eax = 0x251A;
    if (!do_lk)
      r.x.eax++;
    r.x.edx = (((linaddr & 0xfff) + len) + 4095)>>12;
    r.x.ecx = linaddr - (linaddr & 0xfff);
    r.x.ebx = 0;
    int386 (0x21, &r, &r);
    if (r.w.cflag & 1)
      return -1; /* 8=ENOMEM, 9=bad addr */
    return 0; 
  }
  r.x.eax = 0x600;                /* 0x0600: DPMI Lock Linear Region */
  if (!do_lk) r.x.eax++;          /* 0x0601 DPMI Unlock Linear Region */ 
  r.x.ebx = (linaddr >> 16);      /* Linear address in BX:CX */
  r.x.ecx = (linaddr & 0xFFFF);
  r.x.esi = (len >> 16);       /* Length in SI:DI */
  r.x.edi = (len & 0xFFFF);
  int386 (0x31, &r, &r);
  if (r.w.cflag & 1)
    return -1;
  return 0; 
#endif
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_lock_linear_region (unsigned long addr,unsigned len)
{                                     /* Int 31H Function 0600H [0.9] */
  return __lk_unlk_linear_region(1,addr,len);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_unlock_linear_region (unsigned long addr, unsigned len)
{                                     /* Int 31H Function 0601H [0.9] */
  return __lk_unlk_linear_region(0,addr,len);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_simulate_real_mode_procedure_iret(__dpmi_regs *regs)
{                                     /* Int 31H Function 0302H [0.9] */
  union  REGS  r;
  struct SREGS s;
  segread (&s);
  r.w.ax  = 0x302;
  /* ES:(E)DI= selector:offset of 32H-byte reg buffer */
  s.es    = s.ds;
  r.x.edi = (unsigned)regs;
  r.x.ebx = 0; /* no flags */
  r.x.ecx = 0; /* nothing to copy over stack */
  int386x (0x31, &r, &r, &s);
  if (r.w.cflag & 1)
     return -1;
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_simulate_real_mode_procedure_retf(__dpmi_regs *regs)
{                                     /* Int 31H Function 0301H [0.9] */
  union  REGS  r;
  struct SREGS s;
  segread (&s);
  r.w.ax  = 0x301;
  /* ES:(E)DI= selector:offset of 32H-byte reg buffer */
  s.es    = s.ds;
  r.x.edi = (unsigned)regs;
  r.x.ebx = 0; /* no flags */
  r.x.ecx = 0; /* nothing to copy over stack */
  int386x (0x31, &r, &r, &s);
  if (r.w.cflag & 1)
     return -1;
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_allocate_real_mode_callback(auto void (*_handler)(void), 
                                       __dpmi_regs *regs,
                                       __dpmi_raddr *rmaddr ) 
{                                    /* Int 31H Function 0303H [0.9] */
  union  REGS  r;
  struct SREGS s;

  memset (&r, 0, sizeof(r));
  segread (&s);
  r.w.ax  = 0x303;
  /* ES:(E)DI= selector:offset of 32H-byte reg buffer */
  s.es    = s.ds;
  r.x.edi = (unsigned)regs;
  /* DS:(E)SI= selector:offset of protected mode procedure to call */
  s.ds    = s.cs;
  r.x.esi = (unsigned)_handler;
  int386x (0x31, &r, &r, &s);
  if (r.w.cflag & 1)
     return -1;
  rmaddr->segment = r.w.cx;
  rmaddr->offset16 = r.w.dx;
  return 0; /* success */
}  
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_free_real_mode_callback(__dpmi_raddr *rmaddr)
{
  union REGS r;
  r.x.eax = 0x304;
  r.w.cx = rmaddr->segment;
  r.w.dx = rmaddr->offset16;
  int386 (0x31, &r, &r); 
  if (r.w.cflag & 1)
    return -1;
  return 0;
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_and_set_virtual_interrupt_state(int newstate)
{
  union REGS r;
  if (__IsPharlap())
  {
    int oldstate = __dpmi_get_virtual_interrupt_state();
    if (oldstate && !newstate)
      __asm cli ;
    if (!oldstate && newstate)
      __asm sti ;
    return oldstate;
  }
  r.x.eax = 0x0900;
  if (newstate)
    r.x.eax |= 1;
  int386 (0x31, &r, &r);
  return (0xff & r.x.eax);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_and_disable_virtual_interrupt_state(void)
{
  return __dpmi_get_and_set_virtual_interrupt_state(0);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_and_enable_virtual_interrupt_state(void)
{
  return __dpmi_get_and_set_virtual_interrupt_state(1);
}
#endif

/* -------------------------------------------------------------------- */

#if defined(__DPMI32__) && !defined(__DJGPP__)
int __dpmi_get_virtual_interrupt_state(void)
{
  union REGS r;
  if (__IsPharlap())
  {
    r.x.eax = 0x2534;
    int386 (0x21, &r, &r); 
    return (0xff & r.x.eax);  
  }
  r.x.eax = 0x0902;
  int386 (0x31, &r, &r); 
  return (0xff & r.x.eax);  
}
#endif /* defined(__DPMI32__) && !defined(__DJGPP__) */

/* ==================================================================== */
/*                            move data                                 */
/* ==================================================================== */

#if 0 /* defined(__DPMI32__) && !defined(__DJGPP__) */
      /* INLINE CODE in wdpmi.h */
void dosmemget(unsigned long linoffset, size_t length, void *buffer)
{
  movedata( ___my_dpmi_get_dosds(), linoffset,
            ___my_dpmi_get_ds(), (unsigned)buffer, length );
}
#endif

/* -------------------------------------------------------------------- */

#if 0 /* defined(__DPMI32__) && !defined(__DJGPP__) */
      /* INLINE CODE in wdpmi.h */      
void dosmemput(const void *buffer, size_t length, unsigned long linoffset)
{
  movedata( ___my_dpmi_get_ds(), (unsigned)buffer,
            ___my_dpmi_get_dosds(), linoffset, length );
}
#endif 

/* ===================================================================== */
/*                            lock wrappers                              */
/* ===================================================================== */

#if defined(__DPMI32__)
static int __lock_unlock_based_address(int do_lock, unsigned segsel,
                                 const void *address, unsigned int length)
{
  unsigned long base;
  int rc = __dpmi_get_segment_base_address( segsel, &base );
  if (rc == 0)
  {
    base += (unsigned long)address;
    rc = __lk_unlk_linear_region(do_lock, base, length);
  } 
  return rc;
}
                                    
int __my_dpmi_lock_code (const void *address, unsigned length)
{
  return __lock_unlock_based_address(1, __my_dpmi_get_cs(), address, length );
}

int __my_dpmi_lock_data (const void *address, unsigned length)
{
  return __lock_unlock_based_address(1, __my_dpmi_get_ds(), address, length );
}

int __my_dpmi_unlock_code (const void *address, unsigned length)
{
  return __lock_unlock_based_address(0, __my_dpmi_get_cs(), address, length );
}

int __my_dpmi_unlock_data (const void *address, unsigned length)
{
  return __lock_unlock_based_address(0, __my_dpmi_get_ds(), address, length );
}

#endif /* __DPMI32__ */

/* ======================================================================= */
/*                     DPMI callback/interrupt helper                      */
/* ======================================================================= */

#if defined(__DPMI32__) /* all DPMI extenders including pharlap */
static int __assert_cs_base_and_limit(unsigned int safe_ds)
{
  /* Make sure our wrappers ... 
  ** 1) are cs addressable. Technically speaking, the 
  **    RM callback wrapper is allowed to be in DS, but 
  **    PM isr handlers must be in code.
  ** 2) That the near calls made to the pm_entry don't 
  **    blow up (that pm_entry is near). 
  ** Since we are FLAT, these test should always succeed.
  */ 
  unsigned int cssel = __my_dpmi_get_cs();
  unsigned int dssel = __my_dpmi_get_ds();
  unsigned long csbase, dsbase, cslimit, dslimit;

  /* Our limits comparison below would be meaningless 
  ** unless bases were the same - should always be so
  ** (by definition) for FLAT.
  */
  if (__dpmi_get_segment_base_address ( cssel, &csbase ) || 
      __dpmi_get_segment_base_address ( dssel, &dsbase ))
  {
    return -1;
  }
  if (csbase != dsbase)
  {
    return -1;
  }
  cslimit = __dpmi_get_segment_limit(cssel);
  dslimit = __dpmi_get_segment_limit(dssel);
  if (cslimit == 0 || dslimit == 0)
  {
    return -1;
  }
  if (cslimit < dslimit)
  {
    if (__dpmi_set_segment_limit(cssel, dslimit) != 0)
      return -1;
  }
  if (safe_ds != dssel) /* probably only necessary for DJGPP */
  {
    cssel = safe_ds;
    if (__dpmi_get_segment_base_address ( cssel, &csbase ) != 0)
      return -1;
    if (csbase != dsbase)
      return -1;
    cslimit = __dpmi_get_segment_limit(cssel);
    if (cslimit == 0)
      return -1;
    if (cslimit < dslimit)
    {
      if (__dpmi_set_segment_limit(cssel, dslimit) != 0)
        return -1;
    }
  }
  return 0;
}
#endif

/* ======================================================================= */
/*                   DPMI protected mode interrupt wrapper                 */
/* ======================================================================= */

#if defined(__DPMI32__) /* all DPMI extenders including pharlap */
#include <setjmp.h>

static void __pm_isr_helper(__dpmi_regs *regs)
{
#if defined(__WATCOMC__)
  if (setjmp( ((unsigned *)(((unsigned char *)regs)+sizeof(*regs))) ) == 0)
#elif defined(__DJGPP__)
  if (setjmp( (((jmp_buf *)(((unsigned char *)regs)+sizeof(*regs)))[0]) )==0)
#else
  #error "adjust this to suit your compiler's setjmp.h"
  if (setjmp( ((jmp_buf)(((unsigned char *)regs)+sizeof(*regs))) ) == 0)
#endif
  {
    typedef void (*xxx_isr_t)(__dpmi_regs *);
    xxx_isr_t isr = (xxx_isr_t)(regs->d.res);
    regs->d.res = (long)__pm_isr_helper;
    (*isr)( regs );
    /*doesn't get here if __my_dpmi_chain_pmiv() is called */
    regs->d.res = (long)isr;
  }
  return;
}

/* -------------------------------------------------------------------- */

void __my_dpmi_chain_pmiv(__dpmi_regs *regs)
{
  if (regs && regs->d.res == ((unsigned long)__pm_isr_helper))
  {
#if defined(__WATCOMC__)
    longjmp( ((unsigned *)(((unsigned char *)regs)+sizeof(*regs))), 1 );
#elif defined(__DJGPP__)
    longjmp( (((jmp_buf *)(((unsigned char *)regs)+sizeof(*regs)))[0]), 1 );
#else
    #error "adjust this to suit your compiler's setjmp.h"
    longjmp( ((jmp_buf)(((unsigned char *)regs)+sizeof(*regs))), 1 );
#endif
  }
  /* _printk("PANIC! bad arg to __my_dpmi_chain_pmiv()\n"); */
} 

/* -------------------------------------------------------------------- */

static int __lock_unlock_pm_isr_helpers(int do_lock)
{
  static unsigned int __pm_isr_helper_lock_count = 0;
  if (do_lock)
  {
    if (__pm_isr_helper_lock_count == 0)
    {
      /* We have to lock functions separately because
      ** Watcom's unused function removal feature gives the
      ** compiler/linker the liberty to shuffle things about,
      ** and the two functions are not necessarily anywhere
      ** near each other at run time.
      */
      if (__my_dpmi_lock_code((const void *)__pm_isr_helper,256) != 0)
        __pm_isr_helper_lock_count--;
      else if (__my_dpmi_lock_code((const void *)__my_dpmi_chain_pmiv,256) != 0)
      {
        __my_dpmi_unlock_code((const void *)__pm_isr_helper,256);
        __pm_isr_helper_lock_count--;
      }
    }
    if ((++__pm_isr_helper_lock_count) == 0)
    {
      --__pm_isr_helper_lock_count;
      return -1;
    }
  }
  else if (((int)__pm_isr_helper_lock_count) != -1)
  {
    if ((--__pm_isr_helper_lock_count) == 0)
    {
      __my_dpmi_unlock_code((const void *)__pm_isr_helper,256);
      __my_dpmi_unlock_code((const void *)__my_dpmi_chain_pmiv,256);
    }
  }
  return 0;
}

/* -------------------------------------------------------------------- */

int __my_dpmi_hook_pmiv( int vector, void (*user_isr)(__dpmi_regs *regs) )
{
  static const unsigned char wrapper[] = {
  /*0000*/ 0xEB,0x2E,                  /* jmp       wrapper_entry     */
  /*0002*/ 0x00,0x00,                  /* padding     : db 0, 0       */
#define PMISRWO_SIG             0x04
  /*0004*/ 0x00,0x00,0x00,0x00,        /* signature   : dd 0          */
  /*0008*/ 0x00,0x00,0x00,0x00,        /* reentry_chk : dd 0          */
#define PMISRWO_WRAPSIZE        0x0C
  /*000C*/ 0x00,0x00,0x00,0x00,        /* wrapper_size: dd 0          */
#define PMISRWO_CODESIZE        0x10
  /*0010*/ 0x00,0x00,0x00,0x00,        /* code_size   : dd 0          */
#define PMISRWO_STACKLOWPTR     0x14
  /*0014*/ 0x00,0x00,0x00,0x00,        /* STACKLOW_ptr: dd 0          */
  /*0018*/ 0x00,0x00,0x00,0x00,        /* old_STACKLOW: dd 0          */
#define PMISRWO_VECTOR          0x1C
  /*001C*/ 0x00,0x00,0x00,0x00,        /* vector      : dd 0          */
#define PMISRWO_ISRHELPER       0x20
  /*0020*/ 0x00,0x00,0x00,0x00,        /* isr_helper  : dd 0          */
#define PMISRWO_USERISRPTR      0x24
  /*0024*/ 0x00,0x00,0x00,0x00,        /* user_isr    : dd 0          */
#define PMISRWO_CHAIN_OFF       0x28 
  /*0028*/ 0x00,0x00,0x00,0x00,        /* oldisr_offset: dd 0         */
#define PMISRWO_CHAIN_SEL       0x2C 
  /*002C*/ 0x00,0x00,0x00,0x00,        /* oldisr_segment: dd 0        */

  /*0030*/                             /* X$13:                       */
  /*0030*/ 0x90,                       /* nop                         */
  /*0031*/ 0x90,                       /* nop                         */
  /*0032*/ 0x90,                       /* nop                         */
  /*0033*/ 0x90,                       /* nop                         */
  /*0034*/ 0x9C,                       /* pushfd                      */
  /*0035*/ 0x1E,                       /* push      ds                */
  /*0036*/ 0x50,                       /* push      eax               */
  /*0037*/ 0x53,                       /* push      ebx               */
#define PMISRWO_SELF_OFF        (0x38+1)
  /*0038*/ 0xBB,0x78,0x56,0x34,0x12,   /* mov       ebx,0x12345678    */
#define PMISRWO_SELF_SEL        (0x3D+1)
  /*003D*/ 0xB8,0x78,0x56,0x34,0x12,   /* mov       eax,0x12345678    */
  /*0042*/ 0x8E,0xD8,                  /* mov       ds,ax             */
  /*0044*/ 0xFE,0x43,0x08,             /* inc       byte ptr [ebx+0x8]*/
  /*0047*/ 0x80,0x7B,0x08,0x01,        /* cmp       byte ptr [ebx+0x8],0x01*/
  /*004B*/ 0x0F,0x85,0xF9,0x00,0x00,0x00, /* jne       X$16           */
  /*0051*/ 0x89,0xF8,                  /* mov       eax,edi           */
  /*0053*/ 0x89,0xDF,                  /* mov       edi,ebx           */
  /*0055*/ 0x03,0x7B,0x10,             /* add       edi,[ebx+0x10]    */
  /*0058*/ 0x83,0xC7,0x04,             /* add       edi,0x00000004    */
  /*005B*/ 0x83,0xE7,0xFC,             /* and       edi,0xfffffffc    */
  /*005E*/ 0x89,0x07,                  /* mov       [edi],eax         */
  /*0060*/ 0x89,0x77,0x04,             /* mov       [edi+0x4],esi     */
  /*0063*/ 0x89,0x6F,0x08,             /* mov       [edi+0x8],ebp     */
  /*0066*/ 0x8B,0x43,0x24,             /* mov       eax,[ebx+0x24]    */
  /*0069*/ 0x89,0x47,0x0C,             /* mov       [edi+0xc],eax     */
  /*006C*/ 0x8B,0x04,0x24,             /* mov       eax,[esp]         */
  /*006F*/ 0x89,0x47,0x10,             /* mov       [edi+0x10],eax    */
  /*0072*/ 0x89,0x57,0x14,             /* mov       [edi+0x14],edx    */
  /*0075*/ 0x89,0x4F,0x18,             /* mov       [edi+0x18],ecx    */
  /*0078*/ 0x8B,0x44,0x24,0x04,        /* mov       eax,[esp+0x4]     */
  /*007C*/ 0x89,0x47,0x1C,             /* mov       [edi+0x1c],eax    */
  /*007F*/ 0x8B,0x44,0x24,0x0C,        /* mov       eax,[esp+0xc]     */
  /*0083*/ 0x89,0x47,0x20,             /* mov       [edi+0x20],eax    */
  /*0086*/ 0x66,0x8C,0x47,0x22,        /* mov       [edi+0x22],es     */
  /*008A*/ 0x8B,0x44,0x24,0x08,        /* mov       eax,[esp+0x8]     */
  /*008E*/ 0x89,0x47,0x24,             /* mov       [edi+0x24],eax    */
  /*0091*/ 0x66,0x8C,0x67,0x26,        /* mov       [edi+0x26],fs     */
  /*0095*/ 0x66,0x8C,0x6F,0x28,        /* mov       [edi+0x28],gs     */
  /*0099*/ 0x8B,0x44,0x24,0x10,        /* mov       eax,[esp+0x10]    */
  /*009D*/ 0x89,0x47,0x2A,             /* mov       [edi+0x2a],eax    */
  /*00A0*/ 0x8B,0x44,0x24,0x14,        /* mov       eax,[esp+0x14]    */
  /*00A4*/ 0x89,0x47,0x2C,             /* mov       [edi+0x2c],eax    */
  /*00A7*/ 0x66,0x89,0x67,0x2E,        /* mov       [edi+0x2e],sp     */
  /*00AB*/ 0x66,0x8C,0x57,0x30,        /* mov       [edi+0x30],ss     */
  /*00AF*/ 0x8B,0x43,0x14,             /* mov       eax,[ebx+0x14]    */
  /*00B2*/ 0x09,0xC0,                  /* or        eax,eax           */
  /*00B4*/ 0x74,0x12,                  /* je        X$14              */
  /*00B6*/ 0x52,                       /* push      edx               */
  /*00B7*/ 0x89,0xFA,                  /* mov       edx,edi           */
  /*00B9*/ 0x81,0xC2,0x80,0x00,0x00,0x00, /* add       edx,0x00000080 */
  /*00BF*/ 0x83,0xE2,0xFC,             /* and       edx,0xfffffffc    */
  /*00C2*/ 0x87,0x10,                  /* xchg      edx,[eax]         */
  /*00C4*/ 0x89,0x53,0x18,             /* mov       [ebx+0x18],edx    */
  /*00C7*/ 0x5A,                       /* pop       edx               */
  /*00C8*/                             /* X$14:                       */
  /*00C8*/ 0x1E,                       /* push      ds                */
  /*00C9*/ 0x53,                       /* push      ebx               */
  /*00CA*/ 0x57,                       /* push      edi               */
  /*00CB*/ 0x89,0xDD,                  /* mov       ebp,ebx           */
  /*00CD*/ 0x03,0x6B,0x0C,             /* add       ebp,[ebx+0xc]     */
  /*00D0*/ 0x83,0xED,0x04,             /* sub       ebp,0x00000004    */
  /*00D3*/ 0x83,0xE5,0xFC,             /* and       ebp,0xfffffffc    */
  /*00D6*/ 0x66,0x8C,0xD1,             /* mov       cx,ss             */
  /*00D9*/ 0x66,0x8C,0xD8,             /* mov       ax,ds             */
  /*00DC*/ 0x87,0xEC,                  /* xchg      ebp,esp           */
  /*00DE*/ 0x8E,0xD0,                  /* mov       ss,ax             */
  /*00E0*/ 0x8E,0xC0,                  /* mov       es,ax             */
  /*00E2*/ 0x8E,0xE0,                  /* mov       fs,ax             */
  /*00E4*/ 0x8E,0xE8,                  /* mov       gs,ax             */
  /*00E6*/ 0xFC,                       /* cld                         */
  /*00E7*/ 0x51,                       /* push      ecx               */
  /*00E8*/ 0x55,                       /* push      ebp               */
  /*00E9*/ 0x89,0xE5,                  /* mov       ebp,esp           */
  /*00EB*/ 0x89,0xF8,                  /* mov       eax,edi           */
  /*00ED*/ 0x50,                       /* push      eax               */
  /*00EE*/ 0xFF,0x53,0x20,             /* call      dword ptr [ebx+0x20]*/
  /*00F1*/ 0xFA,                       /* cli                         */
  /*00F2*/ 0x89,0xEC,                  /* mov       esp,ebp           */
  /*00F4*/ 0x5D,                       /* pop       ebp               */
  /*00F5*/ 0x59,                       /* pop       ecx               */
  /*00F6*/ 0x89,0xEC,                  /* mov       esp,ebp           */
  /*00F8*/ 0x8E,0xD1,                  /* mov       ss,cx             */
  /*00FA*/ 0x5F,                       /* pop       edi               */
  /*00FB*/ 0x5B,                       /* pop       ebx               */
  /*00FC*/ 0x1F,                       /* pop       ds                */
  /*00FD*/ 0x8B,0x43,0x14,             /* mov       eax,[ebx+0x14]    */
  /*0100*/ 0x09,0xC0,                  /* or        eax,eax           */
  /*0102*/ 0x74,0x05,                  /* je        X$15              */
  /*0104*/ 0x8B,0x53,0x18,             /* mov       edx,[ebx+0x18]    */
  /*0107*/ 0x89,0x10,                  /* mov       [eax],edx         */
  /*0109*/                             /* X$15:                       */
  /*0109*/ 0x8E,0x6F,0x28,             /* mov       gs,[edi+0x28]     */
  /*010C*/ 0x8E,0x67,0x26,             /* mov       fs,[edi+0x26]     */
  /*010F*/ 0x66,0x8B,0x57,0x24,        /* mov       dx,[edi+0x24]     */
  /*0113*/ 0x8E,0x47,0x22,             /* mov       es,[edi+0x22]     */
  /*0116*/ 0x66,0x8B,0x4F,0x20,        /* mov       cx,[edi+0x20]     */
  /*011A*/ 0x66,0x89,0x54,0x24,0x08,   /* mov       [esp+0x8],dx      */
  /*011F*/ 0x66,0x89,0x4C,0x24,0x0C,   /* mov       [esp+0xc],cx      */
  /*0124*/ 0x66,0x89,0x4C,0x24,0x18,   /* mov       [esp+0x18],cx     */
  /*0129*/ 0x8B,0x47,0x1C,             /* mov       eax,[edi+0x1c]    */
  /*012C*/ 0x8B,0x4F,0x10,             /* mov       ecx,[edi+0x10]    */
  /*012F*/ 0x89,0x0C,0x24,             /* mov       [esp],ecx         */
  /*0132*/ 0x89,0x44,0x24,0x04,        /* mov       [esp+0x4],eax     */
  /*0136*/ 0x8B,0x4F,0x18,             /* mov       ecx,[edi+0x18]    */
  /*0139*/ 0x8B,0x57,0x14,             /* mov       edx,[edi+0x14]    */
  /*013C*/ 0x8B,0x47,0x0C,             /* mov       eax,[edi+0xc]     */
  /*013F*/ 0x8B,0x6F,0x08,             /* mov       ebp,[edi+0x8]     */
  /*0142*/ 0x8B,0x77,0x04,             /* mov       esi,[edi+0x4]     */
  /*0145*/ 0x8B,0x3F,                  /* mov       edi,[edi]         */
  /*0147*/ 0x3B,0x43,0x24,             /* cmp       eax,[ebx+0x24]    */
  /*014A*/                             /* X$16:                       */
  /*014A*/ 0x9F,                       /* lahf                        */
  /*014B*/ 0xFE,0x4B,0x08,             /* dec       byte ptr [ebx+0x8]*/
  /*014E*/ 0x9E,                       /* sahf                        */
  /*014F*/ 0x74,0x14,                  /* je        X$17              */
  /*0151*/ 0x8B,0x43,0x28,             /* mov       eax,[ebx+0x28]    */
  /*0154*/ 0x0B,0x43,0x2C,             /* or        eax,[ebx+0x2c]    */
  /*0157*/ 0x74,0x0C,                  /* je        X$17              */
  /*0159*/ 0x5B,                       /* pop       ebx               */
  /*015A*/ 0x58,                       /* pop       eax               */
  /*015B*/ 0x1F,                       /* pop       ds                */
  /*015C*/ 0x9D,                       /* popfd                       */
#define PMISRWO_JMP_FIXUP       (0x15D+3)
  /*015D*/ 0x2E,0xFF,0x2D,0x90,0x90,0x90,0x90, /* jmp       fword ptr cs:0x90909090*/
  /*0164*/ 0xCF,                       /* iretd                       */
  /*0165*/                             /* X$17:                       */
  /*0165*/ 0x5B,                       /* pop       ebx               */
  /*0166*/ 0x58,                       /* pop       eax               */
  /*0167*/ 0x1F,                       /* pop       ds                */
  /*0168*/ 0x9D,                       /* popfd                       */
  /*0169*/ 0xCF                        /* iretd                       */
  };

  if (vector >= 0 && vector <= 0xff && user_isr)
  {
    if (__lock_unlock_pm_isr_helpers(1) == 0)
    {
      unsigned short wrap_sel;
      #if defined(__DJGPP__) /* signal safe ds */
      wrap_sel = (unsigned short)__djgpp_ds_alias;
      #else
      wrap_sel = (unsigned short)__my_dpmi_get_ds();
      #endif

      if (__assert_cs_base_and_limit(wrap_sel) == 0)
      {
        #define PMISR_STACK_SIZE 8192u

        unsigned char *wrap;
        unsigned int wrap_size, code_size;

        code_size = sizeof(wrapper) + (4 - (sizeof(wrapper) & 3));
        wrap_size = code_size + PMISR_STACK_SIZE;

        wrap = (unsigned char *)malloc( wrap_size );
        if (wrap)
        {
          if (__my_dpmi_lock_data (wrap, wrap_size ) == 0)
          { 
            __dpmi_paddr pm_int;
            memset( wrap, 0, wrap_size );
            memcpy( wrap, wrapper, code_size );

            __dpmi_get_protected_mode_interrupt_vector(vector, &pm_int);
   
            *((unsigned char**)&wrap[PMISRWO_SIG])       = &wrap[PMISRWO_SIG];
            *((unsigned long *)&wrap[PMISRWO_WRAPSIZE])  = wrap_size;
            *((unsigned long *)&wrap[PMISRWO_CODESIZE])  = code_size;
            #if defined(__WATCOMC__)
            *((unsigned long *)&wrap[PMISRWO_STACKLOWPTR])= (long)&_STACKLOW;
            #elif defined(__DJGPP__)
            //*((unsigned long *)&wrap[PMISRWO_STACKLOWPTR])= (long)&__djgpp_stack_limit;
            #endif
            *((unsigned long *)&wrap[PMISRWO_VECTOR])    = vector;
            *((void **        )&wrap[PMISRWO_ISRHELPER]) = (void *)__pm_isr_helper;
            *((void **        )&wrap[PMISRWO_USERISRPTR])= (void *)user_isr;
            *((unsigned char**)&wrap[PMISRWO_SELF_OFF])  = wrap;
            *((unsigned long *)&wrap[PMISRWO_SELF_SEL])  = wrap_sel;

            *((unsigned long *)&wrap[PMISRWO_CHAIN_OFF]) = pm_int.offset32;
            *((unsigned long *)&wrap[PMISRWO_CHAIN_SEL]) = pm_int.selector;
            *((unsigned char**)&wrap[PMISRWO_JMP_FIXUP]) = &wrap[PMISRWO_CHAIN_OFF];

            pm_int.selector = (unsigned short)__my_dpmi_get_cs();
            pm_int.offset32 = (unsigned long)wrap;
            if (__dpmi_set_protected_mode_interrupt_vector(vector, &pm_int) == 0)
            {
              return (int)wrap;
            }
            __my_dpmi_unlock_data (wrap, wrap_size );

          } /* if (__my_dpmi_lock_data (wrap, wrap_size ) == 0) */
          free(wrap);
        } /* if (wrap) */
      } /* if (__assert_cs_base_and_limit(wrap_sel) == 0) */
      __lock_unlock_pm_isr_helpers(0);
    } /* __lock_unlock_pm_isr_helpers(1) */
  } /* if (vector >= 0 && vector <= 0xff) */
  return 0;
}

/* -------------------------------------------------------------------- */

int __my_dpmi_unhook_pmiv( int hook )
{
  if (hook)
  {
    unsigned char *wrap = (unsigned char *)hook;
    if ( *((unsigned char**)&wrap[PMISRWO_SIG]) == &wrap[PMISRWO_SIG] 
      && *((unsigned char**)&wrap[PMISRWO_SELF_OFF]) == wrap )
    { 
      __dpmi_paddr pm_int;      
      unsigned int wrap_size = *((unsigned int *)(&wrap[PMISRWO_WRAPSIZE]));
      unsigned int vector = *((unsigned int *)(&wrap[PMISRWO_VECTOR]));
      pm_int.offset32 = *((unsigned long *)&wrap[PMISRWO_CHAIN_OFF]);
      pm_int.selector = *((unsigned short *)&wrap[PMISRWO_CHAIN_SEL]);

      if (__dpmi_set_protected_mode_interrupt_vector(vector, &pm_int) == 0)
      {
        __my_dpmi_unlock_data (wrap, wrap_size );
        __lock_unlock_pm_isr_helpers(0);
        free( wrap ); 
        return 0;
      }
    }
  }          
  return -1;    
}
#endif /* __DPMI32__ */

/* ======================================================================= */
/*                     DPMI real-mode callback wrapper                     */
/* ======================================================================= */

#if defined(__DPMI32__) && /* all DPMI extenders including pharlap */ \
   !defined(DONT_WANT_RMCB_SUPPORT)

/* **WARNING!** **WARNING!** **WARNING!** NTVDM trashes EIP if */
/* an RMCB gets called while an Int 21h call is in progress. In our */
/* RMCB wrapper this results in DS corruption (other rmcb wrappers, eg */
/* GO32, have other registers that get trashed) */

#define FILL 0x00

#define WRAPTYPE_RMCB_RAW     0  /* pm isr */
#define WRAPTYPE_RMCB_CHAIN   1  /* real mode isr + chain */
#define WRAPTYPE_RMCB_IRET    2  /* real mode isr + iret */
#define WRAPTYPE_RMCB_RETF    3  /* real mode far call */

static void *__mk_wrapper(int wrap_type, int vector, void (*pm_entry)(void),
                          __dpmi_regs *rmcb_regs )
{
  #define WRAPOFF_SIGNATURE   0x04
  #define WRAPOFF_REENTRY_CHK 0x08
  #define WRAPOFF_WRAPSIZE    0x0C
  #define WRAPOFF_CODESIZE    0x10
  #define WRAPOFF_STACKLOWPTR 0x14
  #define WRAPOFF_RMCB_PTR    0x1C
  #define WRAPOFF_WRAP_TYPE   0x20
  #define WRAPOFF_VECTOR      0x24
  #define WRAPOFF_OLDISR      0x28
  #define WRAPOFF_REGSPTR     0x2C
  #define WRAPOFF_PMISR_OFF32 0x30 
  #define WRAPOFF_PMISR_SEL   0x34

  #define WRAPPER_SIG_RMCB ((((long)'W')<<24)+(((long)'R')<<16)+(((long)'A')<<8)+'P')
  static unsigned char wrapper[] = {
  /*0000*/ 0xEB,0x36,                  /* jmp short wrapper_enty      */
  /*0002*/ 0x00,0x00,                  /* padding:      dw 0          */
  /*0004*/    FILL,FILL,FILL,FILL,     /* signature:    dd 0          */
  /*0008*/    FILL,FILL,FILL,FILL,     /* reentry_chk:  dd 0          */
  /*000C*/    FILL,FILL,FILL,FILL,     /* wrapper_size: dd 0          */
  /*0010*/    FILL,FILL,FILL,FILL,     /* code_size:    dd 0          */
  /*0014*/    FILL,FILL,FILL,FILL,     /* STACKLOW_ptr: dd 0          */
  /*0018*/    FILL,FILL,FILL,FILL,     /* old_STACKLOW: dd 0          */
  /*001C*/    FILL,FILL,FILL,FILL,     /* pm_entry:     dd 0          */
  /*0020*/    FILL,FILL,FILL,FILL,     /* wrapper_type: dd 0          */
  /*0024*/    FILL,FILL,FILL,FILL,     /* vector:       dd 0          */
  /*0028*/    FILL,FILL,FILL,FILL,     /* old_isr:      dd 0          */
  /*002C*/    FILL,FILL,FILL,FILL,     /* regs_ptr:     dd 0          */
  /*0030*/ 0x00,0x00,0x00,0x00,        /* oldpmisr_off: dd 0          */
  /*0034*/ 0x00,0x00,0x00,0x00,        /* oldpmisr_sel: dd 0          */

  /*0038*/                             /* wrapper_enty:               */
  /*0038*/ 0x50,                       /* push      eax               */
  /*0039*/ 0x53,                       /* push      ebx               */

  /*003A*/ 0x55,                       /* push      ebp               */
  /*003B*/ 0x51,                       /* push      ecx               */
  /*003C*/ 0x52,                       /* push      edx               */
  /*003D*/ 0x1E,                       /* push      ds                */
  #define WRAPOFF_RMCB_SELF_OFF32  0x3F
  #define WRAPOFF_RMCB_SELF_SEL    0x44
  /*003E*/ 0xBB,  FILL,FILL,FILL,FILL, /* mov       ebx,<wrap_off32>  */
  /*0043*/ 0xB8,  FILL,FILL,FILL,FILL, /* mov       eax,<wrap_sel>    */
  /*0048*/ 0x66,0x8C,0xD9,             /* mov       cx,ds             */
  /*004B*/ 0x8E,0xD8,                  /* mov       ds,ax             */
  /*004D*/ 0x0F,0xB7,0xC9,             /* movzx     ecx,cx            */

  /* the next 5 lines are a fix for NTVDM (even on Win2K) to         **/
  /* workaround a NTVDM's DPMI server bug - only the the SI and DI   **/
  /* parts of ESI and EDI are valid.                                 **/

  /*0050*/ 0x89,0xC8,                  /* mov       eax,ecx           */
  /*0052*/ 0x0F,0x03,0xC0,             /* lsl       eax,eax           */
  /*0055*/ 0x39,0xF0,                  /* cmp       eax,esi           */
  /*0057*/ 0x77,0x06,                  /* ja        end_NTfix         */
  /*0059*/ 0x0F,0xB7,0xF6,             /* movzx     esi,si            */
  /*005C*/ 0x0F,0xB7,0xFF,             /* movzx     edi,di            */
  /*005F*/                             /* end_NTfix:                  */

  /*005F*/ 0xFF,0x43,0x08,             /* inc       dword ptr [ebx+0x8]*/
  /*0062*/ 0x83,0x7B,0x08,0x01,        /* cmp       dword ptr [ebx+0x8],1*/
  /*0066*/ 0x75,0x62,                  /* jne       wrapper_fini      */

  /** WatcomC has a stack overflow guard global called _STACKLOW which   **/
  /** needs to be adjusted before stack switch, and restored thereafter. **/
  /** new_esp      := wrapper_offset + wrapper_size, so...               **/
  /** new_STACKLOW := new_esp - code_size := wrapper_offset + code_size  **/

  /*0068*/ 0x8B,0x53,0x14,             /* mov       edx,[ebx+0x14]    */
  /*006B*/ 0x09,0xD2,                  /* or        edx,edx           */
  /*006D*/ 0x74,0x10,                  /* je        _STACKLOW_set_end */
  /*006F*/ 0x89,0xD8,                  /* mov       eax,ebx           */
  /*0071*/ 0x03,0x43,0x10,             /* add       eax,[ebx+0x10]    */
  /*0074*/ 0x83,0xC0,0x04,             /* add       eax,0x00000004    */
  /*0077*/ 0x83,0xE0,0xFC,             /* and       eax,0xfffffffc    */
  /*007A*/ 0x87,0x02,                  /* xchg      eax,[edx]         */
  /*007C*/ 0x89,0x43,0x18,             /* mov       [ebx+0x18],eax    */
  /*007F*/                             /* _STACKLOW_set_end:          */

  /* as I said before:... new_esp := wrapper_offset + wrapper_size    */

  /*007F*/ 0x89,0xDD,                  /* mov       ebp,ebx           */
  /*0081*/ 0x03,0x6B,0x0C,             /* add       ebp,[ebx+0xc]     */
  /*0084*/ 0x83,0xED,0x04,             /* sub       ebp,0x00000004    */
  /*0087*/ 0x83,0xE5,0xFC,             /* and       ebp,0xfffffffc    */
  /*008A*/ 0x66,0x8C,0xD2,             /* mov       dx,ss             */
  /*008D*/ 0x66,0x8C,0xD8,             /* mov       ax,ds             */
  /*0090*/ 0x8E,0xD0,                  /* mov       ss,ax             */
  /*0092*/ 0x87,0xEC,                  /* xchg      ebp,esp           */

  /* save state */

  /*0094*/ 0x52,                       /* push      edx               */
  /*0095*/ 0x55,                       /* push      ebp               */
  /*0096*/ 0x0F,0xA0,                  /* push      fs                */
  /*0098*/ 0x0F,0xA8,                  /* push      gs                */
  /*009A*/ 0x1E,                       /* push      ds                */
  /*009B*/ 0x53,                       /* push      ebx               */

  /* setup args, initialize segregs, call, restore args (needed later) */

  /*009C*/ 0x51,                       /* push      ecx               */
  /*009D*/ 0x56,                       /* push      esi               */
  /*009E*/ 0x06,                       /* push      es                */
  /*009F*/ 0x57,                       /* push      edi               */
  /*00A0*/ 0x66,0x8C,0xD8,             /* mov       ax,ds             */
  /*00A3*/ 0x8E,0xE0,                  /* mov       fs,ax             */
  /*00A5*/ 0x8E,0xE8,                  /* mov       gs,ax             */
  /*00A7*/ 0x8E,0xC0,                  /* mov       es,ax             */
  /*00A9*/ 0xFC,                       /* cld                         */
  /*00AA*/ 0xFF,0x53,0x1C,             /* call      dword ptr [ebx+0x1c]*/
  /*00AD*/ 0xFA,                       /* cli                         */
  /*00AE*/ 0x5F,                       /* pop       edi               */
  /*00AF*/ 0x07,                       /* pop       es                */
  /*00B0*/ 0x5E,                       /* pop       esi               */
  /*00B1*/ 0x59,                       /* pop       ecx               */

  /* restore state */

  /*00B2*/ 0x5B,                       /* pop       ebx               */
  /*00B3*/ 0x1F,                       /* pop       ds                */
  /*00B4*/ 0x0F,0xA9,                  /* pop       gs                */
  /*00B6*/ 0x0F,0xA1,                  /* pop       fs                */
  /*00B8*/ 0x5D,                       /* pop       ebp               */
  /*00B9*/ 0x5A,                       /* pop       edx               */

  /* restore stack */

  /*00BA*/ 0x8E,0xD2,                  /* mov       ss,dx             */
  /*00BC*/ 0x89,0xEC,                  /* mov       esp,ebp           */

  /* restore _STACKLOW */

  /*00BE*/ 0x8B,0x53,0x14,             /* mov       edx,[ebx+0x14]    */
  /*00C1*/ 0x09,0xD2,                  /* or        edx,edx           */
  /*00C3*/ 0x74,0x05,                  /* je        X$38              */
  /*00C5*/ 0x8B,0x43,0x18,             /* mov       eax,[ebx+0x18]    */
  /*00C8*/ 0x89,0x02,                  /* mov       [edx],eax         */
  /*00CA*/                             /* X$38:                       */

                                       /* wrapper_fini: */
  /*00CA*/ 0xFF,0x4B,0x08,             /* dec       dword ptr [ebx+0x8]*/

  /* get wrapper type in eax, old_isr_vect in ebx */
  
  /*00CD*/ 0x8B,0x43,0x20,             /* mov       eax,[ebx+0x20]    */
  /*00D0*/ 0x8B,0x5B,0x28,             /* mov       ebx,[ebx+0x28]    */
  /*00D3*/ 0x1F,                       /* pop       ds                */
  /*00D4*/ 0x5A,                       /* pop       edx               */
  /*00D5*/ 0x59,                       /* pop       ecx               */
  /*00D6*/ 0x5D,                       /* pop       ebp               */

  /* fixup return frame depending on RMCB type */

  /*00D7*/ 0x83,0xF8,0x01,             /* cmp       eax,0x00000001    */
  /*00DA*/ 0x72,0x29,                  /* jb        go_iret           */
                                       /* RMCB_CHAIN: */
  /*00DC*/ 0x26,0x89,0x5F,0x2A,        /* mov       es:[edi+0x2a],ebx */
  /*00E0*/ 0x74,0x23,                  /* je        go_iret           */

                                       /* RMCB_IRET or RMCB_RETF: */
  /*00E2*/ 0x8B,0x1E,                  /* mov       ebx,[esi]         */
  /*00E4*/ 0x26,0x89,0x5F,0x2A,        /* mov       es:[edi+0x2a],ebx */
  /*00E8*/ 0xBB,0x04,0x00,0x00,0x00,   /* mov       ebx,0x00000004    */
  /*00ED*/ 0x83,0xF8,0x02,             /* cmp       eax,0x00000002    */
  /*00F0*/ 0x75,0x0E,                  /* jne       pop_on_rmstack    */
  /*00F2*/ 0x66,0x8B,0x5E,0x04,        /* mov       bx,[esi+0x4]      */
  /*00F6*/ 0x66,0x26,0x89,0x5F,0x20,   /* mov       es:[edi+0x20],bx  */
  /*00FB*/ 0xBB,0x06,0x00,0x00,0x00,   /* mov       ebx,0x00000006    */
  /*0100*/                             /* pop_on_rmstack:             */
  /*0100*/ 0x66,0x26,0x01,0x5F,0x2E,   /* add       es:[edi+0x2e],bx  */

  /*0105*/                             /* go_iret:                    */
  /*0105*/ 0x5B,                       /* pop       ebx               */
  /*0106*/ 0x58,                       /* pop       eax               */
  /*0107*/ 0xCF                        /* iretd                       */
  };
  unsigned int wrap_len = 16300; /*stack size is this minus sizeof(wrapper)*/
  unsigned char *wrap = (unsigned char *)0;

  wrap = (unsigned char *)malloc(wrap_len);

  if (wrap)
  {
    unsigned short wrap_sel;
    #if defined(__DJGPP__) /* signal safe ds */
    wrap_sel = (unsigned short)__djgpp_ds_alias;
    #else
    wrap_sel = (unsigned short)__my_dpmi_get_ds();
    #endif
    if (__assert_cs_base_and_limit(wrap_sel) == 0)
    {
      if (__my_dpmi_lock_data(wrap,wrap_len) == 0)
      {
        memset( wrap, 0, wrap_len ); 
        memcpy( wrap, wrapper, sizeof(wrapper) );

        *((unsigned long *)(&wrap[WRAPOFF_VECTOR]))      = vector;
        *((unsigned long *)(&wrap[WRAPOFF_SIGNATURE]))   = WRAPPER_SIG_RMCB;
        *((unsigned long *)(&wrap[WRAPOFF_REENTRY_CHK])) = 0;        
        *((unsigned long *)(&wrap[WRAPOFF_WRAPSIZE]))    = wrap_len;
        *((unsigned long *)(&wrap[WRAPOFF_CODESIZE]))    = sizeof(wrapper);
        #if defined(__WATCOMC__)
        {
          #ifndef __cplusplus
          extern unsigned _STACKLOW;
          #endif
          *((unsigned long *)(&wrap[WRAPOFF_STACKLOWPTR])) = (long)(&_STACKLOW);
        }
        #endif
        *((void **        )(&wrap[WRAPOFF_RMCB_PTR]))   = (void *)pm_entry;
        *((unsigned long *)(&wrap[WRAPOFF_WRAP_TYPE]))  = wrap_type;    
        *((unsigned short*)(&wrap[WRAPOFF_RMCB_SELF_SEL]))   = wrap_sel;      
        *((void **        )(&wrap[WRAPOFF_RMCB_SELF_OFF32])) = (void *)wrap;  
    
        if (wrap_type == WRAPTYPE_RMCB_IRET || 
            wrap_type == WRAPTYPE_RMCB_RETF ||
            wrap_type == WRAPTYPE_RMCB_CHAIN)
        {
          if (__my_dpmi_lock_data( rmcb_regs, sizeof(*rmcb_regs) ) == 0)
          {
            *((void **)(&wrap[WRAPOFF_REGSPTR])) = (void *)rmcb_regs;
            if (wrap_type == WRAPTYPE_RMCB_CHAIN)
            {
              __dpmi_raddr tmp;
              __dpmi_get_real_mode_interrupt_vector(vector, &tmp);
              *((unsigned short *)(&wrap[WRAPOFF_OLDISR+0])) = tmp.offset16;
              *((unsigned short *)(&wrap[WRAPOFF_OLDISR+2])) = tmp.segment;
            }
            return wrap;
          }
        }
        else /* WRAPTYPE_RMCB_RAW */
        {
          return wrap;
        }
      } /* if (__my_dpmi_lock_data(wrap,wrap_len) == 0) */
    } /* if (__assert_cs_base_and_limit(wrap_sel) == 0) */
    free(wrap);
  } 
  return 0;
}

/* --------------------------------------------------------------------- */

static void __free_wrapper(void *_wrap)
{
  unsigned char *wrap = (unsigned char *)_wrap;
  if (wrap)
  {
    if (*((unsigned long *)(&wrap[WRAPOFF_SIGNATURE])) == WRAPPER_SIG_RMCB
    &&  *((void **        )(&wrap[WRAPOFF_RMCB_SELF_OFF32])) == _wrap)
    {
      unsigned int wrap_len = *((unsigned int *)(&wrap[WRAPOFF_WRAPSIZE]));
      int wrap_type = *((unsigned int *)(&wrap[WRAPOFF_WRAP_TYPE]));

      if (wrap_type == WRAPTYPE_RMCB_IRET || 
          wrap_type == WRAPTYPE_RMCB_RETF ||
          wrap_type == WRAPTYPE_RMCB_CHAIN)
      {
        if (wrap_type == WRAPTYPE_RMCB_CHAIN)
        {
          unsigned char vector = wrap[WRAPOFF_VECTOR]; 
          __dpmi_raddr rm_int;
          rm_int.offset16 = *((unsigned short *)(&wrap[WRAPOFF_OLDISR+0]));
          rm_int.segment  = *((unsigned short *)(&wrap[WRAPOFF_OLDISR+2]));
          __dpmi_set_real_mode_interrupt_vector(vector, &rm_int);
        }
        __my_dpmi_unlock_data( *((void **)(&wrap[WRAPOFF_REGSPTR])),
                                sizeof(__dpmi_regs) );
      }
      __my_dpmi_unlock_data(wrap,wrap_len);
      free(wrap);
    }
  }
  return;
}

/* --------------------------------------------------------------------- */

#pragma pack(1)
struct __my_rmcbw
{
  unsigned long signature;
  void *wrapper;
  unsigned short cb_sel;
  unsigned char jmp_isn;
  unsigned short real_rmcb_off;
  unsigned short real_rmcb_seg;
}; /* size==15 */
#pragma pack()

/* --------------------------------------------------------------------- */

static int __mk_rmcb( int wrap_type, int vector, void (*func)(void), 
                                     __dpmi_regs *regs,
                                     __dpmi_raddr *rmaddr )
{
  void *wrap = __mk_wrapper( wrap_type, vector, func, regs );
  if (wrap)
  {
    struct __my_rmcbw cb;
    int cb_sel, cb_seg;
    cb_seg = __dpmi_allocate_dos_memory ( (sizeof(cb)+15)>>4, &cb_sel );
    if (cb_seg != -1)
    {
      typedef void (*xxx_rmcb_t)(void);
      if (__dpmi_allocate_real_mode_callback ( (xxx_rmcb_t)wrap,
                                                 regs, rmaddr ) == 0)
      {
        cb.signature = WRAPPER_SIG_RMCB;
        cb.wrapper = wrap;
        cb.cb_sel = (unsigned short)cb_sel;
        cb.jmp_isn = 0xEA;    /* jmp absolute */
        cb.real_rmcb_off = rmaddr->offset16;
        cb.real_rmcb_seg = rmaddr->segment;
        dosmemput( &cb, sizeof(cb), (((unsigned long)cb_seg)<<4) );

        rmaddr->segment = (unsigned short)cb_seg;
        rmaddr->offset16 = (short)(((char *)&(cb.jmp_isn))-((char *)&cb));
        return (0); /* success */
      }
      __dpmi_free_dos_memory(cb_sel);
    } /* if (cb_seg != -1) */
    __free_wrapper(wrap);    
  } /* if (wrap) */
  return -1; /* 0x8015; */
}
#endif /* __DPMI32__ && !DONT_WANT_RMCB_SUPPORT */

/* --------------------------------------------------------------------- */

#if defined(__DPMI32__)
int __my_dpmi_free_real_mode_callback( __dpmi_raddr *rmaddr )
{
#if defined(DONT_WANT_RMCB_SUPPORT)
  rmaddr = rmaddr; 
  return -1;
#else
  struct __my_rmcbw cb;
  if (rmaddr->offset16 == (short)(((char *)&(cb.jmp_isn))-((char *)&cb)) )
  {
    unsigned long lineaddr = (((unsigned long)(rmaddr->segment))<<4);
    dosmemget(lineaddr, sizeof(cb), &cb); 
    if (cb.signature == WRAPPER_SIG_RMCB)
    {
      unsigned int cb_sel = cb.cb_sel;
      __dpmi_raddr tmp;
      tmp.offset16 = cb.real_rmcb_off;
      tmp.segment = cb.real_rmcb_seg;
      __dpmi_free_real_mode_callback ( &tmp );
      __free_wrapper(cb.wrapper);
      memset( &cb, 0, sizeof(cb) );
      dosmemput( &cb, sizeof(cb), lineaddr );
      __dpmi_free_dos_memory(cb_sel);
      return 0;
    }
  }
  return __dpmi_free_real_mode_callback( rmaddr );
#endif
}
#endif /* __DPMI32__ */

/* --------------------------------------------------------------------- */

#if defined(__DPMI32__)
int __my_dpmi_allocate_real_mode_callback_retf(__dpmi_rmcb_handler_t func,
                                                __dpmi_regs *regs,
                                                __dpmi_raddr *rmaddr )
{
  #if defined(DONT_WANT_RMCB_SUPPORT)
  func = func;
  regs = regs;
  rmaddr = rmaddr;
  return -1;
  #else
  return __mk_rmcb( WRAPTYPE_RMCB_RETF, 0, func, regs, rmaddr );
  #endif
}
#endif /* __DPMI32__ */

/* --------------------------------------------------------------------- */

#if defined(__DPMI32__)
int __my_dpmi_allocate_real_mode_callback_iret(__dpmi_rmcb_handler_t func,
                                                __dpmi_regs *regs,
                                                __dpmi_raddr *rmaddr )
{
  #if defined(DONT_WANT_RMCB_SUPPORT)
  func = func;
  regs = regs;
  rmaddr = rmaddr;
  return -1;
  #else
  return __mk_rmcb( WRAPTYPE_RMCB_IRET, 0, func, regs, rmaddr );
  #endif
}
#endif /* __DPMI32__ */

/* --------------------------------------------------------------------- */

#if defined(__DPMI32__)
int __my_dpmi_allocate_real_mode_callback_ichain(__dpmi_rmcb_handler_t func,
                                                  int vector,
                                                  __dpmi_regs *regs,
                                                  __dpmi_raddr *rmaddr )
{
  #if defined(DONT_WANT_RMCB_SUPPORT)
  func = func;
  regs = regs;
  rmaddr = rmaddr;
  vector = vector;
  return -1;
  #else
  return __mk_rmcb( WRAPTYPE_RMCB_CHAIN, vector, func, regs, rmaddr );
  #endif
}
#endif /* __DPMI32__ */

/* =================================================================== */
/*                          WATCOM 386                                 */
/* =================================================================== */

#if defined(__WATCOMC__) && defined(__386__)

  #if defined(__SW_3R)
    #pragma library ("clib3r.lib");
  #elif defined(__SW_3S)
    #pragma library ("clib3s.lib");
  #else
    #error huh? wcc386 but not __SW_3R or __SW_3S?
  #endif

  #define __compiler_specific_init() (0) /* nothing to do */

#endif /* __WATCOMC__ && __386__ */

/* =================================================================== */
/*                          BORLAND386                                 */
/* =================================================================== */

#if defined(__BORLANDC__) && defined(__FLAT__)

  void __bcc32_segread (struct SREGS *sreg)
  {
    sreg->ds = _DS;
    sreg->es = _ES;
    sreg->ss = _SS;
    sreg->cs = _CS;
    __asm db 66h,8ch,0e0h /* mov ax,fs */
    sreg->fs = _AX;
    __asm db 66h,8ch,0e8h /* mov ax,gs */
    sreg->gs = _AX;
  }
  int __bcc32_int386 (int intno, union REGS *ireg, union REGS *oreg)
  {
    struct SREGS sreg;
    segread (&sreg);
    return int386x (intno, ireg, oreg, &sreg);
  }
  int __bcc32_int386x (int intno, union REGS *ireg, union REGS *oreg, struct SREGS *sreg)
  {
    /*
    ** int386x assumes that ss is executable.
    ** If that assumption is not valid then this code will need to be
    ** redone in real asm using a jump table for the 'Int xx; ret' call. 
    */
    __dpmi_regs regpack;

    regpack.r_ax = ireg->x.eax;
    regpack.r_bx = ireg->x.ebx;
    regpack.r_cx = ireg->x.ecx;
    regpack.r_dx = ireg->x.edx;
    regpack.r_si = ireg->x.esi;
    regpack.r_di = ireg->x.edi;
    regpack.r_ds = sreg->ds;
    regpack.r_es = sreg->es;

    __asm lea  eax, regpack
    __asm mov  edx, intno
    __asm push gs
    __asm push fs
    __asm push ds
    __asm push es
    __asm push ebp
    __asm push ebx
    __asm push edi
    __asm push esi

    /* create instruction buf: 0xCD,intno,0xC3 */
    __asm and  edx,0ffh     /* intno */ 
    __asm mov  dh,0C3h      /* 0xC3 == ret */
    __asm shl  edx,8
    __asm mov  dl,0CDh      /* 0xCD == int */ 

    __asm push eax     /* save regpack offset */
    __asm push edx     /* create instruction buffer */

    __asm mov  ebp,eax /* regpack offset */
    __asm mov  edi,[ebp+0]
    __asm mov  esi,[ebp+4]
    __asm mov  ebx,[ebp+16]
    __asm mov  edx,[ebp+20]
    __asm mov  ecx,[ebp+24]
    __asm mov  eax,[ebp+28]
    __asm mov  es, [ebp+34]
    __asm mov  ds, [ebp+36]

    __asm mov  ebp,esp     /* load pointer to isn buffer */
    __asm call ebp
    __asm pop  ebp         /* destroy instruction buffer */
    __asm pop  ebp         /* restore regpack offset */

    __asm mov  [ebp+0],edi
    __asm mov  [ebp+4],esi
    __asm mov  [ebp+16],ebx
    __asm mov  [ebp+20],edx
    __asm mov  [ebp+24],ecx
    __asm mov  [ebp+28],eax
    __asm pushfd
    __asm pop  eax
    __asm mov  [ebp+32],ax
    __asm mov  [ebp+34],es
    __asm mov  [ebp+36],ds

    __asm pop  esi
    __asm pop  edi
    __asm pop  ebx
    __asm pop  ebp
    __asm pop  es
    __asm pop  ds
    __asm pop  fs
    __asm pop  gs

    oreg->x.eax   = regpack.r_ax;
    oreg->x.ebx   = regpack.r_bx;
    oreg->x.ecx   = regpack.r_cx;
    oreg->x.edx   = regpack.r_dx;
    oreg->x.esi   = regpack.r_si;
    oreg->x.edi   = regpack.r_di;
    oreg->x.flags = regpack.r_flags;
    sreg->ds      = regpack.r_ds;
    sreg->es      = regpack.r_es;
    return (int)regpack.r_ax;
  }

#if 0
  void __bcc32_movedata( unsigned src_sel, unsigned src_off,
                         unsigned dst_sel, unsigned dst_off,
                         unsigned length )
  {
    __asm push ds
    __asm push es
    __asm push esi
    __asm push edi
    __asm mov  ecx,dst_sel
    __asm mov  edi,dst_off
    __asm mov  es,cx
    __asm mov  esi,src_off
    __asm mov  eax,src_sel
    __asm mov  ecx,length
    __asm mov  ds,ax
    __asm cld
    __asm mov  eax,ecx
    __asm shr  ecx,2
    __asm and  ax,3
    __asm rep  movsd
    __asm mov  cx,ax
    __asm rep  movsb
    __asm pop  edi
    __asm pop  esi
    __asm pop  es
    __asm pop  ds
    return;
  }
#endif

  #define __compiler_specific_init() (0) /* nothing to do */

#endif /* defined(__BORLANDC__) && defined(__FLAT__) */

/* =================================================================== */
/*                              GCC                                    */
/* =================================================================== */

#if defined(__DJGPP__) /* requires CWSDPMI (or compatible) implementation */
/* 
** no dpmi init necessary - if linked against a DMPI implementation that
** does not support a particular DJGPP DPMI API call, that call will fail,
** and that is something that we have to check for anyway.
*/
#define __compiler_specific_init() (0) /* nothing to do */
#if defined(__cplusplus) 
/* emit this into every object file that is compiled with c++ */
/* we don't actually use c++ elements, and don't want lib???cx */ 
__asm__(".weak ___gxx_personality_v0\n ___gxx_personality_v0 = _errno\n");
#endif
#endif /* __DJGPP__ */

/* =================================================================== */
/*                            HIGH C                                   */
/* =================================================================== */

#if defined(__HIGHC__)
#define __compiler_specific_init() (0) /* nothing to do */
#endif

/* =================================================================== */
/*                            EVERYONE!                                */
/* =================================================================== */

#if defined( DPMI_INIT_TRACE)
#undef DPMI_INIT_TRACE
#define DPMI_INIT_TRACE(x) _printk x
#else
#define DPMI_INIT_TRACE(x) /* nothing */
#endif

static int __internal_extender_init(void)
{
#if !defined(__DPMI32__) 
  #if (UINT_MAX > USHRT_MAX)
  #error Something is badly broken. __DPMI32__ was not defined.
  return -1;
  #else
  return 0;
  #endif
#else
  union REGS r;
  struct SREGS s;
  unsigned short pharsig;
  unsigned long csbase, dsbase, limit;

  DPMI_INIT_TRACE(("PHAR check\n"));
  memset( &r, 0, sizeof(r));
  #if defined(__DJGPP__) /* damn naivity (_BORLAND_DOS_REGS) */
  r.d.ebx = 0x50484152; /* "PHAR" */
  r.d.eax = 0x3000; /* ah=30h */
  int386(0x21, &r, &r);
  pharsig = (unsigned short)(r.d.eax >> 16);
  #else
  r.x.ebx = 0x50484152; /* "PHAR" */
  r.x.eax = 0x3000; /* ah=30h */
  int386(0x21, &r, &r);
  pharsig = (unsigned short)(r.x.eax >> 16);
  #endif

  DPMI_INIT_TRACE(("DPMI inst check\n"));
  /* The DPMI installation check doesn't work for DOS4GW or other */
  /* extenders when the DPMI host is per-process only */
  memset( &r, 0, sizeof(r));
  r.w.ax = 0x1687;
  int386 (0x2f, &r, &r );
  if (r.w.ax != 0) 
  {  
    DPMI_INIT_TRACE(("DPMI ver check\n"));
    memset( &r, 0, sizeof(r));
    r.w.ax = 0x0400; /* Get DPMI Version */
    r.w.cflag = 1;
    int386 (0x31, &r, &r );
    if (!(r.w.cflag & 1) && (r.w.ax==90 /*0.90*/ || r.w.ax==0x100 /*1.00*/))
      r.w.ax = 0; 
  }
  if (r.w.ax != 0 || (r.w.bx & 1)==0 || (r.w.cx & 0xff) < 3)
  {
    if (pharsig == 0x4458) /* 'DX' */
      _printk("Init error: This application requires DPMI (do not link with -NODPMI)\n");
    else if (pharsig == 0x4243) /* 'BC' */ /* This should never happen */
      _printk("Init error: Detected Intel Code Builder but no DPMI.\n");
    else
    {
      unsigned char c1 = (unsigned char)(pharsig & 0xff);
      unsigned char c2 = (unsigned char)(pharsig >> 8);
      _printk("Init error: Unrecognized DOS extender 0x%04x ('%c%c')\n",
         pharsig, ((c1 < ' ' || c1 > 0x7e) ? '?' : c1),
                  ((c2 < ' ' || c2 > 0x7e) ? '?' : c2) );
    }
    return -1;
  }

  /* ---------------------------------------------- */

  DPMI_INIT_TRACE(("selector cmp check\n"));
  #if defined(__DJGPP__)
  s.ds = _my_ds();
  s.ss = _my_ss();
  s.cs = _my_cs();
  #else
  segread( &s );
  #endif
  if (s.ds != s.ss)
  {
    _printk("Init error: SS and DS are not equal. (DS=0x%04x, SS=0x%04x)\n",
                                               s.ss,s.ds);
    return -1;
  }
  /* this check is primarily for HIGHC which defines both 
   * _my_ds() and _my_cs() as _mwgetcs().
   * I simply cannot imagine that ANY extender would actually have cs 
   * identical to ds (leave alone a "professional" one like pharlap). 
   * Two identical selectors is not the same as two selectors with the same
   * base and limit.
  */
  if (s.ds != _my_ds())
  {
    _printk("Init error: segread().ds != _my_ds(). (.ds=0x%04x, _my_ds()=0x%04x)\n",
                                               s.ds, _my_ds() );
    return -1;
  }
  if (s.cs != _my_cs())
  {
    _printk("Init error: segread().cs != _my_cs(). (.cs=0x%04x, _my_cs()=0x%04x)\n",
                                               s.cs, _my_cs() );
    return -1;
  }

  /* ---------------------------------------------- */

  DPMI_INIT_TRACE(("get seg base\n"));
  /* CS and DS base must be equal or rm callbacks/pm interrupts will crash
  ** since the C functions being called into have 'near' returns.
  */
  DPMI_INIT_TRACE(("cssel=%x, dssel=%x\n", _my_cs(), _my_ds() ));
  if (__dpmi_get_segment_base_address ( _my_cs(), &csbase ) || 
      __dpmi_get_segment_base_address ( _my_ds(), &dsbase ))
  {
    _printk("Init error: Unable to get CS or DS base\n");
    return -1;
  }
  if (csbase != dsbase)   
  {
    _printk("Init error: CS base != DS base (CSb=0x%08x, DSb=0x%08x)\n",
                                            csbase, dsbase );
    return -1;
  }
  DPMI_INIT_TRACE(("csbase=%x, dsbase=%x\n", csbase, dsbase ));

  /* ---------------------------------------------- */

  /*
  ** Validate the DOS memory (mem below 1M) selector
  ** Note that the call to __my_dpmi_get_dosds() may
  ** also be a call to allocate the selector.
  */
  s.ds = __my_dpmi_get_dosds(); /* aka _dos_ds */
  if ((s.ds == ((unsigned short)-1)) || (s.ds == 0)) 
  {
    _printk("Init error: Unable to get DOSmem selector (0x%04x)\n", s.ds);
    return -1;
  }
  else if (__dpmi_get_segment_base_address ( s.ds, &dsbase ) != 0)
  {
    _printk("Init error: Unable to get DOSmem selector base\n");
    return -1;
  }
  else if (dsbase != 0)
  {
    _printk("Init error: DOSmem selector base is not zero (0x%lx)\n", dsbase);
    return -1;
  }
  else if ((limit = __dpmi_get_segment_limit ( s.ds )) < 0xfffff)
  {
    _printk("Init error: Invalid DOSmem selector limit (0x%lx)\n", limit);
    return -1;
  }
  DPMI_INIT_TRACE(("dos_ds = 0x%04x\n", s.ds ));

  /* ---------------------------------------------- */

  /* allocate a selector for writing to CS
  */
  if (__my_dpmi_get_csalias() == -1)
  {
    _printk("PANIC! Unable to allocate writable CS selector\n");
    return -1;
  }
  DPMI_INIT_TRACE(("cs_alias = 0x%04x\n", __my_dpmi_get_csalias() ));

  /* ---------------------------------------------- */

  /*
  ** for pharlap targets, increase minreal to max possible
  */
  if (pharsig == 0x4458 /* 'DX' */)
  {                        /* increase minreal to max possible */
    DPMI_INIT_TRACE(("PHAR mem fixup\n"));
    #if defined(__DJGPP__) /* damn naivity (_BORLAND_DOS_REGS) */
    r.d.eax = 0x2525;
    r.d.ebx = 0;
    r.d.ecx = 0;
    #else
    r.x.eax = 0x2525;
    r.x.ebx = 0;
    r.x.ecx = 0;
    #endif
    int386 (0x21, &r, &r);
    if (r.w.cflag & 1)
    {
      #if defined(__DJGPP__) /* damn naivity (_BORLAND_DOS_REGS) */
      r.d.eax = 0x2525;
      r.d.ebx = r.d.ecx;
      #else
      r.x.eax = 0x2525;
      r.x.ebx = r.x.ecx;
      #endif
      int386 (0x21, &r, &r);
    }
  }

  DPMI_INIT_TRACE(("end 386 extender init\n"));
  return __compiler_specific_init();
#endif /* __DPMI32__ */
}

static void __ctor_dpmi(void) 
{
  if (__internal_extender_init() != 0)
    exit(1); /* message already printed */
  return;
} 

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_dpmi
#include "ctordtor.h"
