/*
 * Service provider for the 
 * Internet Control Message Protocol, "ICMP", (RFC 792)
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: icmp.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __KERN_ICMP_H
#define __KERN_ICMP_H(x) __tkern_ ## x

#if defined(__WATCOMC__) && !defined(_ENABLE_AUTODEPEND)
#pragma read_only_file;
#endif

/* =================================================================== */
/*                        transport-only access                        */
/* =================================================================== */

#if (defined(__OSILAYER__) && (__OSILAYER__ == 3 || __OSILAYER__ == 4))

#define icmp_mask_req              __KERN_ICMP_H(icmp_mask_req)
#define icmp_router_solic          __KERN_ICMP_H(icmp_router_solic)

extern uint32_t  icmp_mask_req(int timeout); /*used by IPBOOT. Yields! */
extern int       icmp_router_solic(void);    /*used by ARP resolver. Yields!*/

/* ------------------------------------------------------------ */

enum /* ICMP conditions forwarded to (and sent by) the transport providers. */
{    
  /* Must be contiguous numbering, and start at 1. These codes are
   * subject to change and cannot be relied on to keep the same values
   * or to remain in any particular order. IE, don't use xlat tables or
   * make visible/usable from userland.
  */
  /*"Unreachable Net" doesn't need a separate code (same as unreachable host)*/
  SOICMPERR_UNREACH_HOST  = 1, /* also unreachable net/bad tos/eaccess */
  SOICMPERR_UNREACH_PROTO = 2,
  SOICMPERR_UNREACH_PORT  = 3,
#  define SOICMPERR_ISUNREACH(x)   \
         (x >= SOICMPERR_UNREACH_HOST && x <= SOICMPERR_UNREACH_PORT)
  SOICMPERR_CANTFRAG      = 4, /* Fragmentation needed and DF set */
  SOICMPERR_TIMXCEED_FRAG = 5,
  SOICMPERR_TIMXCEED_TTL  = 6,
  SOICMPERR_SOURCEQUENCH  = 7,
  SOICMPERR_PARAMPROB     = 8
};

#define icmp_retour __KERN_ICMP_H(icmp_retour)
extern void icmp_retour( int soicmp_err, const void FAR *orig_ipdgram,
              const void FAR *hwaddr/* optional */, int callers_debug_lev );

/* ------------------------------------------------------------ */

#define icmp_echo_reply_hook_add  __KERN_ICMP_H(icmp_echo_reply_hook_add)
#define icmp_echo_reply_hook_del  __KERN_ICMP_H(icmp_echo_reply_hook_del)

typedef void (*icmp_echo_reply_hook_t)(const void FAR *_ippkt);
extern int    icmp_echo_reply_hook_add(icmp_echo_reply_hook_t);
extern int    icmp_echo_reply_hook_del(icmp_echo_reply_hook_t);

/* ------------------------------------------------------------ */

#endif /* __OSILAYER__ 3/4 */
#endif /* __KERN_ICMP_H */
