/*
 * Multicast Group operation stubs for use by transports only.
 *
 * These are only stubs - they do not actually do anything unless
 * the user's application references one of the sock_mcast_*() funcs.
 * 
 * When an application uses sock_mcast_*(), ul_mcast.c gets linked.
 * ul_mcast.c has a reference to __IPPROTO_IGMP__ which is in igmp.c
 * When igmp.c gets linked, a constructor there calls _ip_mcast_setgopfn()
 * with a pointer to a routine to do group operations.
 *
 * Ergo, the overhead (and to a lesser extent, code bloat) of igmp and 
 * multicast support is non-existant unless the user uses sock_mcast_*().
 *
 * Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 * 
*/
#define __COMPILING_IP_MCAST__

#include "target.h"  /* standard types/includes */
#include "ipmcast.h" /* ourselves */

static struct
{
  /* group_op_fn is set (via _ip_mcast_setgopfn()) by a constructor in igmp.c */
  int (*group_op_fn)(uint32_t ina, int op);
} my_statics = {0};


void _ip_mcast_setgopfn(int reserved, int (*fn)(uint32_t ina, int op))
{
  reserved = reserved;
  my_statics.group_op_fn = fn;
}


int _ip_mcast_group_op(uint32_t ina, int op)
{
  if (my_statics.group_op_fn)
    return (*(my_statics.group_op_fn))( ina, op );
  return 0;  
}
