/*
 * Multicast Group operation stubs for use by transports only.
 *
 * These are only stubs - they do not actually do anything unless
 * the user's application references one of the sock_mcast_*() funcs.
 * 
 * When an application uses sock_mcast_*(), ul_mcast.c gets linked.
 * ul_mcast.c has a reference to __IPPROTO_IGMP__ which is in igmp.c
 * When igmp.c gets linked, a constructor there calls _ip_mcast_setgopfn()
 * with a pointer to a routine to do group operations.
 *
 * Ergo, the overhead (and to a lesser extent, code bloat) of igmp and 
 * multicast support is non-existant unless the user uses sock_mcast_*().
 *
 * Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 * 
*/
#ifndef __KERN_IP_MCAST_H
#define __KERN_IP_MCAST_H(x) __tkern_ ## x

#if defined(__COMPILING_IGMP__) || defined(__COMPILING_IP_MCAST__)
/* _ip_mcast_setgopfn() is only for use by IGMP */
#define _ip_mcast_setgopfn     __KERN_IP_MCAST_H(ip_mcast_setgopfn)
extern void _ip_mcast_setgopfn(int reserved, int (*fn)(uint32_t ina, int op));
#endif

#if !defined(__COMPILING_IGMP__)
#define _ip_mcast_group_op     __KERN_IP_MCAST_H(ip_mcast_group_op)
extern int _ip_mcast_group_op(uint32_t ina, int op);
#define _ip_mcast_join_group(__ina)   _ip_mcast_group_op((__ina),+1)
#define _ip_mcast_leave_group(__ina)  _ip_mcast_group_op((__ina),-1)
#define _ip_mcast_is_in_group(__ina)  _ip_mcast_group_op((__ina),0)
#endif

#endif /* __KERN_IP_MCAST_H */
