/*
 * Link-layer (OSI layer 2) interface service stub provider.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: linklayr.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
/*
 * Link-layer Interface (OSI layer 2).
 *
 * No layer above this one needs to know what an on-media packet 
 * looks like, or how hardware addresses are formatted etc.
*/

#ifndef __KERN_LINKLAYR_H
#define __KERN_LINKLAYR_H(x) __tkern_ ## x

#if (defined(__OSILAYER__) && (__OSILAYER__ == 2 || __OSILAYER__ == 3))

/* 'hwtype' field originally defined for arp but used by lots of other 
** protocols (including arp, bootp, dhcp) as an 'address type' specifier. 
** See _eth_get_hwtype()/http://www.iana.org/assignments/arp-parameters
*/
#define RFC826_HWTYPE_BLUEBOOK    ((uint8_t)(0x01)) /* DIX */
#define RFC826_HWTYPE_AX25        ((uint8_t)(0x03)) /* AX.25 */
#define RFC826_HWTYPE_IEEE802     ((uint8_t)(0x06)) /* IEEE802.{3/4/5} */
#define RFC826_HWTYPE_ARCNET      ((uint8_t)(0x07)) /* ArcNet */
#define RFC828_HWTYPE_LOCALTALK   ((uint8_t)(0x0B)) /* IP/DDP */
#define RFC826_HWTYPE_SERIALLINE  ((uint8_t)(0x14)) /* Serial Link */


#define MAX_LINKADDR_SIZE 6 /* Max hardware address len of any HW type */
/* Actually, the maximum is 16, but we don't use more than 6 anywhere */

/* ---------------------------------------------------------------------- */

/*
** types passed to _eth_send()
** Same as ETHERTYPE_* in ethertype.h
*/
#define ETHERTYPE_PUP              0x0200  /* PUP protocol */
#define ETHERTYPE_IP               0x0800  /* IP protocol */
#define ETHERTYPE_ARP              0x0806  /* Addr. resolution protocol */
#define ETHERTYPE_REVARP           0x8035  /* reverse Addr. resolution protocol */
#define ETHERTYPE_VLAN             0x8100  /* IEEE 802.1Q VLAN tagging */
#define ETHERTYPE_IPV6             0x86dd  /* IPv6 */
#define ETHERTYPE_LOOPBACK         0x9000  /* used to test interfaces */

/* ---------------------------------------------------------------------- */

/* device flags 
*/
#define ETH_IFFLAGS_UP      0x0001
#define ETH_IFFLAGS_LBONLY  0x0002 /* only loopback (PD is nullpkt) */
#define ETH_IFFLAGS_P2P     0x0010 /* implies not MCAST/BCAST */
#define ETH_IFFLAGS_MCAST   0x0100

/* ---------------------------------------------------------------------- */

/*
** mcast modes
*/
#define ETH_MCAST_MODE_OFF        0x00
#define ETH_MCAST_MODE_LIMITED    0x01
#define ETH_MCAST_MODE_UNLIMITED  0x02

/* ---------------------------------------------------------------------- */

#define _eth_get_ifflags          __KERN_LINKLAYR_H(eth_get_ifflags)
#define _eth_get_hwtype           __KERN_LINKLAYR_H(eth_get_hwtype)
#define _eth_get_mtu              __KERN_LINKLAYR_H(eth_get_mtu)
#define _eth_get_hwaddr           __KERN_LINKLAYR_H(eth_get_hwaddr)
#define _eth_get_hwaddrlen        __KERN_LINKLAYR_H(eth_get_hwaddrlen)
#define _eth_get_hwaddrbc         __KERN_LINKLAYR_H(eth_get_hwaddrbc)
#define _eth_send                 __KERN_LINKLAYR_H(eth_send)
#define _eth_numbufs_set          __KERN_LINKLAYR_H(eth_numbufs_set)
#define _eth_mcast_set_mode       __KERN_LINKLAYR_H(eth_mcast_set_mode)
#define _eth_mcast_set_list       __KERN_LINKLAYR_H(eth_mcast_set_list)
#define _eth_mcast_get_max_groups __KERN_LINKLAYR_H(eth_mcast_get_max_groups)
#define _eth_register_pkthandler  __KERN_LINKLAYR_H(eth_register_pkthandler)

extern int             _eth_get_ifflags          (void);
extern uint8_t            _eth_get_hwtype           (void);
extern uint16_t            _eth_get_mtu              (void);
extern const uint8_t FAR *_eth_get_hwaddr           (void);
extern uint8_t            _eth_get_hwaddrlen        (void);
extern const uint8_t FAR *_eth_get_hwaddrbc         (void);
extern int             _eth_send( uint16_t ethtype, const void FAR *dest_hw_addr,
                                  const void FAR *pkt, unsigned int pkt_len);
extern int             _eth_numbufs_set          (unsigned int count);
extern int             _eth_mcast_set_mode       (int mode);
extern int             _eth_mcast_set_list(const void *hwaddrlist, 
                              unsigned int listlen);
extern int             _eth_mcast_get_max_groups (void);
extern int             _eth_register_pkthandler( uint16_t ethtype,
                              void (*handler)( const void FAR *netpkt,
                              unsigned int netpkt_len,
                              const void *src_hw_addr,
                              int is_brdcast ) );

/* ---------------------------------------------------------------------- */
/*
** MASSIVE HACK ALERT: _eth_poll SHOULD be static, but isn't because
** it needs to called directly to support a massive hack in ipboot.c.
** See comments there for more information.
*/
#define _eth_poll              __KERN_LINKLAYR_H(eth_poll)
extern void _eth_poll(void);

/* ---------------------------------------------------------------------- */

#endif /* __OSILAYER__ 2/3 */
#endif /* __KERN_LINKLAYR_H */

