/*
 * Portable (model and real/protected mode independant) 
 * Packet Driver Interface.
 * 
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * Portions Copyright (c) 1986, 1989 FTP Software Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: pktdrvr.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __PDAPI_H__
#define __PDAPI_H__

#if defined(__WATCOMC__) && !defined(_ENABLE_AUTODEPEND)
#pragma read_only_file;
#endif

enum pdapi_driver_errnos
{       
  PDERR_BAD_HANDLE     =  1, /*Invalid handle number,*/
  PDERR_NO_CLASS       =  2, /*No interfaces of specified class found,*/
  PDERR_NO_TYPE        =  3, /*No interfaces of specified type found,*/
  PDERR_NO_NUMBER      =  4, /*No interfaces of specified number found,*/
  PDERR_BAD_TYPE       =  5, /*Bad packet type specified,*/
  PDERR_NO_MULTICAST   =  6, /*This interface does not support multicast,*/
  PDERR_CANT_TERMINATE =  7, /*This packet driver cannot terminate,*/
  PDERR_BAD_MODE       =  8, /*An invalid receiver mode was specified,*/
  PDERR_NO_SPACE       =  9, /*Operation failed because of insufficient space,*/
  PDERR_TYPE_INUSE     = 10, /*Type had previously been accessed, and not released*/
  PDERR_BAD_COMMAND    = 11, /*The command was out of range, or not implemented,*/
  PDERR_CANT_SEND      = 12, /*The packet couldn't be sent (usually hw error),*/
  PDERR_CANT_SET       = 13, /*HW address couldn't be changed (> 1 handle open)*/
  PDERR_BAD_ADDRESS    = 14, /*Hardware address has bad length or format,*/
  PDERR_CANT_RESET     = 15, /*Couldn't reset interface (more than 1 handle open).*/
  PDERR_BASIC_DVR      = 16  /*Extended/High-perf command not supported */
};

enum pdapi_driver_classes
{
  PDCLASS_DIX          =  1, /* DIX Ethernet */
  PDCLASS_IEEE802_5    =  3, /* IEEE 802.5 (4Mbit Token Ring w/ 802.2 hdrs) */
  PDCLASS_LOCALTALK    =  5, /* (Packet drivers support IP/DDP only) */
  PDCLASS_SLIP         =  6, /* Serial Link IP (IP only of course) */
  PDCLASS_ARCNETRAW    =  8, /* Raw ARCNET (_not_ IEEE 878.2) */
  PDCLASS_IEEE802_3    = 11, /* IEEE 802.3 (10Mbit Ethernet w/ 802.2 hdrs) */
  PDCLASS_FDDI802_2    = 12  /* FDDI (with 802.2 hdrs) */
};

enum pdapi_driver_recvmodes
{
  PDMODE_OFF            = 1,  /* turn off receiver */
  PDMODE_DIRECTED       = 2,  /* unicast packets destined only for us */
  PDMODE_BROADCAST      = 3,  /* PDMODE_DIRECT + broadcast packets */
  PDMODE_MCAST_ASLISTED = 4,  /* PDMODE_BROADCAST + pdapi_set_multicast_list() */
  PDMODE_MCAST_ALL      = 5,  /* PDMODE_BROADCAST + all multicast */
  PDMODE_PROMISCUOUS    = 6   /* all incoming packets are accepted */
};

struct pdapi_driver_info
{
  unsigned short drvr_version; /* bx */
  unsigned char  if_class;     /* ch */
  unsigned short if_typ;       /* dx */
  unsigned char  if_number;    /* cl */
  unsigned char  drvr_functionality; /* al */
  char drvr_name[64];          /* ds:si */
};

struct pdapi_driver_params
{
  unsigned char  major_rev;      /* Revision of Packet Driver spec */
  unsigned char  minor_rev;      /*  this driver conforms to. */
  unsigned char  addr_len;       /* Length of a MAC-layer address */
  unsigned short ll_mtu;         /* MTU, including MAC headers */
  unsigned short multicast_aval; /* Buffer size for multicast addr */
  unsigned short rcv_bufs;       /* (# of back-to-back MTU rcvs) - 1 */
  unsigned short xmit_bufs;      /* (# of successive xmits) - 1 */
  unsigned short int_num;        /* interrupt for post-EOI processing */
};

extern int pdapi_get_driver_info( int intno, unsigned short handle,
                                  struct pdapi_driver_info FAR *info );
extern int pdapi_get_driver_params( int intno, 
                                    struct pdapi_driver_params FAR *params );
extern int pdapi_get_address( int intno, unsigned short handle, 
                                void FAR *addrbuf, unsigned int FAR *addrlen );
extern int pdapi_set_address( int intno, const void FAR *addrbuf, 
                                  unsigned int addrlen );
extern int pdapi_get_recv_mode( int intno, unsigned short handle, int FAR *mode );
extern int pdapi_set_recv_mode( int intno, unsigned short handle, int mode );
extern int pdapi_get_multicast_list( int intno, void FAR *addrlstbuf, 
                                     unsigned int FAR *buflen );
extern int pdapi_set_multicast_list( int intno, const void FAR *addrlstbuf, 
                                     unsigned int buflen );
extern int pdapi_send_pkt(int intno, const unsigned short FAR *lb_to_handle,
                          const void FAR *buffer, unsigned int buflen);

extern int pdapi_poll_receiver(int intno);
extern int pdapi_set_recvbuf_count( int intno, unsigned int newcount );

extern int pdapi_open_driver(int intno, unsigned short ethproto,
      int (*rcvr)( int intno, unsigned short handle, unsigned int len, 
                          const void FAR *buf ), unsigned short FAR *handle );
extern int pdapi_close_driver(int intno, unsigned short handle);
extern int pdapi_find_driver(int first_intno_to_try);

#endif /* !__PDAPI_H__ */

