/*
 * Target (compiler/C-library) specific things.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: target.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#define DONT_FORCE_BYTE_PACKING /* are structs packed _BY_DEFAULT_? */
#undef  WANT_STACK_CHECKING     /* disable stack checking for this module */

#include "printk.h"  /* _printk() */
#include "timer.h"   /* timer_*() */
#include "blocker.h" /* blocker_*() */
#include "dpmi.h"    /* __dpmi_*() */
#include "target.h"  /* ourselves */

/* =================================================================== */
/*          EFAULT support (BELONGS/DECLARED in target.*)              */
/* =================================================================== */

static int __valid_addr (int is_const, const void *ptr, unsigned int len)
{
#if defined(__DPMI32__)
  unsigned long addr = (unsigned long)ptr;

  if (!ptr)
   return 0;

  if (addr < 0x1000 || (addr >= 0xFFFFFFFFL - len))
     return (0);

  if (is_const)
  {
    if ((addr + len) <= __dpmi_get_segment_limit(_my_cs()))
      return (1);
  }
  if ((addr + len) > __dpmi_get_segment_limit(_my_ds()))
    return (0);

#else /* real mode targets */

  if (!ptr || (sizeof(ptr) == 4 && (((unsigned long)ptr) >> 16) == 0))
   return 0;

  if (((unsigned short)ptr) >= (0xffffu - len)) 
    return 0;

  is_const = is_const; /* unused */

#endif

  return (1);
}

int valid_addr_rw (void *ptr, unsigned int len) 
{ 
  return __valid_addr(0, ptr, len); 
}

int valid_addr_ro (const void *ptr, unsigned int len) 
{ 
  return __valid_addr(1, ptr, len); 
}

/* =================================================================== */
/*            atomic lock primitive (BELONGS/DECLARED in target.*)     */
/* =================================================================== */

#if !defined(HAVE_INLINE_ATOMIC_LOCK) && (UINT_MAX == USHRT_MAX)
int fastlock_trylock(fastlock_t _far *lockptr)
{
  int lacquired = 0;
  __asm push ds
  __asm lds bx,lockptr
  __asm mov ax,1
  __asm xchg [bx],al
  __asm pop ds
  __asm xor al,1
  __asm mov lacquired,ax
  return lacquired;
}
#endif

#if !defined(HAVE_INLINE_ATOMIC_LOCK) && (UINT_MAX != USHRT_MAX)
int fastlock_trylock(fastlock_t *lockptr)
{
  int lacquired = 0;
  __asm mov edx,lockptr
  __asm mov eax,1
  __asm xchg [edx],al
  __asm xor al,1
  __asm mov lacquired,eax
  return lacquired;
}
#endif

/*  =================================================================== */
/*                byte swapping (BELONGS/DECLARED in target.*)          */
/* ==================================================================== */

#if !defined(HAVE_INLINE_BYTEORDER_SWAP) /* see target.h */
unsigned long __long_swap__(unsigned long val)
{
#if defined(__BORLAND__) && defined(__FLAT__)
  __asm mov eax, val
  __asm xchg ah, al
  __asm rol eax, 16
  __asm xchg ah, al
  __asm mov val, eax
  return val;
#elif defined(__TURBOC__)
  __asm mov  ax,word ptr [val+0]
  __asm xchg ah,al
  __asm xchg ax,word ptr [val+2]
  __asm xchg ah,al
  __asm mov  word ptr [val+0],ax
  return val;
#else
  return (val<<24)|((val&0xff00)<<8)|((val&0xff0000)>>8)|(val>>24);  
#endif
}
#endif

#if !defined(HAVE_INLINE_BYTEORDER_SWAP) /* see target.h */
unsigned short __short_swap__(unsigned short val)
{
#if defined(__TURBOC__) /* which includes borland386 */
  __asm mov ax, val
  __asm xchg ah, al
  __asm mov val, ax
  return val;
#else
  return ((val >> 8)|(val << 8));
#endif
}
#endif

/* =================================================================== */
/*                           PANIC support                             */
/* =================================================================== */

#if defined(__GNUC__)
static __inline__ void __bios_putch(unsigned int c)
{
  __asm__ __volatile__("movb $0x0e,%%ah\n\t"
                       "int $0x10\n\t" : "=a"(c/* dummy */) : "0"(c) );
}
#elif defined(__TURBOC__)
#define __bios_putch(__c)                    \
{                                            \
  _AL = (unsigned char)__c;                  \
  _AH = 0x0e;                                \
  __emit__(0xCD,0x10);                       \
}
#elif defined(__WATCOMC__)
extern void __bios_putch(unsigned int c);
#pragma aux __bios_putch="mov ah,0x0e""int 0x10" parm [ax] modify exact [ax];
#endif

void __runtime_notice(const char *p)
{
  /* this has to be (DOS) interrupt safe */
  while (*p)
  {
    if (*p == '\n')
      __bios_putch('\r');
    __bios_putch(*p++);
  }
  return;
}  
void __fatal_runtime_error(const char *p)
{
  __runtime_notice(p);
  exit(255);
}

/* ===================================================================== */
/*                         C-Library replacements                        */
/* ===================================================================== */

/* All the LIBC functions for clock() use the are non-monotonic
 * and freak out applications if the time-of-day changes. 
 * In addition, the Watcom version uses time() [ick!] which is isn't
 * very efficient.
 *
 * This version of clock() is a) monotonic b) linear and c) provides
 * true CLOCKS_PER_SEC resolution on a P5+ (32bit targets only) and
 * better than 0.055s resolution everywhere else.
 *
 * NOTE that there is no need to subtract the values obtained from 
 * timer_*_count() from a base init value. timer_*_count() functions 
 * are all self-initializing and zero-based.
*/
#include <time.h>
#if defined(__WATCOMC__)
clock_t clock(void)
{
  #define __init_clock clock /* self initializing */
  #if (CLOCKS_PER_SEC == 100) /* watcom c <= 9.0 */
  return (clock_t)timer_hsec_count();
  #elif (CLOCKS_PER_SEC == 1000) /* watcom c >= 10.0 */
  return (clock_t)timer_msec_count();
  #else
  #error huh?
  #endif
}
#elif defined(__TURBOC__)
clock_t clock(void)
{
  #define __init_clock clock /* self initializing */
  #if (CLOCKS_PER_SEC == 18.2) /* old turboc */
  return (clock_t)timer_tick_count();
  #elif (CLOCKS_PER_SEC == 1000.0) /* newer tc/bc */
  return (clock_t)timer_msec_count();
  #else
  #error huh?
  #endif
}
#elif defined(__DJGPP__)
static clock_t __init_clock(void)
{
  unsigned long t;
#if 0
  #pragma weak _clock
  if (clock)
#endif
  {
    static int my_bss_count = -1;
    if (my_bss_count < __bss_count) /* always LT compare! */
    {
      int cs_alias = __my_dpmi_get_csalias();
      if (cs_alias != -1)
      {
        char buf[8];
        t = (unsigned)clock;
        memcpy( buf,  (const void *)t, sizeof(buf) ); /* force page in */
        if (__my_dpmi_lock_code( (const void *)t, sizeof(buf)) == 0)
        {
          t = ((unsigned)__init_clock) - ((unsigned)clock);
          buf[0] = 0xe9;    /* "jmp" (relative) */
          *((unsigned long *)(&buf[1])) = t - 5; /* 5==sizeof('jmp xyz') */
          movedata(_my_ds(), (unsigned)(&buf[0]), cs_alias, (unsigned)clock, 5);
        }  
      }
      my_bss_count = __bss_count;
    }
  }
  /* the timer_xxx functions are __bss_count aware */
  #if (CLOCKS_PER_SEC == 91) /* 91 == (18.2 * 5) */
  t = timer_tick_count();
  return (clock_t)(t+(t << 2)); /* leal eax,[eax+eax*4] */
  #else
  #error "Huh?"
  #endif
}
#else  /* not __BORLANDC__|__TURBOC__|__DJGPP__ */
#define __init_clock() NO_OPERATION()
#endif 

/* ===================================================================== */
/*                        stack checking                                 */
/* ===================================================================== */

#if defined(__cplusplus)
extern "C" {
#endif
#if defined(__WATCOMC__)
  extern unsigned int _SP(void);
  #if defined(__386__)
  #pragma aux _SP = "mov eax,esp" value [eax] modify exact [eax];
  #else
  #pragma aux _SP = "mov ax,sp" value [ax] modify exact [ax];
  #endif
#elif defined(__DJGPP__)
  extern unsigned _stklen;
  extern unsigned __djgpp_stack_limit;
#elif defined(__HIGHC__)
  extern unsigned int _mwstack_limit[];
#else
  void *__try_alloca__(size_t sz) /* not static!, not inline! */
  {
    void *v = alloca(sz);
    return v;
  }
#endif
#if defined(__cplusplus)
}
#endif

static const char *_my_utoa(unsigned val, char *buf, unsigned int buflen)
{
  buf += buflen;
  *--buf = '\0';
  while (val)
  {
    *--buf = (char)('0'+(val%10));
    val /= 10;
  }
  if (!*buf)
    *--buf = '0';
  return buf;
}

static void __check_stack_limit(void)
{
  #if (UINT_MAX > USHRT_MAX)
  #define NEEDED_STK (8u * 1024u)
  #else
  #define NEEDED_STK (4u * 1024u)
  #endif
  unsigned values[3];
  const char *msg = (const char *)0;
  values[0] = 0;

  #if defined(__HIGHC__)
  {
    unsigned int avail[2]; /* array to force on the stack */
    avail[1] = (((unsigned int)&avail[0]) - _mwstack_limit[1]);
    if (avail[1] < NEEDED_STK)
    {
      values[0] = avail[1];
      values[1] = NEEDED_STK-avail[1];
      values[2] = NEEDED_STK+2048;
      msg = "Relink your application with '-stack %u'\n";
    }
  }
  #elif defined(__WATCOMC__)
  {
    unsigned int avail = _SP() - _STACKLOW;
    if (avail < NEEDED_STK)
    {
      unsigned int _stklen = _STACKTOP - _STACKLOW;
      unsigned reqd = (_stklen + (NEEDED_STK - avail));
      values[0] = avail;
      values[1] = NEEDED_STK-avail;
      values[2] = (reqd + 1023u)/1024u;
      msg = "Relink your application with 'option stack=%uK'\n";
    }
  }
  #elif defined(__DJGPP__)
  {
    unsigned int sp, avail;
    __asm__ __volatile__ ("movl %%esp,%k0\n" : "=r" (sp) : );
    avail = (sp - __djgpp_stack_limit);
    if (avail < NEEDED_STK)
    {
      values[0] = avail;
      values[1] = NEEDED_STK-avail;
      values[2] = _stklen + (NEEDED_STK - avail);
      msg = "Use 'stubedit.exe' or "
            "declare 'int _stklen = %u;' in your application.\n";
    }
  }
  #else
  {
    size_t avail = 0, next = NEEDED_STK << 1;
    while (avail < NEEDED_STK)
    {
      if (__try_alloca__(avail + next))
        avail += next;
      else if (next > (sizeof(next)<<1))
        next >>= 1;
      else
        break;
    }
    if (avail < NEEDED_STK)
    {
      values[0] = avail;
      values[1] = NEEDED_STK-avail;
      values[2] = _stklen + (NEEDED_STK - avail);
      msg = "Use 'unsigned _Cdecl _stklen = %u;' somewhere (in global scope).\n";
    }
  }
  #endif
  
  if (msg)
  {
    unsigned int i;
    char buf[80];
    __runtime_notice("Insufficient available stack space (");
    __runtime_notice(_my_utoa(values[0], buf, sizeof(buf)));
    __runtime_notice("), ");
    __runtime_notice(_my_utoa(values[1], buf, sizeof(buf)));
    __runtime_notice(" bytes more are needed.\n");
    i = 0;
    while (*msg && *msg != '%')
      buf[i++] = *msg++;
    buf[i] = '\0';
    __runtime_notice(buf);
    __runtime_notice(_my_utoa(values[2], buf, sizeof(buf)));
    __runtime_notice(++msg);
    exit(255);
  }
  return;
}

/* ===================================================================== */
/*                        lib link order test                            */
/* ===================================================================== */

static void __check_link_order(void)
{
#if !defined(__DJGPP__)
  /* if delay() is not very close to kern_delay() then the 
   * standard C library was probably linked before our stuff.
   * (which is a common occurence under TurboC/BCC, but possible
   * for all).
   *
   * This test cannot be in timer module because the offset
   * would be then be resolved at compile time, and would thus
   * be meaningless.
  */
  const char FAR *addr1 = (const char FAR *)delay;
  const char FAR *addr2 = (const char FAR *)blocker_cdelay;
  if ((addr1 < addr2 && (addr2 - addr1) > 0x300) ||
      (addr2 < addr1 && (addr1 - addr2) > 0x300))
  {
    _printk("PANIC: bad library link order\n");
    exit(255);
  }
  /* _printk("delay=%p, kern_delay=%p\n", addr1, addr2); */
#endif
  return;
}

/* ===================================================================== */
/*                  prime the heaps (so allocs are faster)               */
/* ===================================================================== */

static void __check_memory(void)
{
  void FAR *ptrs[10];
  unsigned int i, n = 0;
  unsigned int totalkb = 0;
  unsigned int maxsize = USHRT_MAX;
  unsigned int cursize;

  ptrs[0] = (void FAR *)0;
  for (i = 0; i < (sizeof(ptrs)/sizeof(ptrs[0])); i++)
  {
    ptrs[i] = (void FAR *)0;
    cursize = maxsize;
    while (cursize)
    {
      ptrs[i] = FAR_MALLOC(cursize);
      if (ptrs[i])
        break;
      if (i == 0 && cursize > (maxsize - 1024u))
        cursize -= 16u;
      else
        cursize -= 1024u;
    }
    if (!ptrs[i])
      break;
    n++;
    totalkb += cursize>>10u;
    if (i == 0)
      maxsize = cursize;
    else if (cursize < maxsize)
      break;
  }
  for (i = 0; i < n; i++)
  {
    FAR_FREE(ptrs[i]);
  }
  /*
  _printk("Primed (far) mem = %ukB in %u segments (largest = %u)\n", 
                              totalkb, n, maxsize );
  */

  #if defined(__TINY__) || defined(__SMALL__) || defined(__MEDIUM__)
  {
    cursize = USHRT_MAX;
    while (cursize)
    {
      void *v = malloc(cursize);
      if (v)
      {
        free(v);
        if (cursize != USHRT_MAX)
        {
          maxsize = cursize + 1024u; 
          while (cursize < maxsize)
          {
            v = malloc(cursize + 16);
            if (!v)
              break;
            free(v); 
            cursize += 16;
          }
        }
        break;
      }
      cursize -= 1024u;
    }
    /* _printk("Max alloc (near) mem = %u bytes\n", cursize ); 
    */
  }
  #endif

  return;
}

#if defined(__WATCOMC__) && !defined(__386__)
/* 
** "Once a call to _fmalloc or halloc has been made, 
** you may not be able to allocate any memory with malloc unless 
** space has been reserved for the near heap using either malloc, 
** sbrk or _nheapgrow."
*/
#include "ctordtor.h"
#pragma off(unreferenced)
/* constructors auto-generated with __SETUP_CONSTRUCTOR_FOR__+ctordtor.h */
/* are INIT_PRIORITY_PROGRAM-1, so do this one prio higher.              */
static XI( __ctor_heapgrow, _heapgrow, INIT_PRIORITY_PROGRAM-2 );
#pragma on(unreferenced)
#endif

/* ===================================================================== */
/*                     struct member alignment check                     */
/* ===================================================================== */

/* 
** Check that EITHER '#pragma pack(1)' is supported, OR the library
** is compiled with default packing on byte boundaries.
**
** Note that we #defined DONT_FORCE_BYTE_PACKING before including
** target.h, otherwise target.h will turn on byte packing by default
** using compiler-specific pragmas.
*/
#pragma pack(1) /* align all structure members on byte boundaries */
struct _al_check_t
{
  uint8_t  m1;
  uint16_t m2;
  uint32_t m3;
  struct {
     uint16_t m4a:1, m4b:2, m4c:3, m4d:4, m4e:6;
  } m4;
  struct {
     int m5a:1, m5b:2, m5c:3, m5d:4, m5e:6, m6f:16; 
  } m5;
  uint32_t m6;
  uint16_t m7;
  uint8_t  m8;
};
#pragma pack() /* restore default packing */


static void __check_alignment(void)
{
  struct _al_check_t alc;
  if ( ( (((uint8_t *)&alc.m1) + sizeof(alc.m1)) != ((uint8_t *)&alc.m2) ) ||
       ( (((uint8_t *)&alc.m2) + sizeof(alc.m2)) != ((uint8_t *)&alc.m3) ) ||
       ( (((uint8_t *)&alc.m3) + sizeof(alc.m3)) != ((uint8_t *)&alc.m4) ) ||
       ( (((uint8_t *)&alc.m4) + sizeof(alc.m4)) != ((uint8_t *)&alc.m5) ) ||
       ( (((uint8_t *)&alc.m5) + sizeof(alc.m5)) != ((uint8_t *)&alc.m6) ) ||
       ( (((uint8_t *)&alc.m6) + sizeof(alc.m6)) != ((uint8_t *)&alc.m7) ) ||
       ( (((uint8_t *)&alc.m7) + sizeof(alc.m7)) != ((uint8_t *)&alc.m8) ) ||
       sizeof(alc.m4) != sizeof(uint16_t) ||
       sizeof(alc.m5) != sizeof(uint32_t) || 
       ( (sizeof(alc.m1)+sizeof(alc.m2)+sizeof(alc.m3)+sizeof(alc.m4)+
          sizeof(alc.m5)+sizeof(alc.m6)+sizeof(alc.m7)+sizeof(alc.m8)) 
          != sizeof(alc) )
     )
  {
   /*
   ** '#pragma pack(1)' is not supported by this compiler and the library
   ** was not compiled with default packing on byte boundaries.
   */
    __fatal_runtime_error("PANIC: struct member alignment error\n"
    "Rebuild the library with byte alignment (packing on byte boundaries).\n");
  }
}

/* ===================================================================== */
/*                        initialization                                 */
/* ===================================================================== */

static void __ctor_target(void)
{
#if defined(__DJGPP__)     /* _osmajor/_osminor not set in crt0.o */
  _get_dos_version (0);
#endif

  __check_alignment();   /* dies on error */
  __check_stack_limit(); /* dies on error */
  __check_link_order();  /* dies on error */
  __check_memory();
  __init_clock();

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_target
#include "ctordtor.h"
