/*
 * Target (compiler/C-library) specific things.
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: target.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __KERN_TARGET__
#define __KERN_TARGET__(x) __tkern_ ## x

#if defined(__WATCOMC__) && !defined(_ENABLE_AUTODEPEND)
#pragma read_only_file;
#endif

#undef KERNEL_COMPILER_RECOGNIZED

/* ==================================================================== */
/*         TurboC/BorlandC includes and portability work-arounds        */
/* ==================================================================== */

#if defined(__TURBOC__) || defined(__BORLANDC__)
  #define KERNEL_COMPILER_RECOGNIZED

  #include <limits.h>
  #include <stdlib.h> /* sys_nerr/_sys_nerr/_MAX_PATH */
  #include <string.h>
  #include <conio.h>  
  #include <dos.h>

  #if (__TURBOC__ <= 0x301)  /* TCC <= 2.01 */
    #include <mem.h>
    #include <alloc.h>
    #define _cdecl _Cdecl
    #define _far   far
  #else
    #include <memory.h>
    #include <malloc.h>
  #endif
  #if defined(__FLAT__) && defined(HAVE_PHARLAP_SDK)
  #include <pharlap.h>
  #endif

  #if defined(__TURBOC__) && (__TURBOC__ <= 0x410) /* TCC/BCC <= 3.1 */
    #define  __get_sys_nerr_ptr() (&(sys_nerr))
    /* this and older versions also do not have a 'w' member in REGS */
  #else
    #define  __get_sys_nerr_ptr() (&(_sys_nerr))
  #endif

  /* tc/bcc complains if an if() is followed by a ';' */
  #define NO_OPERATION()         ((void)0)            /* nop */
  #define UNREFERENCED(__var__)  ((void)(__var__))
  #define __bss_count 0

  /* getenv() is rewired in target.c */
  extern char _FAR * __kgetenv(const char _FAR *envvar);
  #undef  getenv
  #define getenv __kgetenv

  #if (defined(__BORLANDC__) && (defined(__FLAT__) || defined(__HUGE__) || \
                                 defined(__LARGE__) || defined(__MEDIUM__)) )
     /* make string/memory functions inline */
     #define strlen        __strlen__
     #define strncpy       __strncpy__
     #define strrchr       __strrchr__
     #define strncmp       __strncmp__
     #define memset        __memset__
     #define memcpy        __memcpy__
     #define memcmp        __memcmp__
     #define memchr        __memchr__
     #define movedata      __movedata__ /* this one is important for __FLAT__*/ 
  #endif
  #if defined(__FLAT__)
    extern int __bcc32_int386(int intno, union REGS *ireg, union REGS *oreg);
    #define int386x  __bcc32_int386x
    extern int __bcc32_int386x(int intno, union REGS *ireg, union REGS *oreg, 
                               struct SREGS *sreg);
    #define int386   __bcc32_int386
    extern void __bcc32_segread(struct SREGS *sreg);
    #define segread  __bcc32_segread
    exterm void __bcc32_movedata( unsigned src_sel, unsigned src_off,
                                  unsigned dst_sel, unsigned dst_off,
                                  unsigned length );
    #define movedata __bcc32_movedata

    #define stackset(stk)  __asm { mov  ax,ss;    \
                                   mov  ebx, esp; \
                                   mov  cx,  ds;  \
                                   mov  ss,  cx;  \
                                   lea  esp, stk; \
                                   push eax;      \
                                   push ebx;      \
                                 }
    #define stackrestore() __asm { lss esp, [esp] }
    #undef  FP_SEG /* not safe and not needed */
    #undef  FP_OFF /* not safe and not needed */
  #endif

#endif

/* ==================================================================== */
/*         Metaware HighC #includes and portability work-arounds        */
/* ==================================================================== */

#if defined(__HIGHC__)
  #define KERNEL_COMPILER_RECOGNIZED

  #include <limits.h>
  #include <stdlib.h>
  #include <string.h>
  #include <conio.h>  
  #include <dos.h>
  #include <pharlap.h>
  #include <init.h>  /* _mwlsl(), _msgetcs() */

  #if (UINT_MAX == USHRT_MAX)
    #error 16bit target is not supported
  #endif

  #define __get_sys_nerr_ptr() (&(_sys_nerr))
  #define _enable()            _inline(0xFB)       /* sti */
  #define _disable()           _inline(0xFA)       /* cli */
  #define NO_OPERATION()       /* no complaint if if() is followed by ';' */
  #define UNREFERENCED(__v__)  ((void)(__v__))
  #define __bss_count 0

  #undef  FP_SEG /* not safe and not needed */
  #undef  FP_OFF /* not safe and not needed */

#endif

/* ==================================================================== */
/*         Microsoft C std #includes and portability work-arounds       */
/* ==================================================================== */

#if defined(_MSC_VER)
  #define KERNEL_COMPILER_RECOGNIZED

  #include <limits.h>
  #include <stdlib.h>
  #include <string.h>
  #include <malloc.h>
  #include <conio.h>
  #include <dos.h>

  #if (UINT_MAX > USHRT_MAX)
    #error 32bit target is not supported
  #endif
  /* Microsoft C doesn't have __SMALL__,  __LARGE__  */
  #undef __TINY__
  #undef __SMALL__
  #undef __COMPACT__  
  #undef __LARGE__  
  #undef __MEDIUM__  
  #undef __HUGE__  
  #if defined(M_I86TM)    /* <= 64K code + data */
    #define __TINY__
  #elif defined(M_I86SM)  /* <= 64K code, <= 64K data */ 
    #define __SMALL__
  #elif defined(M_I86MM)  /* <= 1M code, <= 64K data */
    #define __MEDIUM__
  #elif defined(M_I86CM)  /* <= 64K code, <= 1023K data */
    #define __COMPACT__
  #elif defined(M_I86LM)  /* <= 1M code + data, 64K ptr limit */
    #define __LARGE__
  #elif defined(M_I86HM)  /* <= 1M code + data, 1M ptr limit */
    #define __HUGE__
  #else
    #error Unknown memory model
  #endif

  #define  __get_sys_nerr_ptr() (&(_sys_nerr))
  #define NO_OPERATION()        /* no complaint if if() is followed by ';' */
  #define UNREFERENCED(__v__)   ((void)(__v__))
  #define __bss_count 0
  #define MK_FP(s,o)            ((void _far *)(((unsigned long)(s) << 16) + \
                                         (unsigned long)(o)))

#endif

/* ==================================================================== */
/*               GnuC std #includes and portability work-arounds        */
/* ==================================================================== */

#if defined(__GNUC__)
  #define KERNEL_COMPILER_RECOGNIZED

  #if (__GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 7))
    #error I need GCC 2.7.x or later
  #endif
  #if !defined(__DJGPP__)
    #error GNUC only in combination with DJGPP
  #endif

  #include <limits.h>
  #include <stdlib.h>
  #include <string.h>
  #include <unistd.h>
  #include <conio.h>  /* must be included before dos.h */
  #include <dos.h>
  #include <libc/bss.h> /* extern int __bss_count */
  #include <libc/farptrgs.h> /* _farpeekl() */

  #define __get_sys_nerr_ptr() (&(sys_nerr))
  /* #define _enable()            __asm__ __volatile__ ("sti") */
  /* #define _disable()           __asm__ __volatile__ ("cli") */
  #define NO_OPERATION() ((void)0) /* warning: empty body in an if-statement */
  #define UNREFERENCED(__v__)  ((void)(__v__))

  #if defined(__cplusplus) 
  /* emit this into every object file that is compiled with c++ */
  /* we don't actually use c++ elements, and don't want lib???cx */ 
  __asm__(".weak ___gxx_personality_v0\n ___gxx_personality_v0 = _errno\n");
  /* this statement causes U references to __gxx_personality_v0 to vanish. */
  #endif

#endif

/* ==================================================================== */
/*               Watcom C #includes and portability work-arounds        */
/* ==================================================================== */

#if defined(__WATCOMC__)
  #define KERNEL_COMPILER_RECOGNIZED

  #include <limits.h>
  #include <dos.h>
  #include <i86.h>
  #include <conio.h>  /* inp()/outp() */
  #include <malloc.h> /* stackaval */
  #include <stdlib.h> /* sys_nerr */
  #include <string.h>
  #if defined(__386__) && defined(HAVE_PHARLAP_SDK)
  #include <pharlap.h>
  #endif

  #if !defined(__SMALL__) && !defined(__TINY__) && !defined(__COMPACT__)
  #pragma intrinsic(strcmp,memset)
  #pragma intrinsic(_enable,_disable)
  #pragma intrinsic(movedata)
  #endif

  #if !defined(__cplusplus)
  #pragma disable_message(604); /*'must look ahead to determine whether ... */
                   /* ... construct is a declaration/type or an expression' */
  #pragma disable_message(594); /*'construct resolved as a declaration/type'*/
  #endif

  #define __get_sys_nerr_ptr()  (&(sys_nerr))
  #define NO_OPERATION()        /* no complaint if if() is followed by ';' */
  #define UNREFERENCED(__v__)   ((void)(__v__))
  #define __bss_count 0

  #if defined(__386__)
  extern void stackset2(void *data, unsigned int stacksz, unsigned *lowptr);
  #pragma aux stackset2 = "push dword ptr [esi]" "push esi" "mov [esi],edx" \
                          "add edx,eax" "sub edx,4" "and dl,0fch" \
                          "mov  ax, ss" "mov ebx,esp" "mov cx,ds" "mov ss,cx" \
                          "mov  esp, edx" "push eax" "push ebx" \
                          parm [edx] [eax] [esi] modify exact [eax ebx ecx edx];   
  extern void stackrestore (void);
  #pragma aux stackrestore = "lss esp, [esp]" \
                             "pop eax"        \
                             "pop dword ptr [eax]" \
                             modify exact [eax];
  #else
  extern void stackset2(void _far *data, unsigned int stacksz, 
                       unsigned _far *lowptr);
  #pragma aux stackset2 = "mov  es,cx"   \
                          "push word ptr es:[bx]" \
                          "push es"      \
                          "push bx"      \
                          "mov  es:[bx],ax" \
                          "add  ax,si"   \
                          "sub  ax,2"    \
                          "and  al,0feh" \
                          "mov  cx,ss"   \
                          "mov  bx,sp"   \
                          "mov  ss,dx"   \
                          "mov  sp,ax"   \
                          "push cx"      \
                          "push bx"      \
               parm [ax dx] [si] [bx cx] modify exact [ax bx cx dx es];
  extern void stackrestore(void);
  #pragma aux stackrestore = "pop bx"    \
                             "pop ax"    \
                             "mov ss,ax" \
                             "mov sp,bx" \
                             "pop bx"    \
                             "mov ax,es" \
                             "pop es"    \
                             "pop word ptr es:[bx]" \
                             "mov es,ax" \
                             modify exact [bx ax];
  #endif /* stackset/stackrestore */

  #ifdef __cplusplus
  #define CSYMBOL "C"
  #else
  #define CSYMBOL
  #endif
  extern CSYMBOL unsigned _STACKLOW;
  extern CSYMBOL unsigned _STACKTOP;

  #define stackset(__data,__stksize) \
          stackset2(__data,__stksize,&_STACKLOW) 

  /* The library is compiled with -zl (no default libraries) in order
   * to not require any specific mathlib. But -zl also suppresses
   * inclusion of the clib, so add that manually here.
  */
  #if defined(__386__) && defined(__SW_3R)
  #pragma library ("clib3r.lib");
  #elif defined(__386__) && defined(__SW_3S)
  #pragma library ("clib3s.lib");
  #elif defined(__TINY__) || defined(__SMALL__)
  #pragma library ("clibs.lib");
  #elif defined(__MEDIUM__)
  #pragma library ("clibm.lib");
  #elif defined(__COMPACT__)
  #pragma library ("clibc.lib");
  #elif defined(__LARGE__)
  #pragma library ("clibl.lib");
  #elif defined(__HUGE__)
  #pragma library ("clibh.lib");
  #endif  

#endif /* WATCOMC */

/* ==================================================================== */
/*                        end of compiler stuff                         */
/* ==================================================================== */

#if !defined(KERNEL_COMPILER_RECOGNIZED)
#error UNRECOGNIZED COMPILER
#endif
#undef KERNEL_COMPILER_RECOGNIZED /* no longer needed */

/* =================================================================== */
/*                           PANIC support                             */
/* =================================================================== */

#define __fatal_runtime_error __KERN_TARGET__(fatal_runtime_error)
#define __runtime_notice      __KERN_TARGET__(runtime_notice)

extern void __fatal_runtime_error(const char *p);
extern void __runtime_notice(const char *p);
#define __nonfatal_runtime_error __runtime_notice

/* =================================================================== */
/*                           EFAULT support                            */
/* =================================================================== */

#define valid_addr_rw    __KERN_TARGET__(valid_addr_rw)
#define valid_addr_ro    __KERN_TARGET__(valid_addr_ro)
#define valid_addr       valid_addr_rw

extern int valid_addr_rw (void *addr, unsigned int len);
extern int valid_addr_ro (const void *addr, unsigned int len);

/* =================================================================== */
/*                             byte swapping                           */
/* =================================================================== */

#if defined(__GNUC__)
# define HAVE_INLINE_BYTEORDER_SWAP
  static __inline__ unsigned long __long_swap__(unsigned long x)
  {
    __asm__ ("xchgb %b0, %h0\n\t"   /* swap lower bytes  */ \
             "rorl  $16,%0\n\t"     /* swap words        */ \
             "xchgb %b0,%h0"        /* swap higher bytes */ \
            : "=q" (x) : "0" (x));
    return (x);
  }
  static __inline__ unsigned short __short_swap__(unsigned short x)
  {
    __asm__ ("xchgb %b0, %h0" : "=q" (x) : "0" (x));
    return (x);
  }
#elif defined(__TURBOC__) && defined(__cplusplus)
# define HAVE_INLINE_BYTEORDER_SWAP
 inline unsigned long __long_swap__(unsigned long val)
 {
   #if defined(__FLAT__)
   _EAX = val;
   __emit__(0x86, 0xC4);       /* xchg ah,al */
   __emit__(0xC1, 0xC8, 0x10); /* ror eax,16 */
   __emit__(0x86, 0xC4);       /* xchg ah,al */
   return _EAX;
   #else
   _AX = *(((unsigned short *)&val)+0);
   _DX = *(((unsigned short *)&val)+1);
   __emit__(0x86, 0xC4);       /* xchg ah,al */
   __emit__(0x92);             /* xchg dx,ax */
   __emit__(0x86, 0xC4);       /* xchg ah,al */
   *(((unsigned short *)&val)+1) = _DX;
   *(((unsigned short *)&val)+0) = _AX;
   return val;
   #endif
 }
 static inline unsigned short __short_swap__(unsigned short val)   
 {
   _AX = val;
   __emit__(0x86, 0xC4);       /* xchg ah,al */
   return _AX;
 }
#elif defined(__WATCOMC__)
# define HAVE_INLINE_BYTEORDER_SWAP
  extern unsigned short __short_swap__(unsigned short x);
  extern unsigned long  __long_swap__(unsigned long x);
  #if defined(__386__)
    #pragma aux  __long_swap__ = \
                 "xchg al, ah"  \
                 "ror  eax, 16" \
                 "xchg al, ah"  \
                 parm [eax] value [eax] modify exact [eax];
  #else
    #pragma aux  __long_swap__ = \
                 "xchg al, dh"  \
                 "xchg ah, dl"  \
                 parm [dx ax] value [dx ax] modify exact [dx ax];
  #endif
  #pragma aux  __short_swap__ = \
                "xchg al, ah" \
                parm [ax] value [ax] modify exact [ax];
#else
   #define __long_swap__    __KERN_TARGET__(long_swap__)
   #define __short_swap__   __KERN_TARGET__(short_swap__)
   extern unsigned long  __long_swap__(unsigned long val);
   extern unsigned short __short_swap__(unsigned short val);
#endif

#define HTONL(x) __long_swap__(x)
#define HTONS(x) __short_swap__(x)
#define NTOHL(x) __long_swap__(x)
#define NTOHS(x) __short_swap__(x)

/* =================================================================== */
/*                           atomic locks                              */
/* =================================================================== */

typedef struct { volatile int spl; } fastlock_t; 
#define FASTLOCK_INITIALIZER_UNLOCKED {0}

#define fastlock_unlock(__m) (__m)->spl = 0
#define fastlock_trylock __KERN_TARGET__(fastlock_trylock)

#if defined(__WATCOMC__) && defined(__386__)
extern int fastlock_trylock(fastlock_t *m);
#pragma aux fastlock_trylock = "mov eax,1" "xchg [edx],eax" "xor eax, 1" \
                      parm [edx] value [eax] modify exact [eax];
#define HAVE_INLINE_ATOMIC_LOCK
#elif defined(__WATCOMC__) && !defined(__386__)
extern int fastlock_trylock(fastlock_t _far *m);
/* the following isn't as unecessarily convoluted as it appears. */
#pragma aux fastlock_trylock = "push ds" "xchg dx,ax" "mov ds,ax" \
  "xchg dx,bx" "mov ax,1" "xchg [bx],al" "xor al,1" "mov bx,dx" "pop ds" \
                      parm [ax dx] value [ax] modify exact [ax dx];
#define HAVE_INLINE_ATOMIC_LOCK
#elif defined(__BORLANDC__) && defined(__FLAT__)
static __inline int fastlock_trylock(fastlock_t *m) 
{
  /* BCC can't do inline assembler in inline functions */
  _EDX = (unsigned long)&(m->spl);
  _EAX = 1;
  __emit__(0x87, 0x02); /* xchg [edx],eax */
  return _EAX ^ 1;
}
#define HAVE_INLINE_ATOMIC_LOCK
#elif defined(__GNUC__)
static __inline int fastlock_trylock(fastlock_t *m)
{
  int lacquired = 0;
  /* gcc is sometimes too clever */
  struct __fool_gcc_volatile { unsigned long a[100]; };
  __asm__ __volatile__(
                "movl $1,%0    \n\t" \
                "xchgl %0,%1   \n\t" \
                "xorl $1,%0    \n\t"
                : "=r"(lacquired)
                : "m"(*((volatile struct __fool_gcc_volatile *)(&(m->spl))))
                : "memory");
  return lacquired;
}
#define HAVE_INLINE_ATOMIC_LOCK
#elif (UINT_MAX == USHRT_MAX)
extern int fastlock_trylock(fastlock_t _far *m);
#else
extern int fastlock_trylock(fastlock_t *m);
#endif

/* =================================================================== */
/*                        far heap functions                           */
/* =================================================================== */

#if (UINT_MAX == USHRT_MAX) /* 16 bit target */
  #define FAR _far /* no reason why bufs need to be on the local heap */
  #define FAR_MEMCPY  _fmemcpy
  #define FAR_MEMMOVE _fmemmove
  #define FAR_MEMCMP  _fmemcmp
  #define FAR_MEMSET  _fmemset
  #define FAR_MALLOC  _fmalloc
  #define FAR_REALLOC _frealloc
  #define FAR_FREE    _ffree
  #define FAR_STRLEN  _fstrlen
  #define FAR_STRCMP  _fstrcmp
  #define FAR_STRCMPI _fstrcmpi
#else
  #define FAR
  #define FAR_MEMCPY   memcpy
  #define FAR_MEMMOVE  memmove
  #define FAR_MEMCMP   memcmp
  #define FAR_MEMSET   memset
  #define FAR_MALLOC   malloc
  #define FAR_REALLOC  realloc
  #define FAR_FREE     free
  #define FAR_STRLEN   strlen
  #define FAR_STRCMP   strcmp
  #define FAR_STRCMPI  strcmpi
#endif

/* =================================================================== */
/*                          basic types                                */
/* =================================================================== */

typedef unsigned char  uint8_t;   /*  8 bit unsigned */
typedef unsigned short uint16_t;  /* 16 bit unsigned */
typedef unsigned long  uint32_t;  /* 32 bit unsigned */

/* =================================================================== */
/*                          disable stack checking                     */
/* =================================================================== */

#if !defined(WANT_STACK_CHECKING)
#  if defined(__HIGHC__) 
#    pragma off(check_stack)
#  elif defined(__WATCOMC__)
#    pragma off(check_stack)
#  elif (defined(__TURBOC__) && (__TURBOC__ <= 0x301))
     /* nothing - prevent from falling into next if */
#  elif (defined(__TURBOC__) || defined(__BORLANDC__))
#    pragma option -N- 
#  elif defined(__GNUC__)
     /* nothing - no stack check support */
#  else
#    error huh?
#  endif
#endif

/* =================================================================== */
/*                          force byte alignment                       */
/* =================================================================== */

#if !defined(DONT_FORCE_BYTE_PACKING)
  /* sample only - otherwise all modules would have to include target.h */
#elif (defined(_MSC_VER) && (_MSC_VER >= 700)) || \
   (defined(__BORLANDC__) && defined(__FLAT__))
#  pragma warning(disable:4103)
#  pragma pack(1)
#elif defined(__CCDL__)
#  pragma packed
#elif defined(__HIGHC__)
#  pragma Off(Align_members)
#elif defined(__WATCOMC__) && (__WATCOMC__ >= 1000)
#  pragma pack(__push,1); /* #pragma pack(1) works too */
#elif defined(__TURBOC__)
#  pragma option -a-
#elif defined(__GNUC__)
#  pragma pack(1)
#else
#  error huh?
#endif

/* =================================================================== */
/*                    32bit multiply and divide                        */
/* =================================================================== */

#if 1 /* (UINT_MAX > USHRT_MAX) */
#define MUL32(__a,__i) (((uint32_t)(__a))*((uint32_t)(__i)))
#define DIV32(__a,__i) ((__a)/(__i)) 
#define MOD32(__a,__i) ((__a)%(__i)) 
#else
extern uint32_t __kern_MUL32(uint32_t and, uint32_t ier);
extern uint32_t __kern_DIV32(uint32_t and, uint32_t ier);
extern uint32_t __kern_MOD32(uint32_t and, uint32_t ier);
#define MUL32(__a,__i) __kern_MUL32((__a),(__i))
#define DIV32(__a,__i) __kern_DIV32((__a),(__i))
#define MOD32(__a,__i) __kern_MOD32((__a),(__i))
#endif

#endif  /* __KERN_TARGET__ */

