/*
 * TCP transport provider
 *
 * History of tcp.c
 *
 *  1986/04/16: Created by Geof Cooper for the 68000 (TinyTCP)
 *                   Copyright (C) 1986, IMAGEN Corporation
 *                   "This code may be duplicated in whole or in part 
 *                   provided that [1] there is no commercial gain involved 
 *                   in the duplication, and [2] that this copyright notice 
 *                   is preserved on all copies.  Any other duplication
 *                   requires written notice of the author." 
 *              Ported to the 8086 by R. Rodman, 
 *                   changes by him are (L)GPL, but the original copyright
 *                   remains.
 *                   In case you were wondering... this is when 'BIG_ENDIAN',
 *                   and 'LITTLE_ENDIAN' came along.
 *  1990-1993 : Modified and updated by Erick Engelke for use with 
 *                   packet drivers.  Initial implementation of 
 *                   Nagle Algo, TCP PUSH and other important things.
 *                   Copyright (c) 1990, 1991, 1992, 1993 Erick Engelke
 *                   "You may freely use this library to produce programs 
 *                   which you may give away or sell in source format, or 
 *                   as compiled executables, or both.
 *                   You may not sell this library or a portion thereof or 
 *                   distribute modified versions the library code in 
 *                   either source or OBJect or LIBrary format without 
 *                   written permission from the author.  The author still 
 *                   maintains copyright privileges and will assert the 
 *                   privileges of ownership for purposes of royalties 
 *                   such distributions."
 *  1997 : Geof Cooper and Rick Rodman release TinyTCP 
 *                   under the Gnu Public License. 
 * ~1998 : "The TCP/IP software used in the Rabbit 2000 TCP/IP Development 
 *                   Kit is designed for use only with Rabbit Semiconductor 
 *                   chips, and is used under licence from Erick Engelke.". 
 *                   La, la, la.
 *  May 2001 : Partial to complete rewrite of the TinyTCP library
 *                   by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *                   including ~25% rewrite of this file (tcp.c) with pieces
 *                   incorporated from Erick Engelke's release of 
 *                   TinyTCP (stealthily renamed Wattcp by Erick), 
 *                   with corrections and modifications by S. Lawson, 
 *                   Quentin Smart, Gisle Vanem and numerous others 
 *                   none of which Erick bothered to credit either.
 * 
 * Licenses for the library:
 *   - (L)GPL for code by Geof Cooper and Rick Rodman. This covers 
 *     pieces throughout the library, and are noted as such.
 *     The GNU General Public License is at
 *        http://www.gnu.org/copyleft/gpl.html
 *     The GNU Lesser Public License is at 
 *        http://www.gnu.org/copyleft/lesser.html
 *   - Some pieces of tcp.c are Copyright (C) Erick Engelke, who gives you
 *     the license to "freely use" the Erick Engelke portions of tcp.c "to 
 *     produce programs which you may give away or sell in source format, 
 *     or as compiled executables, or both. You may not sell (the 
 *     Erick Engelke portions of tcp.c) or a portion thereof or distribute 
 *     modified versions therof in either source or OBJect or LIBrary format
 *     without written permission from the author.  The author still 
 *     maintains copyright privileges and will assert the privileges of 
 *     ownership for purposes of royalties such distributions."
 *   - Portions of IGMP and multicast support are by Jim Martin
 *     and are noted as such.
 *   - Everything else is my (Cyrus Patel) code, and irrespective of 
 *     whether it is here in tcp.c or elsewhere in the library, is 
 *     herewith placed under (L)GPL for users of the GNU compiler kit, 
 *     and a BSD-style license for everyone else - In other words, do 
 *     what you wish with it, don't blame anyone if it breaks, and don't 
 *     pretend its yours. 
 *   - When in doubt, the (L)GPL applies since everything is based
 *     on Geof Cooper's work.
 *
 * *all* functions in the protocol layers (which this module is) are static.
 * entry points are exported with a dispatch table that is
 * registered from a constructor.
 *
 * The only globals made visible by pctcp.c are:
 * int __IPPROTO_TCP__ = IPPROTO_TCP;
 * int __SOCK_STREAM__ = 1; // same value as BSD's SOCK_STREAM
 * These may be used to force linkage.
 *
*/
#define __OSILAYER__ 4
#define WANT_STACK_CHECKING /* enable stack checking for this module */


#include "target.h"   /* standard types/includes */
#include "ipbase.h"   /* _ip_*() */
#include "inchksum.h" /* checksum() */
#include "debug.h"    /* KMSG(), DBGPRINTF() */
#include "icmp.h"     /* _icmp_*() */
#include "timer.h"    /* wintr_*() */
#include "blocker.h"  /* blocker_*() */
#include "loadcfg.h"  /* _conf_*() */
#include "inaddr.h"   /* is_broadcast() etc */
#include "sigblk.h"   /* set_cbreak_state() */
#include "printk.h"   /* printk() */
#include "utports.h"  /* lport_{reserve|release}() */
#include "utsio.h"    /* dispatch table */
#include "userland.h" /* WSO_* constants */

#define EXPORTED static
#undef IPPROTO_UDP /* undefining UDP sez "this is TCP" */

/* ==================================================================== */
/*              globals that may be referenced to force linkage         */
/* ==================================================================== */

/* globals that may be referenced to force linkage */
int __IPPROTO_TCP__ = IPPROTO_TCP;
int __SOCK_STREAM__ = 1; /* same value as BSD's SOCK_STREAM */

/* ==================================================================== */
/*                   TCP protocol control block definition              */
/* ==================================================================== */

typedef struct tcp_pcb /* Try to keep members on natural boundaries */
{                      /* (words on word-boundary, dwords on dword boundary) */
    uint32_t   pcb_marker_top;

    struct tcp_pcb FAR *next;   /* link to next tcp-socket */
    struct sockio_info FAR *sio_link; /* link to user space */

    uint8_t   xmit_timeout_s;   /* per-socket copy of ... */     
    uint8_t   recv_timeout_s;   /* ... global 'sock_delay'. */   
    uint8_t   ttl;              /* Time To Live  */              
    uint8_t   tos;              /* Terms of service (priority) */
    uint32_t  locflags;         /* local option flags */
                              
    uint32_t  myaddr;           /* my ip-address */                   
    uint32_t  hisaddr;          /* peer's internet address */         
    uint16_t   hisport;         /* peer's source port */              
    uint16_t   myport;          /* my source port */                  
            
    uint32_t  acknum;           /* data ACK'ed */
    uint32_t  seqnum;           /* sequence number */

/* Initial sequence number - stored in the pcb in HOST order (dont cnvt it!)*/
#define INIT_SEQ()  Random(0,ULONG_MAX)

    uint8_t   unhappy;          /* flag, indicates retransmitting segt's */
    uint8_t   recent;           /* 1 if recently transmitted */
    uint8_t   rigid;
    uint8_t   stress;

    uint8_t   state;            /* tcp_StateXXX (both UDP/TCP) */
    uint8_t   send_wscale;      /* to-do!!: window scales shifts, tx/rx */
    uint8_t   recv_wscale;            
#define TCP_MAX_WINSHIFT 14     /* maximum window shift */
    uint8_t   oobchar;          /* OOB buffer. Valid if WSO_LF_URG_AVAIL */
 
    uint8_t   flags;            /* tcp flags for last packet sent */
    uint8_t   cwindow;          /* Congestion window */
    uint8_t   wwindow;          /* Van Jacobson's algorithm */
    uint8_t   karn_count;         

                                /* Van Jacobson's Algorithm; */
#define MAX_VJSA    80000UL     /* max std. average */
#define MAX_VJSD    20000UL     /* max std. deviation */  
#define INIT_VJSA   220         /* initial std avg */  

    uint32_t  vj_sa;            /* VJ's alg, standard average */
    uint32_t  vj_sd;            /* VJ's alg, standard deviation */
    uint32_t  vj_last;          /* absolute last xmit time (in msecs) */

    uint32_t  last_rx_time_s;   /* abs time in secs that a pkt was *rcvd* */
                                /* used by retransmitter, keepalives, etc */

    uint32_t  missed_seg[2];    /* S. Lawson - handle one dropped segment */
    unsigned int   max_seg;     /* maximum segment size */
    unsigned int   window;      /* other guy's window */
    int    so_errno;            /* socket error code */          
    int    unacked;             /* unacked data, must be signed */
    unsigned int linger_delay_s;/* secs to wait in TIMEWAIT (2*MSL) */
#define TCPTV_MSL (30u)         /* max seg lifetime */
                                /* RFC 793 sez 2minutes, BSD uses 30s */

    struct                      /* retransmission strategy things */
    {
      unsigned int count;       /* number of times tried retransmission */
      uint32_t timeout_s;       /* time of next transmission */
      uint32_t intvl_ms;        /* retran interval (based on round trip time) */  
                                /* min = 1000ms, max = 64000ms */
#define TCPTV_RTO_INIT    5u    /* retran intvl while in SYNSENT state */
#define TCPTV_RTO_CLOSING 6u    /* retran intvl at >=CLOSEWT state */
#define TCPTV_RTO_DEFLT   3u    /* retran intvl if intvl_ms is 0 (unknown) */
#define TCPTV_RTO_BASE    1u    /* intvl_ms = vjsasd + TCPTV_RTO_BASE */
#define TCPREXMIT_MAXCNT  12    /* maximum number of rexmit tries
                                   without an answer before assuming
                                   peer died. */
     /* RFC-1122 and RFC-2988 state that the initial RTO should 
      * be (at least) ~3 secs. I chose a more conservative 5 seconds 
      * here (which is allowed by the standard). NetBSD uses 6 secs,
      * other BSDs use 3.
     */
    } rexmit;

    /* not all states have a retransmission strategy. The state
     * timeout stuff then prevents a socket getting "stuck" in a state.
    */
    struct /* state timeout management tracked in retransmitter */
    {
      uint16_t state;            /* state last seen */
      unsigned long time;        /* abs time in secs of entry into state */
    } state_to;

    struct                       /* keepalive handling */
    {
#define TCPTV_KEEP_INIT  (75u)   /* initial connect timeout (secs) */
#define TCPTV_KEEP_IDLE  (7200u) /* first keepalive probe: 2hrs (secs) */
#define TCPTV_KEEP_INTVL (75u)   /* secs between unanswered ka probes */
#define TCPTV_KEEP_CNT   (8u)    /* number of unanswered ka probes before
                                    assuming peer died. */

      unsigned long last_send;   /* absolute time of last probe send */
      unsigned int count;        /* number of probes sent */
    } keepalive;

    struct                       /* accept slots */
    {
      struct tcp_pcb FAR *list[5];
      unsigned int maxready;     /* max queue length */
    } bsd_accept;
    struct tcp_pcb FAR *parent;  /* each conn waiting accept has this set */

    struct /* Support for Timestamp option in tcphdr. We don't use tstamps*/
    {      /* ourselves, but we are obliged to echo them if the peer sends. */
      uint32_t  rcvd_num;        /* last (peer's) tstamp rcvd */
      uint32_t  sent_num;        /* last (self->peer) tstamp sentto */
      uint32_t  returned;        /* last (peer->self) tstamp sentback */
    } tsopt;

    struct
    {
      unsigned int usedlen;      /* tx buffer used len */
      uint8_t FAR *buf;          /* tx buffer ptr */
      unsigned int maxlen;       /* tx buffer size */
    } snd_buf;

    struct
    {
      signed int usedlen;        /* Rx length, must be signed */
      uint8_t FAR *buf;          /* Rx buffer ptr */
      unsigned int maxlen;       /* Rx buffer size */
#define MAX_WINDOW(__s__) ((__s__)->rcv_buf.maxlen)
    } rcv_buf;

    struct                       /* crap, backwards-compatible */
    {
      void (*yield_fn)(void);    /* yield while waiting */             
      uint32_t  usertimer_s;     /* ip_timer_init(), ip_timer_expired() */
      unsigned int   phase_timeout_s;    /* sock_timeout()/tcp_listen(,,,timeout) */
                                 /* See __tcp_Retransmitter() for note */ 
    } bc_idiocy;

    uint32_t pcb_marker_bottom;

} tcp_pcb;

/* ==================================================================== */
/*                        TCP specific constants                        */
/* ==================================================================== */

/* These are the TCP options */
#define TCPOPT_EOL       0       /* end-of-option list */
#define TCPOPT_NOP       1       /* no-operation */
#define TCPOPT_MAXSEG    2       /* maximum segment size */
#define TCPOPT_WINDOW    3       /* window scale factor (rfc1072) */
#define TCPOPT_SACKOK    4       /* selective ack ok (rfc1072) */
#define TCPOPT_SACK      5       /* selective ack (rfc1072) */
#define TCPOPT_ECHO      6       /* echo (rfc1072) */
#define TCPOPT_ECHOREPLY 7       /* echo (rfc1072) */
#define TCPOPT_TIMESTAMP 8       /* timestamps (rfc1323) */
#define TCPOPT_CC        11      /* T/TCP CC options (rfc1644) */
#define TCPOPT_CCNEW     12      /* T/TCP CC options (rfc1644) */
#define TCPOPT_CCECHO    13      /* T/TCP CC options (rfc1644) */


/* ----------------------------------------------------------------- */

#pragma pack(1)
typedef struct tcp_Header {
        uint16_t   srcPort;
        uint16_t   dstPort;
        uint32_t  seqnum;
        uint32_t  acknum;
#if defined(__BIG_ENDIAN__)
        unsigned char offset : 4,
                      unused : 4;
#elif defined(__TURBOC__)
        unsigned int  unused : 4,
                      offset : 4;
#else
        unsigned char unused : 4,
                      offset : 4;
#endif
        uint8_t   flags;

#define tcp_FlagFIN   0x01
#define tcp_FlagSYN   0x02
#define tcp_FlagRST   0x04
#define tcp_FlagPUSH  0x08
#define tcp_FlagACK   0x10
#define tcp_FlagURG   0x20
#define tcp_FlagMASK  0x3F

        uint16_t   window;
        uint16_t   checksum;
        uint16_t   urgent;
      } tcp_Header;
#pragma pack()

/* ======================================================================= */
/*                             Runtime configuration                       */
/* ======================================================================= */

static struct 
{
  /* TCP option config flags. */
  int tcp_opt_wscale;       /* not referenced at all */
  int tcp_opt_sackok;       /* not used ('#if 0' out) */

  int tcp_default_mss;      /* size (default = 0 = determine at connect time) */

  /* Misc TCP values */
  int tcp_nonagle;          /* bool (default = 0 = nagle enabled) */
  int tcp_keepalive_intvl;  /* seconds (default = 0 = TCPTV_KEEP_INTVL) */
  int tcp_keepalive_allon;  /* turn on/off keepalives for all pcbs */
  int tcp_rw_timeout;       /* "SOCKDELAY" opt. (0 = use builtin val) */

  int debug_lev;            /* default is zero */

  tcp_pcb FAR *_tcp_allsocs;    /* list of tcp-sockets */

} FAR my_statics;

#define tcp_default_mss     (my_statics.tcp_default_mss)
#define tcp_opt_wscale      (my_statics.tcp_opt_wscale)
#define tcp_opt_sackok      (my_statics.tcp_opt_sackok)
#define tcp_nonagle         (my_statics.tcp_nonagle)
#define tcp_keepalive_intvl (my_statics.tcp_keepalive_intvl)
#define tcp_keepalive_allon (my_statics.tcp_keepalive_allon)
#define tcp_rw_timeout      (my_statics.tcp_rw_timeout)
#define _tcp_allsocs        (my_statics._tcp_allsocs)
#define debug_lev           (my_statics.debug_lev)

static void __conf_set_nagle(const char *value)
{
  /* the option is "NAGLE", and the conf var is "nonagle" */
  tcp_nonagle = (atoi(value) ? 0 : 1);
}

static void __init_private_data(void)
{
  static const struct config_table FAR my_cfg[] = {
    { "MSS",                  ARG_ATOI,     (void FAR *)&tcp_default_mss   },
    { "NAGLE",                ARG_FUNC,     (void FAR *)__conf_set_nagle  },
    { "TCP.OPT.SACK",         ARG_ATOI,     (void FAR *)&tcp_opt_sackok    },
    { "TCP.OPT.WSCALE",       ARG_ATOI,     (void FAR *)&tcp_opt_wscale    },
    { "TCP.KEEPALIVE.INTVL",  ARG_ATOI,     (void FAR *)&tcp_keepalive_intvl },
    { "INACTIVE",             ARG_ATOI,     (void FAR *)&tcp_keepalive_allon },
    { "TCP.DEBUG",            ARG_ATOI,     (void FAR *)&debug_lev         },
    { "SOCKDELAY",  ARG_MULTI|ARG_ATOI,     (void FAR *)&tcp_rw_timeout    },
    { ((const char FAR *)0), 0,    ((void FAR *)0)  }
  };

  _conf_add_confextn( &my_cfg[0], NULL );

  FAR_MEMSET( &my_statics, 0, sizeof(my_statics));
  /* all vars have a default value of zero */

  return;
}

static tcp_pcb FAR *__validate_allsocs(struct sockio_info *__s); /* fwd ref */
static const struct wso_usrreq_tab FAR *__get_dispatch_tab(void); /* fwd ref */

/* ==================================================================== */
/*                         debug helpers                                */
/* ==================================================================== */

#if defined(NDEBUG)
#define __dbg_dump_header( __a, __b, __c, __d ) NO_OPERATION()
#else
static void __dbg_dump_header(const in_Header FAR *inpkt,
                              const tcp_Header FAR *hdr,
                              const void FAR *payload,
                              unsigned int payloadlen )
{
  char scratch[32];
  uint16_t pktid = 0;

  if (!inpkt) /* its output */
    DBGPRINTF(("*TCP OUT-xxxx: header dump\n" ));
  else
  {
    pktid = NTOHS(inpkt->identification);
    DBGPRINTF(("*TCP IN-%04x: header dump\n", pktid ));
  }
  DBGPRINTF(("*  -> srcPort   = %u\n", NTOHS(hdr->srcPort) ));
  DBGPRINTF(("*  -> dstPort   = %u\n", NTOHS(hdr->dstPort) ));
  DBGPRINTF(("*  -> seqnum    = 0x%08lx (%lu)\n", NTOHL(hdr->seqnum), NTOHL(hdr->seqnum) ));
  DBGPRINTF(("*  -> acknum    = 0x%08lx (%lu)\n", NTOHL(hdr->acknum), NTOHL(hdr->acknum) ));
  DBGPRINTF(("*  -> offset    = 0x%x (%u) (hdrlen=%u)\n", hdr->offset, hdr->offset,
                                         ((unsigned int)hdr->offset)<<2 ));
  scratch[0] = '\0';
  if (hdr->flags & tcp_FlagMASK)
  {
    strcpy(scratch, "=");
    if (hdr->flags & tcp_FlagACK)  strcat (scratch, " ACK");
    if (hdr->flags & tcp_FlagFIN)  strcat (scratch, " FIN");
    if (hdr->flags & tcp_FlagSYN)  strcat (scratch, " SYN");
    if (hdr->flags & tcp_FlagPUSH) strcat (scratch, " PSH");
    if (hdr->flags & tcp_FlagRST)  strcat (scratch, " RST");
    if (hdr->flags & tcp_FlagURG)  strcat (scratch, " URG");
  }
  DBGPRINTF(("*  -> flags     = 0x%02x (%u) %s\n", hdr->flags, hdr->flags, scratch ));
  DBGPRINTF(("*  -> window    = 0x%04x (%u)\n", NTOHS(hdr->window), NTOHS(hdr->window) ));
  DBGPRINTF(("*  -> checksum  = 0x%04x\n", NTOHS(hdr->checksum) ));
  DBGPRINTF(("*  -> urgptr    = 0x%04x (%u)\n", NTOHS(hdr->urgent), NTOHS(hdr->urgent) ));

  if (hdr->offset == (sizeof(*hdr)>>2)) /* hdr->offset == (20>>2)) */
    DBGPRINTF(("*  -> options   = (none)\n" )); 
  else
  {
    unsigned int optslen = (((unsigned int)hdr->offset)<<2) - sizeof(*hdr);
    const unsigned char FAR *start = ((const unsigned char FAR *)(hdr + 1));
    const unsigned char FAR *opt = start;

    DBGPRINTF(("*  -> options: (len = %u)\n", optslen ));

    while (opt < start+optslen)
    {
      unsigned int i, num;
      uint32_t val;
      switch (*opt)  /* Note: IP-option includes copy/class bits */
      {
        case TCPOPT_EOL:
             DBGPRINTF(("*     -> EOL\n" ));
             opt = start+optslen;
             break;
        case TCPOPT_NOP:
             DBGPRINTF(("*     -> NOP\n" ));
             opt++;
             break;
        case TCPOPT_MAXSEG:
             DBGPRINTF(("*     -> MSS %u\n", NTOHS(*(const uint16_t*)(opt+2)) ));
             opt += *(opt+1);
             break;
        case TCPOPT_WINDOW:
             DBGPRINTF(("*     -> Wscale %lu\n", (2L << *(opt+2)) ));
             opt += *(opt+1);
             break;
        case TCPOPT_SACK:
             num = (*(opt+1) - 2) / 4;
             DBGPRINTF(("*    -> SACK (count = %d)\n", num));
             for (i = 0; i < num; i++)
             {
               uint32_t left  = NTOHL(*(const uint32_t*)(opt+2+4*i));
               uint32_t right = NTOHL(*(const uint32_t*)(opt+6+4*i));
               DBGPRINTF(("*        -> blk %d: %lu/%lu\n", i+1, left, right));
             }
             opt += *(opt+1);
             break;
        case TCPOPT_SACKOK:
             DBGPRINTF(("*    -> SACK-OK\n"));
             opt += *(opt+1);
             break;
        case TCPOPT_ECHO:
             val = NTOHL(*(const uint32_t*)(opt+2));
             DBGPRINTF(("*     -> Echo %08lX\n", val));
             opt += *(opt+1);
             break;
        case TCPOPT_ECHOREPLY:
             val = NTOHL(*(const uint32_t*)(opt+2));
             DBGPRINTF(("*     -> Echoreply %08lX\n", val));
             opt += *(opt+1);
             break;
        case TCPOPT_TIMESTAMP:
             DBGPRINTF(("*     -> Timestamp %lu/%lu\n",
                     NTOHL(*(const uint32_t*)(opt+2)), 
                     NTOHL(*(const uint32_t*)(opt+6)) ));
             opt += *(opt+1);
             break;
        case TCPOPT_CC:
             val = NTOHL(*(const uint32_t*)(opt+2));
             DBGPRINTF(("*     -> CC %08lX\n", val));
             opt += *(opt+1);
             break;
        case TCPOPT_CCNEW:
             val = NTOHL(*(const uint32_t*)(opt+2));
             DBGPRINTF(("*     -> CCnew %08lX\n", val));
             opt += *(opt+1);
             break;
        case TCPOPT_CCECHO:
             val = NTOHL(*(const uint32_t*)(opt+2));
             DBGPRINTF(("*     -> CCecho %08lX\n", val));
             opt += *(opt+1);
             break;
        default:
             DBGPRINTF(("*     -> unknown opt 0x%x (%d), len=%d\n", 
                        *opt, *opt, opt[1] ));
             opt += *(opt+1);
             break;
      }
    }
  }

  if (inpkt)
  {
    unsigned int hdrlen = ((unsigned int)hdr->offset)<<2;
    if (NTOHS(inpkt->length) > (in_GetHdrLen(inpkt) + hdrlen))
    {
      payload = (const void FAR *)(((const uint8_t FAR *)hdr) + hdrlen);
      payloadlen = NTOHS(inpkt->length) - (in_GetHdrLen(inpkt) + hdrlen);
      if (debug_lev >= 3)
        DBGDUMP(("*TCP IN", pktid, payload, payloadlen ));
      else
        DBGPRINTF(("*  -> payloadlen= %u\n", payloadlen ));
    }
  }
  else
  {
    if (payloadlen && debug_lev >= 3)
      DBGDUMP(("*TCP OUT", 0, payload, payloadlen ));
    else
      DBGPRINTF(("*  -> payloadlen= %u\n", payloadlen ));
  }

  return;
}
#endif

/* --------------------------------------------------------------------- */

#if defined(NDEBUG)
#define __dbg_dump_state( __s ) NO_OPERATION()
#else
static void __dbg_dump_state( tcp_pcb FAR *sock )
{
  char scratch[32];
  uint32_t now;

  DBGPRINTF(("*TCP PCB state dump:\n"));
  DBGPRINTF(("*  -> state         = %-8.8s\n", sock_strstate(sock->state)));
  DBGPRINTF(("*  -> unhappy       = %s\n", (sock->unhappy ? "Yes" : "No") ));
  scratch[0] = '\0';
  if (sock->flags)
  {
    strcpy(scratch, "=");
    if (sock->flags & tcp_FlagACK)  strcat (scratch, " ACK");
    if (sock->flags & tcp_FlagFIN)  strcat (scratch, " FIN");
    if (sock->flags & tcp_FlagSYN)  strcat (scratch, " SYN");
    if (sock->flags & tcp_FlagPUSH) strcat (scratch, " PSH");
    if (sock->flags & tcp_FlagRST)  strcat (scratch, " RST");
    if (sock->flags & tcp_FlagURG)  strcat (scratch, " URG");
  }
  DBGPRINTF(("*  -> flags         = 0x%02x (%u) %s\n", 
                                    sock->flags, sock->flags, scratch ));
  DBGPRINTF(("*  -> missed_seg    = %lu:%lu\n", sock->missed_seg[0], 
                                                sock->missed_seg[1] ));
  DBGPRINTF(("*  -> karn_count    = %d\n", sock->karn_count ));
  DBGPRINTF(("*  -> vj_sa         = %lu\n", sock->vj_sa ));
  DBGPRINTF(("*  -> vj_sd         = %lu\n", sock->vj_sd ));
  DBGPRINTF(("*  -> cwindow       = %d\n", sock->cwindow ));
  DBGPRINTF(("*  -> wwindow       = %d\n", sock->wwindow ));
  DBGPRINTF(("*  -> rexmit.intvl  = %lu\n", sock->rexmit.intvl_ms ));

  now = timer_sec_count();
  if (sock->rexmit.timeout_s == 0UL)
  {
    DBGPRINTF(("*  -> rexmit.timeout_s = 0\n" ));
  }  
  else if (now <= sock->rexmit.timeout_s)
  {
    DBGPRINTF(("*  -> rexmit.timeout_s = %lu (remaining: %lus)\n", 
               sock->rexmit.timeout_s, (sock->rexmit.timeout_s - now)+1 )); 
  }
  else
  {
    DBGPRINTF(("*  -> rexmit.timeout_s = %lu (expired: %lus ago)\n", 
               sock->rexmit.timeout_s, (now - sock->rexmit.timeout_s) )); 
  }

  DBGPRINTF(("*  -> seqnum        = 0x%08lX (%lu)\n", 
                    sock->seqnum, sock->seqnum ));
  DBGPRINTF(("*  -> acknum        = 0x%08lX (%lu)\n", 
                    sock->acknum, sock->acknum ));
  DBGPRINTF(("*  -> unacked       = %d\n", sock->unacked ));
  DBGPRINTF(("*  -> rcv_buf.usedlen = %d\n", sock->rcv_buf.usedlen ));
  DBGPRINTF(("*  -> snd_buf.usedlen = %d\n", sock->snd_buf.usedlen ));

  return;
}
#endif

/* ==================================================================== */
/*                         TCP options                                  */
/* ==================================================================== */

/*
 * TCP option routines.
 * Note: Each of these MUST add multiple of 4 byte of options.
 *
 * Insert MSS option.
 */
static int tcp_opt_maxsegment (tcp_pcb FAR *s, uint8_t *opt, int *availlen)
{
  int len = 0;
  if (4 <= *availlen)
  {
    *opt++ = TCPOPT_MAXSEG;    /* option: MAXSEG,length,mss */
    *opt++ = 4;
    *(uint16_t*) opt = HTONS((uint16_t)s->max_seg);
    len = 4;
  }
  *availlen -= len;
  return (len);
}

/*
 * Insert return TimeStamp. We don't use timestamps ourselves, but if
 * the peer sends one, we are obliged to return it.
*/
static int tcp_opt_ret_tstamp(tcp_pcb FAR *s, uint8_t *opt, int *availlen)
{
  int len = 0;

  /* We got a TS in a previous SYN-ACK or SYN. Send it back (unless RST)*/
  if ((s->flags & tcp_FlagRST) == 0 && (s->locflags & WSO_LF_RCVD_TSTMP))
  {
    s->locflags &= ~WSO_LF_RCVD_TSTMP;  /* don't do this again */

    if (12 <= *availlen)
    {
      s->tsopt.sent_num = timer_msec_count();
      *opt++ = TCPOPT_NOP;     /* NOP,NOP,TIMESTAMP,length,TSval,TSecho */
      *opt++ = TCPOPT_NOP;
      *opt++ = TCPOPT_TIMESTAMP;   
      *opt++ = 10;
      *((uint32_t *)opt) = HTONL(s->tsopt.sent_num); opt += sizeof(uint32_t);
      *((uint32_t *)opt) = HTONL(s->tsopt.rcvd_num); opt += sizeof(uint32_t);
      len = 12;
    }
    *availlen -= len;
  }
  return len;
}

#ifdef NOT_USED
static int tcp_opt_winscale (tcp_pcb FAR *s, uint8_t *opt, int *availlen)
{
  int len = 0;
  if (4 <= *availlen)
  {
    *opt++ = TCPOPT_WINDOW;    /* option: WINDOW,length,wscale */
    *opt++ = 4;
    *(uint16_t*) opt = HTONS(s->send_wscale);
    len = 4;
  }
  *availlen -= len;
  return (len);
}
#endif

#ifdef NOT_USED
static int tcp_opt_sack_ok (tcp_pcb FAR *s, uint8_t *opt, int *availlen)
{
  int len = 0;
  if (4 <= *availlen)
  {
    *opt++ = TCPOPT_SACKOK;
    *opt++ = 2;
    *opt++ = TCPOPT_NOP;
    *opt++ = TCPOPT_NOP;
    len = 4;
  }
  *availlen -= len;
  return (len);
}
#endif

#ifdef NOT_USED
static int tcp_opt_sack (tcp_pcb FAR *s, uint8_t *opt, int *availlen,
                         struct SACK_list *sack)
{
  int len = 0;
  if (sack->num_blk && (2 + (8 * sack->num_blk) + 2) <= *availlen )
  {
    int i;
    len = 2 + (8 * sack->num_blk);

    *opt++ = TCPOPT_SACK;       /* option: SACK,length,left,right,.. */
    *opt++ = len;
    for (i = 0; i < sack->num_blk; i++)
    {
      *((uint32_t *)opt) = HTONL(sack->list[i].left_edge);
      opt += sizeof(uint32_t);
      *((uint32_t *)opt) = HTONL(sack->list[i].right_edge);
      opt += sizeof(uint32_t);
    }
    *opt++ = TCPOPT_NOP;
    *opt++ = TCPOPT_NOP;
    len += 2;
  }
  *availlen -= len;
  return (len);
}
#endif

/*
 * tcp_do_options - Add TCP options to output segment
 */
static int tcp_do_options (tcp_pcb FAR *s, uint8_t *opt, int is_syn)
{
  int   availlen = ((0x0fu << 4u) - sizeof(tcp_Header))-4;
  int   len = 0;

  if (is_syn)
  {
    len += tcp_opt_maxsegment (s, opt+len, &availlen);

#if 0
    if (s->locflags & WSO_LF_REQ_SCALE)
       len += tcp_opt_winscale (s, opt+len, &availlen);

    if (tcp_opt_sackok)
       len += tcp_opt_sack_ok (s, opt+len, &availlen);
#endif
  }
  else if (!is_syn)
  {
    /* If we got a TS in a previous SYN-ACK or SYN we need to return it */
    len += tcp_opt_ret_tstamp(s, opt+len, &availlen );
  }

  /* unlike IP, the option list does not need to end with _EOL,
   * but the length does need to be an even multiple of four.
  */ 
  if (len & 3)
  {
    while ((len & 3) != 3)
      opt[len++] = TCPOPT_NOP;
    opt[len++] = TCPOPT_EOL;
  }

  return (len);
}

/* ==================================================================== */
/*                         raw tcp sender                               */
/* ==================================================================== */

static void SET_REXMIT_TIMEOUT(tcp_pcb FAR *s,uint32_t ms,int incr_only,int line)
{
  unsigned int secs = 1;
  if (ms > 64000ul)
    secs = 64;
  else if (((unsigned int)ms) > 1000u)
    secs = ((((unsigned int)ms) + 499u)/1000u);
  ms = timer_sec_count() + secs;
  if (!incr_only || ms > s->rexmit.timeout_s)
  {
    s->rexmit.timeout_s = ms;
    if (debug_lev >= 3)
      DBGPRINTF(("new rexmit.timeout_s = %u (expires @ %lu) from line %d\n", 
                          secs, s->rexmit.timeout_s, line));
  }
  line = line; /* possibly unreferenced - shaddup compiler */
}


#if defined(NDEBUG)
#define TCP_SEND(s)     _tcp_send    (s, "", 0)
#define TCP_SENDSOON(s) _tcp_sendsoon(s, "", 0)
#elif defined(__FUNCTION__)
#define TCP_SEND(s)     _tcp_send    (s, __FUNCTION__, __LINE__)
#define TCP_SENDSOON(s) _tcp_sendsoon(s, __FUNCTION__, __LINE__)
#else
#define TCP_SEND(s)     _tcp_send    (s, __FILE__, __LINE__)
#define TCP_SENDSOON(s) _tcp_sendsoon(s, __FILE__, __LINE__)
#endif

/*
 * _tcp_send - Format and send an outgoing segment.
 *             Several packets may be sent depending on peer's window.
 * Returns 0 on success, -1 on error.
*/
static int _tcp_sendsoon (tcp_pcb FAR *s, char *file, unsigned line);

static int _tcp_send (tcp_pcb FAR *s, char *file, unsigned line)
{
  int total_sent_len = 0; /* count of data length we've sent */
  int datastart_offs;     /* where data starts in tx-buffer (must be signed) */
  int total_tosend_len;   /* total amount of data to send (must be signed) */
  int pkt_num;            /* 0..s->cwindow-1 */

  blocker_critsect_begin();

  if (s->state == tcp_StateSYNSENT && 
      s->rexmit.intvl_ms == 0) /* first time send */
  {
    tcp_pcb FAR *ds = _tcp_allsocs;
    while (ds) 
    {
      if (ds != s && ds->rexmit.intvl_ms && ds->hisaddr == s->hisaddr)
      {
        s->rexmit.intvl_ms = ds->rexmit.intvl_ms;
        break;
      }
      ds = ds->next;
    }
    if (s->rexmit.intvl_ms == 0)
      s->rexmit.intvl_ms = (TCPTV_RTO_INIT * 1000u);
  }

  s->recent = 0;
  datastart_offs = 0;
  total_sent_len = 0;

  total_tosend_len = s->snd_buf.usedlen;
  if (total_tosend_len > ((int)s->window))
    total_tosend_len = s->window;

  if (s->karn_count != 2) /* not a retransmission */
  {
    /* Morten Terstrup <MorTer@dk-online.dk> found this signed bug
     */
    if (total_tosend_len < s->unacked)
      total_tosend_len = s->unacked;
    total_tosend_len -= s->unacked;
    datastart_offs = s->unacked;
  }

  /* set s->max_seg if unknown, or adjust in case mtu has decreased */
  /* if IP/ether haven't been initialized yet, that will happen now */
  {
    unsigned int ip_mtu = _ip_get_mtu(); /* cause config load and IP boot */
    if (s->max_seg == 0 && tcp_default_mss > 0)
      s->max_seg = tcp_default_mss;
    if (s->max_seg == 0 || s->max_seg > (ip_mtu-(20u+20u))) 
      s->max_seg = (ip_mtu-(20u+20u));
  }

  /* step through our packets
   */
  for (pkt_num = 0; pkt_num < s->cwindow; pkt_num++)
  {
    tcp_PseudoHeader ph;
    const uint8_t FAR *senddataptr = (const uint8_t FAR *)0;
    int senddatalen;     /* how much (user) data in this segment */
    uint8_t hdr_buf[sizeof(tcp_Header)+64/*options*/];
    tcp_Header *hdr = (tcp_Header *)(&hdr_buf[0]);
    int hdr_len = sizeof(*hdr);         

    /* Insert any TCP options after header
     */
    if (pkt_num == 0 && (s->flags & (tcp_FlagSYN|tcp_FlagACK)) == tcp_FlagSYN)
    {
      hdr_len = sizeof(*hdr) + tcp_do_options(s, &hdr_buf[sizeof(*hdr)], 1);
      senddatalen = 0;
    }
    else
    {
      hdr_len = sizeof(*hdr) + tcp_do_options(s, &hdr_buf[sizeof(*hdr)], 0);
      senddatalen = 0;

      if (s->max_seg > (hdr_len - sizeof(*hdr)))
      {
        int data_free = s->max_seg - (hdr_len - sizeof(*hdr));

        senddatalen = total_tosend_len;
        if (senddatalen > data_free)
          senddatalen = data_free;
        if (senddatalen > 0)
        {
          #if 0
          if (s->queuelen)
            senddataptr = s->queue+datastart_offs;
          else
          #endif
            senddataptr = s->snd_buf.buf+datastart_offs;
        }
      }
    }

    /* make tcp header
     */
    hdr->srcPort  = HTONS(s->myport);
    hdr->dstPort  = HTONS(s->hisport);
                    /* seqnum = unacked. no longer total_sent_len */
    hdr->seqnum   = HTONL(s->seqnum + datastart_offs); 
    hdr->acknum   = HTONL(s->acknum);

    hdr->window   = HTONS((uint16_t)(s->rcv_buf.maxlen - s->rcv_buf.usedlen));
    hdr->flags    = s->flags;
    hdr->unused   = 0;
    hdr->checksum = 0;
    hdr->urgent   = 0;
    hdr->offset   = (uint8_t)(((unsigned int)(hdr_len))>>2); /* <= 15 */

    if (s->locflags & WSO_LF_NOPUSH) 
      hdr->flags &= ~tcp_FlagPUSH;

    /* make tcp header checksum
     */
    ph.src      = HTONL(s->myaddr ? s->myaddr : _ip_get_ipaddr());
    ph.dst      = HTONL(s->hisaddr);
    ph.mbz      = 0;
    ph.protocol = IPPROTO_TCP;
    ph.length   = HTONS((uint16_t)(hdr_len + senddatalen));
    ph.checksum = checksum2(hdr, hdr_len, senddataptr, senddatalen );
    hdr->checksum = ~checksum (&ph, sizeof(ph));

    if (debug_lev)
    {
      DBGPRINTF(("hdr_len = %u, senddatalen = %u\n", hdr_len, senddatalen ));
    }

    if (_ip_output_2(NTOHL(ph.src), s->hisaddr, IPPROTO_TCP, s->ttl, s->tos, 
              0, hdr, hdr_len, senddataptr, senddatalen, &s->so_errno ) == -1)
    {
      TCP_SENDSOON (s);
      return -1;
    }

    if (debug_lev >= 2)
    {
      __dbg_dump_header(NULL, hdr, senddataptr, senddatalen);
      __dbg_dump_state( s );
    }

    if (debug_lev)
      DBGPRINTF(("*-TCP OUT-xxxx: done. sent %u\n", hdr_len+senddatalen));

    /* do next packet
     */
    if (senddatalen > 0)
    {
      total_sent_len  += senddatalen;
      datastart_offs   += senddatalen;
      total_tosend_len -= senddatalen;
    }
    if (total_tosend_len <= 0)
       break;
  }

  s->unacked = datastart_offs;

  if (debug_lev)
  {
    DBGPRINTF(("*_tcp_send: Sent total %u data (win %u, cwin %u)\n",
                total_sent_len, s->window, s->cwindow ));
    DBGPRINTF(("*_tcp_send: unacked = %u, SEQ = 0x%lx (%lu), (caller %s:%u)\n",
                s->unacked, s->seqnum, s->seqnum, file, line )); 
  }

  if (s->karn_count == 2) /* that was a retransmission */
  {  
    if (s->rexmit.intvl_ms)
      s->rexmit.intvl_ms = /* increase by 50% */
        (s->rexmit.intvl_ms + s->rexmit.intvl_ms + s->rexmit.intvl_ms) >> 1;
    else                   /* should not happen */
      s->rexmit.intvl_ms = TCPTV_RTO_DEFLT; 
    s->vj_last = 0;
  }
  else
  {
    s->vj_last = 0;
    if (s->unhappy || s->snd_buf.usedlen)
      s->vj_last = timer_msec_count(); /* expect an immediate response */
    s->karn_count = 0;
  }
  
  SET_REXMIT_TIMEOUT(s, s->rexmit.intvl_ms, 0, __LINE__);

  blocker_critsect_end();

  file = file; /* possibly unreferenced - shaddup compiler */
  line = line; /* possibly unreferenced - shaddup compiler */

  return (0); /* success */
}


/*
 * _tcp_sendsoon - schedule a transmission pretty soon.
 *  - this one has an imperfection at midnight, but it
 *    is not significant to the connection performance.
 *
 *    gv: Added - 5 May 2000: Relax retransmission period to
 *        TCPTV_RTO_CLOSING when CLOSEWT state is entered.
 *        Relax retransmission period to TCPTV_RTO_INIT in
 *        SYNSENT state.
 *
 * Returns 0 on success, -1 on error.
 */
static int _tcp_sendsoon (tcp_pcb FAR *s, char *file, unsigned line)
{
  uint32_t timeout;

  if (s->state >= tcp_StateCLOSWT)
    timeout = timer_sec_count() + TCPTV_RTO_CLOSING;
  else 
    timeout = timer_sec_count() + TCPTV_RTO_BASE;

  if (s->rexmit.intvl_ms <= (TCPTV_RTO_BASE * 1000u) && s->recent == 0)
  {                         
    if (s->rexmit.timeout_s < timeout)
    {
      int rc;
      s->karn_count = 0;
      rc = _tcp_send (s, file, line);
      s->recent = 1;
      return (rc);
    }
  } 

  if ((s->unhappy || s->snd_buf.usedlen > 0 || s->karn_count == 1) &&
      (s->rexmit.timeout_s && s->rexmit.timeout_s < timeout))
     return (0);

  if (s->state == tcp_StateSYNSENT)  /* relaxed in SYNSENT state */
    timeout = (TCPTV_RTO_INIT * 1000u);
  else if (s->state >= tcp_StateCLOSWT)
    timeout = (TCPTV_RTO_CLOSING * 1000u);
  else
    timeout = (TCPTV_RTO_BASE * 1000u) + (s->rexmit.intvl_ms >> 4);
  SET_REXMIT_TIMEOUT(s, timeout, 0, __LINE__);

  s->karn_count = 1;

  return (0);
}

/* ==================================================================== */
/*                          send reset                                  */
/* ==================================================================== */

#if defined(__FUNCTION__)
#define TCP_SEND_RESET(s, ip, tcp) \
       _tcp_send_reset(s, ip, tcp, __FUNCTION__, __LINE__)
#else
#define TCP_SEND_RESET(s, ip, tcp) \
       _tcp_send_reset(s, ip, tcp, __FILE__, __LINE__)
#endif

/*
 * Format and send a reset tcp packet
 */
static int _tcp_send_reset (tcp_pcb FAR *s, const in_Header FAR *his_ip, 
                            const tcp_Header FAR *old_tcp, 
                            const char *file, unsigned line)
{
  tcp_PseudoHeader ph;
  uint8_t   oldflags;
  tcp_Header hdr;
  uint32_t  seqnum;
  uint32_t  acknum;
  int rc;

  if (old_tcp->flags & tcp_FlagRST)
     return (-1);

  /* see RFC 793 page 65 for details
  */
  {
    static uint32_t last_RST_time = 0L;
    uint32_t now = timer_tick_count();
    if ((now >> 1) == (last_RST_time >> 1))
      return (-1);
    last_RST_time = now;
  }

  oldflags = old_tcp->flags;
  seqnum = old_tcp->seqnum;
  acknum = old_tcp->acknum;

  if (oldflags & tcp_FlagACK)                        /* ACK */
  {
    seqnum = acknum;
    acknum = 0;
    oldflags = tcp_FlagRST;
  }
  else if ((oldflags & (tcp_FlagSYN|tcp_FlagACK)) == tcp_FlagSYN)  
  {                                              /* SYN, no ACK */
    acknum = HTONL(NTOHL(seqnum) + 1);
    seqnum = 0;
    oldflags = tcp_FlagACK | tcp_FlagRST;
  }
  else
  {
    int temp = NTOHS(his_ip->length) - in_GetHdrLen (his_ip);
    acknum = HTONL(NTOHL(seqnum) + temp);
    seqnum = 0;
    oldflags = tcp_FlagRST;
  }

  if (s && (s->locflags & WSO_LF_NOPUSH))
     oldflags &= ~tcp_FlagPUSH;

  /* tcp header
   */
  hdr.srcPort  = old_tcp->dstPort;
  hdr.dstPort  = old_tcp->srcPort;
  hdr.seqnum   = seqnum;
  hdr.acknum   = acknum;
  hdr.window   = 0;
  hdr.flags    = (uint8_t)(oldflags | tcp_FlagRST);
  hdr.unused   = 0;
  hdr.offset   = sizeof(hdr) / 4;
  hdr.checksum = 0;
  hdr.urgent   = 0;

  ph.src      = his_ip->destination;   /* already network order */
  ph.dst      = his_ip->source;
  ph.mbz      = 0;
  ph.protocol = IPPROTO_TCP;
  ph.length   = HTONS(sizeof(hdr));
  ph.checksum = checksum (&hdr, sizeof(hdr));

  hdr.checksum = ~checksum (&ph, sizeof(ph));

  blocker_critsect_begin();

  DBGPRINTF(("*TCP-xxxx: sending tcp reset to %I...\n", ph.dst));
  DBGPRINTF(("*          (called from %s:%u).\n", file, line ));

  rc = _ip_output_2(NTOHL(ph.src), NTOHL(ph.dst), IPPROTO_TCP, 
                  (uint8_t)(s ? s->ttl : _ip_get_default_ttl()),
                  his_ip->tos, 0, &hdr, sizeof(hdr), NULL, 0, &s->so_errno );

  if (rc == -1)
  {
    if (debug_lev)
      DBGPRINTF(("*-TCP-xxxx attempt to RST failed\n"));
    rc = -1;
  }
  else
  {
    if (debug_lev >= 2)
      __dbg_dump_header(NULL, &hdr, NULL, 0);
    if (debug_lev)
      DBGPRINTF(("*-TCP-xxxx RST sent.\n" ));
    rc = 0;
  }

  blocker_critsect_end();

  file = file; /* possibly unreferenced - shaddup compiler */
  line = line; /* possibly unreferenced - shaddup compiler */

  return rc;
}

/* ==================================================================== */
/*                         sock "close"/"abort"                         */
/* ==================================================================== */

/* 
** Stop a TCP connection. 'stoptype' is an both an error number, 
** as well as an action code:
**  - ECONNRESET: the peer (or proxy) sent a RST
**  - ETIMEDOUT:  lack of response from the peer (keepalive, conn timeout etc) 
**  - ECONNABORTED: user/application called sock_abort()
** If 'stoptype' is not ECONNRESET, __tcp_stop_conn() will send a RST.
*/
static void __tcp_stop_conn(tcp_pcb FAR *s, int stoptype, int line)
{
  if (debug_lev)
    DBGPRINTF(("STOP conn (peer %hI:%u), stoptype=%s, from line=%d\n", 
                s->hisaddr, s->hisport, sock_strerror(stoptype), line ));

  s->snd_buf.usedlen = 0;  /* Flush Tx buffer */

  #if 0 /* why should we do this? */
  if (s->state != tcp_StateCLOSED && s->state != tcp_StateLASTACK)
  {
    #if 0
    if (debug_lev >= 2)
      DBGPRINTF(("zapping s->rcv_buf.usedlen to 0 from %d\n", s->rcv_buf.usedlen ));
    #endif
    s->rcv_buf.usedlen = 0;
  }
  #endif

  if (stoptype == WSO_ECONNRESET)
  {
    /* a reset received while still connecting is a CONNREFUSED error */
    if (s->state < tcp_StateESTAB) /* In the process of connecting */
      stoptype = WSO_ECONNREFUSED;
    /* nothing else to do */
  }
  else /* ECONNABORT, ETIMEDOUT */
  {
    if (s->state != tcp_StateCLOSED)
    {                                          
      s->flags   = (tcp_FlagRST|tcp_FlagACK);
      s->unhappy = 1;
      if (s->state <= tcp_StateSYNREC)
      {
        s->rexmit.timeout_s = 0UL;
      }
      TCP_SEND (s);
    }
  }

  s->state   = tcp_StateCLOSED;
  s->locflags |= WSO_LF_ASYNC_ERROR;
  s->so_errno = stoptype; /* caller can be more specific after we return */

  line = line; /* possibly unreferenced - shaddup compiler */
  return;
}

/* ==================================================================== */

static void __free_pcb(tcp_pcb FAR *s, int was_StateCLOSED) 
{
  int was_child = (s->sio_link == NULL);  
  tcp_pcb FAR *prev = (tcp_pcb FAR *)0;
  tcp_pcb FAR *ds = _tcp_allsocs;

  /* unthread FIRST (otherwise localport_release will find it again) */
  while (ds)
  {
    if (ds == s)
    {
      if (!prev)
        _tcp_allsocs = s->next;
      else
        prev->next = s->next;
      break;
    }
    prev = ds;
    ds = ds->next;
  }

  if (s->myport != 0 && !was_child) 
    localport_release(s->myport, IPPROTO_TCP, 
                      (was_StateCLOSED ? 0 : s->linger_delay_s) );
  s->myport = 0;

  /* free the allocated recv buffer */
  FAR_FREE(s->rcv_buf.buf);
  s->rcv_buf.buf = (uint8_t FAR *)0;
  s->rcv_buf.maxlen = 0;
  s->rcv_buf.usedlen = 0;

  /* free the allocated send buffer */
  FAR_FREE(s->snd_buf.buf);
  s->snd_buf.buf = (uint8_t FAR *)0;
  s->snd_buf.maxlen = 0;
  s->snd_buf.usedlen = 0;

  /* unlink from user space */
  if (s->sio_link)
    FAR_MEMSET( s->sio_link, 0, sizeof(*(s->sio_link)));

  /* free the pcb itself */
  FAR_MEMSET( s, 0, sizeof(*s));
  FAR_FREE(s);

  return;
}

/*
** Really abort a connection. 
** 1) If the socket is pending tcp_accept(), then convert it to a normal
**    socket, recurse to do step 3/4, then unthread and free() since the
**    user will never be able to do a tcp_destroy() on this socket.
** 2) If the socket is listening then recurse for all child sockets
**    that are still pending tcp_accept().
**    Then go straight to step 4.
** 3) If the socket is neither 1) or 2) then send a RST. 
**    Goto step 4.
** 4) set state = tcp_StateCLOSED. Don't release any resources allocated
**    for the socket - tcp_destroy() [which the user must call] will do
**    that.
** 
** Unlike a shutdown socket, an aborted socket is DEAD - it can't send, 
** it can't receive, and is just sitting around waiting to be tcp_destroy()d. 
** It may as well be unthreaded, but isn't because tcp_destroy may need to 
** free resources still allocated for the socket, and for safety, tcp_destroy 
** will refuse to do anything for a socket that isn't on the soclist.
*/
static void __force_abort(tcp_pcb FAR *s)
{
  unsigned int i, n;

  if (debug_lev >= 2)
    DBGPRINTF(("*__force_abort() called for PCB %p\n", s));

  if (s->parent != NULL) /* its pending tcp_accept */
  {
    /* sockets with s->parent != NULL will only be shutdown by internal fxns - 
    ** They are sockets pending tcp_accept() and aren't visible to the user 
    ** yet. No data has been sent or received yet either.
    */
    n = s->parent->bsd_accept.maxready;
    for (i = 0; i < n; i++)
    {
      /* the whole table needs to be searched since there may already 
      ** be gaps when the parent itself is about to be destroyed.
      */
      if (s->parent->bsd_accept.list[i] == s)
      {
        for (--n; i < n; i++)
          s->parent->bsd_accept.list[i] = s->parent->bsd_accept.list[i+1];
        s->parent->bsd_accept.list[n] = NULL;
        break;
      }
    }
    n = s->state;
    /* clear parent (or we'll go into an infinite loop) */
    s->parent = (tcp_pcb FAR *)0; 
    __force_abort(s); /* recurse */
    /* Have to destroy now since the user will never be able to do so. */
    __free_pcb(s, 1 );
    s = NULL; /* 's' is no longer valid */
    return; 
  }

  if (s->state == tcp_StateLISTEN)
  {
    /* listening skts can go straight to tcp_StateCLOSED since
    ** no data has been sent/received yet, but they still need to have 
    ** their children (skts pending tcp_accept()) aborted.
    */
    n = s->bsd_accept.maxready; /* will be zero for non-BSD sox */
    s->bsd_accept.maxready = 0;
    s->state = tcp_StateCLOSED;
    for (i = 0; i < n; i++)
    {
      if (!s->bsd_accept.list[i]) /* done all */
        break;
      __force_abort(s->bsd_accept.list[i]); /* recurse */
    }
  }

  if (s->state != tcp_StateCLOSED)
  {
    __tcp_stop_conn(s, WSO_ECONNABORTED, __LINE__);
  }

  s->state    = tcp_StateCLOSED;
  s->unhappy  = 0;
  s->snd_buf.usedlen  = 0;
  #if 0
  if (debug_lev >= 2)
    DBGPRINTF(("zapping s->rcv_buf.usedlen to 0 from %d\n", s->rcv_buf.usedlen ));
  #endif
  s->rcv_buf.usedlen = 0;
  s->missed_seg[0] = 0;

  return;
}

/*
 *  Shutdown a particular socket/connection. 
 *  1) If its a socket pending BSD tcp_accept() [s->parent != NULL], then
 *     simply abort the connection since the user hasn't seen this socket yet.
 *  2) If its a socket thats listening, then call __force_abort() for
 *     this socket too to destroy all the sockets pending tcp_accept().
 *  3) If its a socket thats connected, and 'how' is > 0, send a FIN.
 *     Must still allow receives if 'how' == 0
 */
static int __pcb_shutdown(tcp_pcb FAR *s, int how)
{
  if (s->parent != NULL) /* its pending tcp_accept - abort it */
  {
    /* sockets with s->parent != NULL will only be shutdown by internal fxns - 
    ** They are sockets pending tcp_accept() and aren't visible to the user 
    ** yet. Since no data has been sent or received yet, simply abort.
    */
    __force_abort(s);
    return 0; /* 's' has been free()d - don't fall through */
  }

  switch (s->state)
  {
    case tcp_StateLISTEN:
    {
      if (how == 1) /* send shutdown ONLY - no such thing for listening sox */
        break;      /* simply fall through */

      /* no data has been sent or received yet - but child 
      ** sockets pending accept still need to be aborted.
      */
      __force_abort(s);
      break;
    }
    case tcp_StateSYNSENT:   /* unlink failed connection */
    {
      s->state = tcp_StateCLOSED;
      how = 2;
      break;
    }
    case tcp_StateSYNREC:
    case tcp_StateESTAB:
    case tcp_StateESTCL:
    {
      if (how == 0)
        break; /* nothing - fallthrough */
        
      if (s->snd_buf.usedlen)      /* must first flush all Tx data */
      {
        s->flags |= (tcp_FlagPUSH | tcp_FlagACK);
        if (s->state < tcp_StateESTCL)
        {
          s->state = tcp_StateESTCL;
          TCP_SENDSOON (s);
        }
        s->unhappy = 1;
        break;
      }

      /* really closing */
      s->flags = (tcp_FlagACK | tcp_FlagFIN);
      /* if (!s->err_msx) s->err_msx = "Connection closed normally";  */
      s->state    = tcp_StateFINWT1;
      s->rexmit.timeout_s = 0UL;                    /* stop RTT timer */
      TCP_SEND (s);
      s->unhappy = 1;
      break;
    }
    case tcp_StateCLOSWT:
    {
      /* need to ACK the FIN and get on with it */
      s->state  = tcp_StateLASTACK;
      s->flags |= tcp_FlagFIN;
      TCP_SEND (s);
      s->unhappy = 1;
      how = 2;
      break;
    }
    case tcp_StateTIMEWT: /* called for timeout on tcp_StateLASTACK */
    {
      s->state = tcp_StateCLOSED;
      how = 2;
      break;
    }
  } /* switch() */

  if (how == 0 || how == 2)
  {
    #if 0
    if (debug_lev >= 2)
      DBGPRINTF(("zapping s->rcv_buf.usedlen to 0 from %d\n", s->rcv_buf.usedlen ));
    #endif
    s->rcv_buf.usedlen = 0;
    s->missed_seg[0] = 0;
    s->locflags |= WSO_LF_CANTRECVMORE;
  }
  if (how == 1 || how == 2)
  {
    s->locflags |= WSO_LF_CANTSENDMORE;
  }

  return 0;  
}

EXPORTED int tcp_shutdown(struct sockio_info *__s, int how)
{
  tcp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
    return __pcb_shutdown(s, how);
  return -1;
}

/* ----------------------------------------------------------------------- */

/*
 * Abort a tcp connection and unthread the socket from the list of sockets.
 * THIS FUNCTION MAY ONLY BE CALLED BY THE USER (and by tcp_create() which
 * needs to do so since the socket * may be reused without having previously
 * destroying the socket).
 *
 * 1) If 's' is a listening socket, first destroy all its 'children' that may
 *    be pending BSD tcp_accept() by calling tcp_shutdown() on the socket.
 *    Then fallthrough and unthread.
 * 2) If its a socket that is pending BSD tcp_accept() then panic.
 *    See comments in code as to why.
 * 3) If its a socket that has already been closed, fall through and unthread.
 * 4) If its not any of the above then it must be in a connected state, 
 *    send a RST, fallthrough and unthread the socket.
*/

/* tcp_destroy is called internally by tcp_create() but otherwise
** may not be called by any other internal function.
*/
static int __pcb_destroy(tcp_pcb FAR *s)
{
  if (s->parent != NULL) 
  {
    /* sockets with s->parent != NULL are sockets pending tcp_accept()
    ** and aren't visible to the user yet. And since tcp_destroy()
    ** can only be called by the user, something must be horribly wrong.
    */
    __fatal_runtime_error("PANIC: tcp_destroy() called for an unaccepted socket\n");
  }
  else
  {
    int was_StateCLOSED;

    __pcb_shutdown( s, 2);
    was_StateCLOSED = (s->state == tcp_StateCLOSED);

    /* force "abort" irrespective of state. 
    ** (listening skts need to have their children aborted) 
    */
    /* -- need some logic to deal with TIMEWT here -- */
    __force_abort(s);

    /* physically destroy the PCB */
    __free_pcb(s, was_StateCLOSED);
  }
  return 0;
}


EXPORTED int tcp_destroy (struct sockio_info *__s)
{
  tcp_pcb FAR *s = __validate_allsocs(__s);
  if (s)
    return __pcb_destroy(s);
  return -1;
}

static void __stack_shutdown(void) 
{                            /* called when the lib is being deinitialized */
  tcp_pcb FAR *s = _tcp_allsocs;
  if (debug_lev)
    DBGPRINTF(("*TCP shutdown begin\n"));
  while (s)
  {
    tcp_pcb FAR *ds = s->next;
    if (!s->parent)   /* its not pending tcp_accept() */
    {                 /* those are destroyed when the listener is destroyed */
      s->linger_delay_s = 0;
      __pcb_destroy(s); 
    }
    s = ds;
  }
  if (debug_lev)
    DBGPRINTF(("*TCP shutdown complete\n"));
  return;
}

/* ==================================================================== */
/*                        periodic stuff                                */
/* ==================================================================== */

/*
** tcp_Retransmitter
*/  
static void __tcp_Retransmitter(tcp_pcb FAR *s)
{
  if (s->state == tcp_StateCLOSED)
  {
    /* nothing */
  }
  else if (s->rexmit.timeout_s == 0UL)
  {
    /* To ensure that a socket doesn't get "stuck" in a state that
     * does not have a retransmission strategy (s->rexmit.timeout_s is zero),
     * this next clause tracks what state was seen last, shutting down the
     * connection when the state doesn't change within the allotted time.
     *
     * s->phase_timeout_s is a user-definable value and has a default of zero.
     * The sock io documentation for tcp_listen() states:
     *    timeout - (TCP only) if timeout is non-zero, the socket will be
     *    shut down if a connection is not received in that many seconds.
     * In old versions of the library, s->phase_timeout_s was called 
     * 's->timeout', was overwritten by state change, and was settable from
     * user space with sock_timeout().
    */ 
    if (s->state_to.state != s->state)
    {
      s->state_to.state = s->state;
      s->state_to.time = timer_sec_count();
    } 
    else
    {
      /* ** all timeouts are in seconds ** */
      static unsigned char FAR to_table[11] = 
      {
        /* tcp_StateLISTEN:  */  0, /* see override below */
        /* tcp_StateSYNSENT: */ 15,
        /* tcp_StateSYNREC:  */ 15,
        /* tcp_StateESTAB:   */  0, /* see override below */
        /* tcp_StateESTCL:   */  0,
        /* tcp_StateFINWT1:  */ 15,
        /* tcp_StateFINWT2:  */ 15,
        /* tcp_StateCLOSWT:  */  7,
        /* tcp_StateCLOSING: */  7,
        /* tcp_StateLASTACK: */  7,
        /* tcp_StateTIMEWT:  */  2
        /*
         * S. Lawson - define a short TIME_WAIT timeout. It should 
         * be 2MSL but it's not really practical for us.
         * 2 secs will hopefully handle the case where ACK must be
         * retransmitted, but can't protect future connections on the 
         * same port from old packets.
        */
      };
      unsigned int timeout = 0;
      if (s->state == tcp_StateLISTEN || s->state == tcp_StateESTAB)
        timeout = s->bc_idiocy.phase_timeout_s;
      else if (s->state < (sizeof(to_table)/sizeof(to_table[0])))
        timeout = to_table[s->state];

      if (timeout && (timer_sec_count() - s->state_to.time) > timeout)
      {
        __tcp_stop_conn(s, WSO_ETIMEDOUT, __LINE__);
      }        
    }
  }
  else if (s->snd_buf.usedlen <= 0 && !s->unhappy && s->karn_count != 1)
  {
    /* nothing - no retransmission necessary */
  }
  else if (timer_sec_count() <= s->rexmit.timeout_s)
  {
    /* nothing - hasn't timedout yet */
  }
  else if (s->karn_count == 2 && /* count is valid */
           s->rexmit.count > TCPREXMIT_MAXCNT) /* too many tries. */
  {
    s->rexmit.timeout_s = 0UL;
    __tcp_stop_conn(s, WSO_ETIMEDOUT, __LINE__ );
  }
  else  /* retransmission strategy timeout */
  {
    if (debug_lev)
      DBGPRINTF(( "regular retran (rexmit.timeout_s=%lus). Unacked=%u\n",
             s->rexmit.timeout_s, s->unacked ));

    /* strategy handles closed windows.  JD + EE
    */
    if (s->window == 0 && s->karn_count == 2)
      s->window = 1;

    if (s->karn_count == 0)
    {
      /* Simple "Slow start" algorithm:
       * Use the backed off RTO - implied, no code necessary.
       * Reduce the congestion window by 25%
       */
      unsigned int cwindow = s->cwindow;
      s->cwindow = (uint8_t)(((cwindow + 1) * 3) >> 2);
      if (s->cwindow == 0)
        s->cwindow = 1;
      s->wwindow = 0;       /* dup ACK counter ? */
     // s->snd_ssthresh = s->cwindow * s->max_seg;  // !!to-do

      /* if really did timeout
       */
      s->karn_count = 2;
      s->unacked    = 0;
      s->rexmit.count = 0;
    }

    s->rexmit.count++;
    s->rexmit.timeout_s = 0UL;           /* stop RTT timer */

    if (s->snd_buf.usedlen)
      s->flags |= (tcp_FlagPUSH | tcp_FlagACK);

#if 0
    if (s->unhappy)
      TCP_STAT (tcps_rexmttimeo++);

    else if (s->flags & tcp_FlagACK)
      TCP_STAT (tcps_delack++);
#endif

    TCP_SEND (s);
  } /* if (timer_sec_count() > s->rexmit.timeout_s) */

  return;
}

/* -------------------------------------------------------------------- */

/*
 * Force the peer to send us a segment by sending a keep-alive packet:
 *   <SEQ=SND.UNA-1><ACK=RCV.NXT><CTL=ACK>
*/
static void __tcp_do_keepalives(tcp_pcb FAR *s)
{
  if (s->locflags & WSO_LF_KEEPALIVE && (s->state == tcp_StateESTAB ||
                                         s->state == tcp_StateESTCL))
  {
    uint32_t now = timer_sec_count();
    unsigned int ka_to = TCPTV_KEEP_INTVL; /* subsequent probe send timeout */

    if (s->keepalive.count == 0) /* first probe send timeout */
      ka_to = TCPTV_KEEP_IDLE; 
    else if (tcp_keepalive_intvl > 0)
      ka_to = tcp_keepalive_intvl; /* subsequent probe send timeout */
    else
      ka_to = TCPTV_KEEP_INTVL; /* subsequent probe send timeout */

    if ((now - s->last_rx_time_s) < ka_to)
    {
      /* reset time and count if we got _any_ packet */
      s->keepalive.last_send = 0;
      s->keepalive.count = 0; 
    }
    else if ((now - s->keepalive.last_send) < ka_to)
    {
      ; /* nothing */
    }
    else if (s->keepalive.count >= TCPTV_KEEP_CNT)
    {
      __tcp_stop_conn(s, WSO_ETIMEDOUT, __LINE__);
      /* s->err_msx = "Connection timed out - no activity"; */
    }
    else
    {
      uint32_t ack     = s->acknum;
      uint32_t seq     = s->seqnum;
      uint8_t  kc      = s->karn_count;
      int   datalen = s->snd_buf.usedlen;
    
      s->acknum     = s->seqnum;
      s->seqnum     = s->unacked - 1;
      s->flags      = tcp_FlagACK;
      s->karn_count = 2;
      s->snd_buf.usedlen = 0;
      TCP_SEND (s);

      s->acknum     = ack;
      s->seqnum     = seq;
      s->karn_count = kc;
      s->snd_buf.usedlen = datalen;

      s->keepalive.last_send = now;
      s->keepalive.count++; 
    }
  }
  return;
}

/* -------------------------------------------------------------------- */

/*
** __tcp_periodic_cb is an periodic callback registered with
** blocker_perdhook_add() when a socket is created
*/
static void __tcp_periodic_cb(void)
{
  tcp_pcb FAR *s, FAR *next;

  __validate_allsocs(NULL);

  for (s = _tcp_allsocs; s; s = next)
  {
    next = s->next;

    __tcp_Retransmitter(s);

    __tcp_do_keepalives(s);
  }
  return;
}

/* ==================================================================== */
/*                         sock open/listen                             */
/* ==================================================================== */

/*
 * Allocate a new protocol control block, and then thread it.
*/
static tcp_pcb FAR *__alloc_pcb(void)
{
  /* Be careful how fields are initialized.
   * Allocating a PCB (or creating a socket) should (by and of itself) not 
   * cause the IP stack to initialize. Stack initialization should 
   * first happen when either data is xferred, or the machine needs 
   * to be reachable, ie, either tcp_connect() or tcp_listen().
   *
   * The only "touchy" field (currently, May 2001) is s->max_seg, which
   * depends on knowning the MTU (and thus on ether being up) to initialize 
   * completely. Ergo, s->max_seg is initially set to zero ("unknown"), and
   * initialized properly in the first TCP_SEND(). 
  */
  tcp_pcb FAR *s = (tcp_pcb FAR *)FAR_MALLOC(sizeof(tcp_pcb));
  #if (USHRT_MAX == UINT_MAX)
  unsigned int rcv_buf_sz = (8u*1024u);
  #else
  unsigned int rcv_buf_sz = (16u*1024u);
  #endif
  unsigned int snd_buf_sz = 2048u;
  uint8_t FAR *rcv_buf = (uint8_t FAR *)FAR_MALLOC(rcv_buf_sz);
  uint8_t FAR *snd_buf = (uint8_t FAR *)FAR_MALLOC(snd_buf_sz);

  if (!s || !rcv_buf || !snd_buf)
  {
    if (snd_buf)
      FAR_FREE(snd_buf);
    if (rcv_buf)
      FAR_FREE(rcv_buf);
    if (s)
      FAR_FREE(s);
    s = (tcp_pcb FAR *)0;
  }
  else
  {
    FAR_MEMSET(s, 0, sizeof(*s));

    s->state        = tcp_StateCLOSED;

    s->rcv_buf.buf     = rcv_buf;
    s->rcv_buf.usedlen = 0;
    s->rcv_buf.maxlen  = rcv_buf_sz;
    s->snd_buf.buf     = snd_buf;
    s->snd_buf.usedlen = 0;
    s->snd_buf.maxlen  = snd_buf_sz;

    /* s->max_seg is a dummy value here (so that IP and ether are not forced 
     * to boot prematurely). The first TCP_SEND() will set it.
    */
    s->max_seg      = 0; /* unknown - first TCP_SEND will set it */
    s->cwindow      = 1;
    s->wwindow      = 0;                      /* slow start VJ algorithm */
    s->vj_sa        = INIT_VJSA;
    s->hisaddr      = 0;
    s->hisport      = 0;
    s->seqnum       = INIT_SEQ();
    s->flags        = 0;
    s->unhappy      = 1;
    s->rexmit.timeout_s = 0;
    s->rexmit.intvl_ms = 0; /* first TCP_SEND() will figure this out */
    s->linger_delay_s = TCPTV_MSL<<1; /* secs, (0 == don't linger) */

    s->myaddr       = 0;
    s->myport       = 0;
    s->ttl          = 60; /* RFC 793 page 50/51 */
    s->tos          = (uint8_t)_ip_get_default_tos();
    s->xmit_timeout_s = 128; /* secs - 0 (forever) ... 255 (4.25min) */
    s->recv_timeout_s = 128; /* secs - 0 (forever) ... 255 (4.25min) */
    if (tcp_rw_timeout > 0)
    {
      if (tcp_rw_timeout > 255)
        tcp_rw_timeout = 255;
      s->xmit_timeout_s = s->recv_timeout_s = (uint8_t)tcp_rw_timeout;
    }

    /* the following options have no effect on listeners, but since a connected
    ** socket may be created from a listening one, do them for all socks.
    */
    s->locflags     = WSO_LF_BLOCKING; /* note! */
    if (tcp_nonagle)
      s->locflags  |= WSO_LF_NONAGLE;
    if (tcp_keepalive_allon > 0)        /* enable keep alives for all sox */
      s->locflags  |= WSO_LF_KEEPALIVE;

    s->pcb_marker_top = s->pcb_marker_bottom = (uint32_t)s;

    s->next         = _tcp_allsocs;    /* insert into chain */
    _tcp_allsocs    = s;
  }

  return s;
}

/* ------------------------------------------------------------------- */

EXPORTED int tcp_create(struct sockio_info *__s)
{
  tcp_pcb FAR *s = __validate_allsocs(__s);

  if (s)
    __pcb_destroy(s);
  else if (!__s)
  {
    errno = EINVAL;
    return -1;
  }
  blocker_yield(0); /* la, la, la */

  if (blocker_perdhook_add( __tcp_periodic_cb )!=0) 
  {
    errno = EMFILE;
    return -1; 
  }

  s = __alloc_pcb();
  if (!s)
    return -1; /* errno is already set to ENOMEM */

  /* make the new PCB addressable from user space */
  s->sio_link     = __s;
  __s->dispatch   = __get_dispatch_tab();
  __s->pcb        = (void FAR *)s;

  return (0);
}

/* ------------------------------------------------------------------- */

static int __pcb_bind( tcp_pcb FAR *s, const uint32_t FAR *laddr, 
                       const uint16_t FAR *lport)
{
  uint32_t addr = 0;
  uint16_t  port = 0;

  if (!laddr) 
    laddr = &addr;
  if (!lport) 
    lport = &port;

  if (s->state != tcp_StateCLOSED || /* must be already bound */
      s->myport != 0)                /* closed but already bound */
  {
    s->so_errno = WSO_EINVAL; /* TOUTSTATE */
    return -1;
  }

  if (*laddr && !is_local_addr(*laddr))
  {
    s->so_errno = WSO_EADDRNOTAVAIL; /* bad address */
    return -1;
  }

  s->myport = localport_reserve(*lport, IPPROTO_TCP, s->locflags);  
  if (s->myport == 0)/* in use and other sock doesn't have REUSEADDR */
  {
    s->so_errno = WSO_EADDRINUSE; 
    return -1;
  }

  s->myaddr = *laddr;
  return (0);
}

EXPORTED int tcp_bind(struct sockio_info *__s, const uint32_t *laddr, const uint16_t *lport)
{
  tcp_pcb FAR *s;

  blocker_yield(0); /* la, la, la */
  s = __validate_allsocs(__s);

  if (s)
    return __pcb_bind(s, laddr, lport);
  return -1;
}

/* ------------------------------------------------------------------- */

EXPORTED int tcp_listen_bsd(struct sockio_info *__s, int qlen)
{
  tcp_pcb FAR *s;

  blocker_yield(0); /* la, la, la */
  s = __validate_allsocs(__s);
  if (!s)
    return -1;

  /* socket must be 
   * - alive (not shutdown),
   * - and either closed+bound or
   *   already listening (if listening, then this is a "change queue" call)
   * - and not be a sock initialized with tcp_listen_wcompat
  */
  if ((s->locflags & WSO_LF_WATTCOMPAT) /* must use tcp_listen_wcompat() */
   || s->myport == 0                /* not bound yet */
   || (s->state != tcp_StateCLOSED && s->state != tcp_StateLISTEN)
   || s->locflags & WSO_LF_CANTRECVMORE) /* has been shutdown */
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (qlen <= 0)
    qlen = 1;
  if (qlen > (int)(sizeof(s->bsd_accept.list)/sizeof(s->bsd_accept.list[0])) )
    qlen = (sizeof(s->bsd_accept.list)/sizeof(s->bsd_accept.list[0]));

  if (s->state == tcp_StateLISTEN) /* already listening */
  {
    unsigned int i;
    for (i = qlen+1; i < s->bsd_accept.maxready; i++)
    {
      if (!s->bsd_accept.list[i]) /* done all */
        break;
      __force_abort(s->bsd_accept.list[i]);
      /* conn is aborted, mem is free()'d, s->bsd_accept.list[i]=NULL */
    }
  }

  s->bsd_accept.maxready = qlen;
  s->state     = tcp_StateLISTEN;
  s->unhappy   = 0;

  return (0);
} 

/* ------------------------------------------------------------------- */

/*
** __find_accept_ready() locates a socket ready for accept,
** and optionally (if foundP != NULL) pops it from the queue.
*/
static int __accept_ready_find(tcp_pcb FAR *s, tcp_pcb FAR * FAR *foundP)
{
  unsigned int i, n;

  if (s->state != tcp_StateLISTEN)
    return 0;

  n = s->bsd_accept.maxready;
  for (i = 0; i < n; i++)
  {
    if (!s->bsd_accept.list[i])
      break;
    if (s->bsd_accept.list[i]->state == tcp_StateESTAB)
    {
      if (foundP) /* request to dequeue */
      {
        *foundP = s->bsd_accept.list[i];
        s->bsd_accept.list[i]->parent = (tcp_pcb FAR *)0;
        for (--n; i < n; i++)
          s->bsd_accept.list[i] = s->bsd_accept.list[i+1];
        s->bsd_accept.list[n] = (tcp_pcb FAR *)0;
      }
      return 1;
    }
    else if (s->bsd_accept.list[i]->state >= tcp_StateESTAB)
    {
      __force_abort(s->bsd_accept.list[i]);
      n = s->bsd_accept.maxready;
    }
  }
  return 0;
}

/* ------------------------------------------------------------------- */

EXPORTED int tcp_accept(struct sockio_info *__s, 
                        struct sockio_info *__news, 
                        uint32_t *raddr, uint16_t *rport)
{
  tcp_pcb FAR *clone, FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  blocker_yield(0); /* la, la, la */

  if (!__news || s->state != tcp_StateLISTEN ||
      s->bsd_accept.maxready < 1) /* wrong kind of (compat) listen */
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (!__accept_ready_find(s, &clone))
  {
    /* If no pending connections are present on the queue, and the 
     * socket is not marked as nonblocking, accept blocks the caller
     * until a connection is present. 
     * If the socket is marked nonblocking and no pending connections
     * are present on the queue, accept returns EWOULDBLOCK.
    */
    uint16_t accerr = WSO_EWOULDBLOCK;

    if (s->locflags & WSO_LF_BLOCKING)
    {
      /* note that blocking accept never times out */
      int oldbrkstate = set_cbreakstate(0);

      accerr = 0;
      while (!accerr)
      {
        if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
          accerr = WSO_EINTR; 
        else if (__accept_ready_find(s, &clone))
          break;
        else if (s->bc_idiocy.yield_fn)
          (*(s->bc_idiocy.yield_fn))();
      } 
      set_cbreakstate(oldbrkstate);

    } /* if (s->locflags & WSO_LF_BLOCKING) */

    if (accerr)
    {
      s->so_errno = accerr;
      return -1;
    }
  }

  /* 
  ** XTI compat note: Everything will work fine if '__s' == '__news'. The
  ** tcp_destroy(__news) will destroy 's', stop the listen, RST all other 
  ** children, free() any allocations made, unthread and free() the pcb.
  */
  s = __validate_allsocs(__news);
  if (s)
    __pcb_destroy(s);

  /* make the PCB addressable from user space */
  s->sio_link     = __news;
  __news->dispatch= __get_dispatch_tab();
  __news->pcb     = (void FAR *)s;

  if (raddr)
    *raddr = clone->hisaddr;
  if (rport)
    *rport = clone->hisport;

  /* child is already threaded, simply return */
  return (0);
}
    
/* ------------------------------------------------------------------- */

/*
 * Initialize a socket for old Wattsock TCP listen function compatibility. 
 * The old wattsock tcp_listen differs from the new in that the
 * old wattsock tcp simply converted a listener into an active connection,
 * while new socks follow BSD accept() semantics.
 *
 * With the old wattsock listen, multiple calls to tcp_listen_wcompat() to 
 * the same local port (lport) are acceptable and constitutes what-a-poo-tcp's
 * mechanism for supporting multiple incoming connections to the same 
 * local port. Each time another host attempts to open a session on that 
 * particular port, another one of the listens will be consumed until such 
 * time as all listens have become established sessions and subsequent 
 * remote host attempts will receive a reset. 
*/
EXPORTED int tcp_listen_wcompat(struct sockio_info *__s, uint16_t lport, uint32_t raddr, uint16_t rport)
{
  tcp_pcb FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  blocker_yield(0); /* la, la, la */

  if (s->myport != 0) /* already bound (implies !closed) */
  {
    s->so_errno = WSO_EINVAL; /* TOUTSTATE */
    return -1;
  }

  if ((s->locflags & WSO_LF_WATTCOMPAT) == 0) /* must use tcp_listen() */
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (is_multicast(raddr) || raddr == ((uint32_t)-1))
  {
    s->so_errno = WSO_EADDRNOTAVAIL;
    return -1;
  }

  s->locflags |= WSO_LF_REUSEADDR; /* share allowed if in other in timewait */
  s->locflags |= WSO_LF_REUSEPORT; /* share allowed if all are reuseport */

  lport = localport_reserve(lport, IPPROTO_TCP, s->locflags );
  if (lport == 0)
  {
    s->so_errno = WSO_EADDRINUSE; /* port in use (shouldn't happen here) */
    return -1;
  }

  s->state     = tcp_StateLISTEN;
  s->myport    = lport;
  s->hisport   = rport;
  s->hisaddr   = raddr;
  s->unhappy   = 0;

  return (0);
}

/* ------------------------------------------------------------------- */

static int __pcb_connect( tcp_pcb FAR *s, 
                      const uint32_t FAR *raddr, const uint16_t FAR *rport )
{
  if (s->state == tcp_StateLISTEN)
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (!raddr || !rport)
  {
    s->so_errno = WSO_EDESTADDRREQ;
    return -1;
  }

  if (s->state <= tcp_StateESTAB && 
     (*raddr != s->hisaddr || *rport != s->hisport))
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (s->state == tcp_StateCLOSED)
  {
    if (*raddr == 0 || *rport == 0 || *raddr == ((uint32_t)-1) ||
        is_multicast(*raddr) || is_broadcast(*raddr))
    {
      s->so_errno = WSO_EINVAL;
      return -1;
    }
    if (s->myport == 0) /* not bound yet */
    {
      uint16_t port = 0; uint32_t addr = 0;
      if (__pcb_bind(s, &addr, &port) != 0)
        return -1;
    }

    /* blow away any old information that might affect a new 
     * connection. Although it would be much easier to
     * memset(), DONT! Local port is bound, most information 
     * in locflags remains, and so do any other options that 
     * the user may have modified (bufsize, timeouts etc).
     * 
     * I'm not sure whether TTL should be kept or reset to default.
    */
    s->so_errno     = 0;
    s->ttl          = 60; /* RFC 793 page 50/51 */
    s->tos          = (uint8_t)_ip_get_default_tos();  
    s->max_seg      = 0; /* unkown - TCP_SEND() below will set it */
    s->locflags    &= WSO_LF_MASK_USER_LF;
    s->seqnum       = INIT_SEQ();
    s->acknum       = 0;
    s->recent       = 0;
    s->rigid        = 0;
    s->stress       = 0;
    s->karn_count   = 0;
    s->vj_sa        = INIT_VJSA;
    s->vj_sd        = 0;
    s->vj_last      = 0;
    s->rcv_buf.usedlen = 0;
    s->snd_buf.usedlen = 0; 
    s->missed_seg[0]= 0;
    s->missed_seg[1]= 0;
    #if 0
    s->queuelen     = 0;
    s->queue        = NULL;
    #endif
    s->unacked      = 0;
    s->cwindow      = 1;
    s->wwindow      = 0;
    s->flags        = tcp_FlagSYN;
    s->state        = tcp_StateSYNSENT;
    s->rexmit.timeout_s = 0;
    s->rexmit.intvl_ms = 0; /* first TCP_SEND() will figure this out */
    s->unhappy      = 1;
    s->hisaddr      = *raddr;
    s->hisport      = *rport;

    /* send the opening SYN - if it fails its because of an error
     * below TCP, eg an ARP resolve failed, or a packet driver
     * error, s->so_errno will be set.
    */
    if (TCP_SEND(s) != 0)  /* FAILED */
    {
      s->state = tcp_StateCLOSED;
    }
  }

  /* 
   * If we got here its because...
   *   first pass connect (in-progress or immediate failure)
   *   second+ pass connect (connected, in-progress or async failure)
   *
   * Resolve the 'in-progress' conditions to either connected or failure.
  */
  if (s->state < tcp_StateESTAB)
  {
    int oldbrkstate;

    if ((s->locflags & WSO_LF_BLOCKING) == 0) /* non-blocking connect */
    {
      /* can't resolve the 'in-progress' condition in this pass */
      s->so_errno = WSO_EINPROGRESS;
      return -1;
    }

    /* blocking connect 
    */
    /* that s->hisaddr == *raddr and s->hisport == *rport
     * has already been asserted at the top of the fxn
     * 
     * The timeout is tracked in the periodic handler.
     * (shunts the state to CLOSED when it expires).
    */
    oldbrkstate = set_cbreakstate(0);
    while (s->state < tcp_StateESTAB)
    {
      if (blocker_yield(YIELDFLAG_HARD|YIELDFLAG_CHKBRK))
      {
        s->so_errno = WSO_EINTR; 
        break;
      }
      if (s->bc_idiocy.yield_fn)
        (*(s->bc_idiocy.yield_fn))();
    }
    set_cbreakstate(oldbrkstate);
  }

  /* 
   * If we got here its because...
   *   first pass connect (connected or immediate failure)
   *   second+ pass connect (connected or async failure)
   *
   * Resolve the 'connected' conditions and return.
  */
  if (s->state == tcp_StateESTAB) 
  {
    /* the result we return depends on how this connect completion 
     * was "noticed", and not on how it was actually completed. 
    */
    if ((s->locflags & WSO_LF_BLOCKING) == 0) /* non-blocking connect */
    {
      /* we noticed it asynchronously */
      s->so_errno = WSO_EALREADY;
      return -1;
    }
    /* From the view of the user, this blocking connect was 
     * _just_ completed.
    */
    return 0;
  }

  /* 
   * All thats left are the failure conditions. Whatever the reason,
   * s->so_errno will have been set:
   * 
   * Synchronous errors:  EHOSTUNREACH (ARP resolve failed), EIO (packet
   *                      driver error), EINTR (from the blocking wait).
   * Asynchronous errors: ETIMEDOUT (polling loop), ExxxUNREACH (ICMP), 
   *                      ECONNREFUSED (ICMP)
  */
  if (s->state != tcp_StateCLOSED) /* may be FINWT1 if timedout */
  {
    int olderr = s->so_errno;
    __force_abort(s); /* => tcp_StateCLOSED */
    s->so_errno = olderr;
  }
  return -1;
}                                      

EXPORTED int tcp_connect(struct sockio_info *__s, const uint32_t *raddr, const uint16_t *rport)
{
  tcp_pcb FAR *s;
  blocker_yield(0); /* la, la, la */
  s = __validate_allsocs(__s);
  if (s)
    return __pcb_connect( s, raddr, rport );
  return -1;
}

/* ==================================================================== */
/*              TCP state machine (called from tcp_handler)             */
/* ==================================================================== */

/*
TCP initiation ("3-way handshake") and connection: The client will do a
 connection initiation, with the tcp SYN flag set, an arbitrary sequence
 number, and no acknowledgement number. The server acknowledges the SYN by
 sending a packet with SYN and ACK set, another random sequence number and the
 acknowledgement number the original sequence. Finally, the client replies back
 with a tcp datagram with the ACK flag set, and the server's ack sequence
 incremented by one. Once the connection is established, each tcp segment
 will be sent with no flags (PSH and URG are optional), the sequence number
 for each packet incremented by the size of the previous tcp segment. After
 the amount of data specified as "window size" has been transferred, the
 peer sending data will wait for an acknowledgement, a tcp segment with the
 ACK flag set and the ack sequence number the one of the last data packet
 that could be received in order. That way, if any segments get lost, they
 will not be acknowledged and can be retransmitted. To end a connection,
 both server and client send a tcp packet with correct sequence numbers and
 the FIN flag set, and if the connection ever de-synchronizes (aborted,
 desynchronized, bad sequence numbers, etc.) the peer that notices the error
 will send a RST packet with correct seq numbers to terminate the connection.

 - Mixter <mixter@newyorkoffice.com>
*/

/*
 * Process the data in an incoming packet.
 * Called from all states where incoming data can be received:
 * SYNSENT, ESTAB, ESTCL, CLOSWT, FIN-WAIT-1 and FIN-WAIT-2
 */
static void tcp_ProcessData (tcp_pcb FAR *s, 
                             const tcp_Header FAR *tcp, 
                             int len, int flags)
{
  long  ldiff, tmpldiff;
  int   diff,  tmpdiff, data_ofs;
  const uint8_t FAR *data;

  if (s->stress > 0)
      s->stress--;

  s->window = NTOHS(tcp->window);
  if (s->window > MAX_WINDOW(s))
      s->window = MAX_WINDOW(s);

  ldiff = s->acknum - NTOHL(tcp->seqnum);

  if (flags & tcp_FlagSYN)
     ldiff--;                         /* back up to 0 */

  diff = abs ((int)ldiff);

  /* find the data portion
   */
  data_ofs = tcp->offset << 2;        /* dword to byte format */
  data = ((const uint8_t FAR *) tcp) + data_ofs;

  if (data_ofs - sizeof(*tcp) > 0)    /* process TCP options */
  {
    const uint8_t FAR *opt = (const uint8_t FAR *)(tcp+1);

    while (opt < data)
    {
      switch (*opt)
      {
        case TCPOPT_EOL:
             opt = data;
             break;

        case TCPOPT_NOP:
             opt++;
             break;

        case TCPOPT_MAXSEG:   /* we are very liberal on MSS stuff */
             if (opt[1] >= (2+2) && (flags & tcp_FlagSYN))
             {
               uint16_t max_seg = NTOHS(*(const uint16_t FAR *)(opt+2));
               if (s->max_seg == 0 || max_seg < s->max_seg)
                 s->max_seg = max_seg;
             }
             opt += opt[1];
             break;

        case TCPOPT_TIMESTAMP:
             if (opt[1] >= (8+2) && ((flags & tcp_FlagSYN) ||
                 ((flags & tcp_FlagACK) && ldiff >= 0))) /* only if sequence number ACKs new data */
             {
               s->tsopt.rcvd_num = NTOHL(*(const uint32_t FAR *)(opt+2));
               s->tsopt.returned = NTOHL(*(const uint32_t FAR *)(opt+6));
               s->locflags |= WSO_LF_RCVD_TSTMP;
             }
             opt += opt[1];

        case TCPOPT_WINDOW:
             if (opt[1] >= (2+2) && (flags & tcp_FlagSYN))
             {
               s->recv_wscale = *(opt+2);
               if (s->recv_wscale > TCP_MAX_WINSHIFT)
                 s->recv_wscale = TCP_MAX_WINSHIFT;
               s->locflags |= WSO_LF_RCVD_SCALE;
             }
             opt += opt[1];
             break;
#if 0
        case TCPOPT_SACK_OK:
             if (opt[1] >= (2+0) && (flags & tcp_FlagSYN))
                s->locflags |= WSO_LF_SACK_PERMIT;
             opt += opt[1];
             break;
#endif

        default:              /* unknown options; type,length,... */
             opt += *(opt+1);
      }
    }
  }


  len -= data_ofs;     /* remove the header length */

  /*
   * SYN/RST segments shouldn't carry any data.
   * Stack identifiers, eg nmap, use this to identify the stack being used.
   * Note however that this combination is legal for IETF's 
   * experimental-track Transactional TCP (T/TCP).
   */
  if ((flags & tcp_FlagSYN) == tcp_FlagSYN ||
      (flags & tcp_FlagRST) == tcp_FlagRST)
  {
    return;
  }

  /* Handle any new data that increments the ACK number
   */
  if (ldiff >= 0)
  {
    data += diff;
    len  -= diff;

    {
      /* no handler, just dump to buffer, should be indexed,
       * handles goofs limit receive size to our window
       */
      if (s->rcv_buf.usedlen >= 0)
      {
        int room = s->rcv_buf.maxlen - s->rcv_buf.usedlen;
        if (len > room)
            len = room;

        if (s->missed_seg[0])   /* don't write into missed segment */
        {
          tmpldiff = s->missed_seg[0] - s->acknum;
          tmpdiff  = abs ((int)tmpldiff);
          if (tmpldiff >= 0 && len > tmpdiff)
             len = tmpdiff;
        }
        if (len > 0)
        {
          s->acknum += len;   /* our new ack begins at end of data */
          FAR_MEMCPY(s->rcv_buf.buf + s->rcv_buf.usedlen, data, len);
          s->rcv_buf.usedlen += len;
        }

        if (s->missed_seg[0] && s->missed_seg[0] == s->acknum)
        {
          tmpldiff = s->missed_seg[1] - s->acknum;
          tmpdiff  = abs ((int)tmpldiff);
          if (tmpldiff > 0)
          {
            s->rcv_buf.usedlen += tmpdiff;
            s->acknum = s->missed_seg[1];
          }
          s->missed_seg[0] = 0L;
          s->missed_seg[1] = 0L;
        }
      }
    }

    s->unhappy = s->snd_buf.usedlen ? 1 : 0;
  }
  else    /* handle one dropped segment */
  {
    uint32_t seqnum = NTOHL(tcp->seqnum);

#if 0
    /* S. Lawson - no out-of-sequence processing of FIN flag
     */
    *flags &= ~tcp_FlagFIN;
#endif

    if (s->missed_seg[0] == 0UL)
    {                           /* just dropped a segment */
      if (len > (int)(s->rcv_buf.maxlen - s->rcv_buf.usedlen - diff))
        len = (int)(s->rcv_buf.maxlen - s->rcv_buf.usedlen - diff);
      if (len > 0)
      {
        FAR_MEMCPY(s->rcv_buf.buf + s->rcv_buf.usedlen + diff, data, len);
        s->missed_seg[0]  = s->missed_seg[1] = seqnum;
        s->missed_seg[1] += len;
      }
    }
    else  /* haven't seen missing segment yet */
    {     
      tmpldiff = s->missed_seg[0] - seqnum;
      tmpdiff  = abs ((int)tmpldiff);

      if (tmpldiff > 0 && len >= tmpdiff)  /* prepend bytes to fragment */
      {   
        FAR_MEMCPY(s->rcv_buf.buf + s->rcv_buf.usedlen + diff, data, tmpdiff);
        s->missed_seg[0] -= tmpdiff;
      }
      tmpldiff = seqnum + len - s->missed_seg[1];
      tmpdiff  = abs ((int)tmpldiff);

      /* append bytes touching fragment
       */
      if (tmpldiff > 0)
      {          
        tmpldiff = s->missed_seg[1] - seqnum;
        if (tmpldiff >= 0)
        {
          /* packet source offset
           */
          int dst, src = abs ((int)tmpldiff);

          /* buffer destination offset
           */
          tmpldiff = s->missed_seg[1] - s->acknum;
          dst = abs ((int)tmpldiff);

          /* length to move
           */
          if (tmpdiff > (int)(s->rcv_buf.maxlen - s->rcv_buf.usedlen - dst))
            tmpdiff = (s->rcv_buf.maxlen - s->rcv_buf.usedlen - dst);
          if (tmpdiff > 0)
          {
            FAR_MEMCPY(s->rcv_buf.buf + s->rcv_buf.usedlen + dst, data + src, tmpdiff);
            s->missed_seg[1] += tmpdiff;
          }
        }
      }
    }
  }
  if (debug_lev)
    DBGPRINTF(("*new s->rcv_buf.usedlen = %d\n", s->rcv_buf.usedlen ));

  return;
}


/* ----------------------------------------------------------------------- */

/*
 * Process the ACK value in received packet, but only if it falls within
 * current window. Discard queued data that have been acknowledged.
 */
static int tcp_ProcessAck (tcp_pcb FAR *s, const tcp_Header FAR *tcp, long *unacked)
{
  long ldiff = (long) (NTOHL(tcp->acknum) - s->seqnum);
  int  diff  = (int) ldiff;

  if (unacked)
     *unacked = ldiff;

  if (debug_lev >= 2)
  {
    DBGPRINTF(("*tcp_ProcessAck(): tcp->acknum=%lu, s->seqnum=%lu. ldiff=%ld\n", 
                         NTOHL(tcp->acknum), s->seqnum, ldiff ));
    DBGPRINTF(("*                  s->snd_buf.usedlen = %d\n", s->snd_buf.usedlen ));
  }

  if (ldiff >= 0L && ((unsigned int)diff) <= s->snd_buf.usedlen)
  {
    #if 0
    if (s->queuelen)
    {
      s->queue    += diff;
      s->queuelen -= diff;
    }
    else 
    #endif
    {
      if (((unsigned int)diff) < s->snd_buf.usedlen)
        FAR_MEMMOVE(s->snd_buf.buf, s->snd_buf.buf+diff, 
                    s->snd_buf.usedlen-diff);
    }

    s->snd_buf.usedlen -= diff;
    s->unacked -= diff;
    s->seqnum  += ldiff;
    return (1);
  }
  return (0);
}

/* ----------------------------------------------------------------------- */

/*
 * LISTEN state
 */
static int __tcp_LISTEN_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                               const tcp_Header FAR *tcp, int flags)
{
  if ((flags & (tcp_FlagSYN|tcp_FlagRST)) == tcp_FlagSYN)
  {   
    if (ip->tos > s->tos)
      s->tos = ip->tos;
    else if (ip->tos < s->tos)
    {
     /* RFC 793 says we should close connection.
      * We best not do that since SunOS ignores TOS
      */
    }
    s->acknum  = NTOHL(tcp->seqnum) + 1;
    s->flags   = (tcp_FlagSYN | tcp_FlagACK);
    s->state   = tcp_StateSYNREC;
    s->unhappy = 1;
    TCP_SEND (s);            /* respond immediately */
  }
  else
  {
    if (!(flags & tcp_FlagRST))  /* don't answer RST */
       TCP_SEND_RESET (NULL, ip, tcp);   /* send a reset */
  }
  return (0);
}

/* ----------------------------------------------------------------------- */

/*
 * SYNSENT state
 */
static int __tcp_SYNSENT_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                              const tcp_Header FAR *tcp, int flags)
{
  if (flags & tcp_FlagSYN)
  {
    if (ip->tos > s->tos)
         s->tos = ip->tos;
    else if (ip->tos < s->tos)
    {
     /* RFC 793 says we should close connection.
      * We best not do that while SunOS ignores TOS
      */
    }

    s->flags   = tcp_FlagACK;

    /* FlagACK means connection established, else SYNREC
     */
    if (flags & tcp_FlagACK)
    {
      /* but is it for the correct session ?
       */
      if (NTOHL(tcp->acknum) == (s->seqnum + 1))
      {
        int len = NTOHS(ip->length) - in_GetHdrLen (ip);

        s->state = tcp_StateESTAB;
        s->seqnum++;                          /* good increment */
        s->acknum = NTOHL(tcp->seqnum) + 1;
        tcp_ProcessData (s, tcp, len, flags); /* someone may try it */
        s->unhappy = 1;                    /* rely on their attempts */
        TCP_SEND (s); /* !! maybe use TCP_SENDSOON() to merge application
                         data into ACK */
      }
      else
      {
        /* wrong ack, force a RST and resend SYN now
         */
        s->flags   = tcp_FlagRST;
        s->unhappy = 1;
        TCP_SEND (s);
        s->seqnum = INIT_SEQ();  /* !! should we set a new seq-num? */
        s->flags  = tcp_FlagSYN;
        TCP_SEND (s);
      }

      /* !!to-do: recalculate RTT-timer
       */
    }
    else
    {
      s->acknum++;
      s->state = tcp_StateSYNREC;
      return (0);
    }
  }
  else
  {
    TCP_SEND_RESET (s, ip, tcp);
    return (0);              /* Added 18-Aug 1999, GV */
  }

  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * SYNREC state
 */
static int __tcp_SYNREC_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                             const tcp_Header FAR *tcp, int flags)
{
  ip = ip; /* unreferenced - shaddup compiler */

  if (flags & tcp_FlagSYN)  /* retransmitted SYN */
  {
    s->flags   = (tcp_FlagSYN | tcp_FlagACK);
    s->unhappy = 1;
    TCP_SEND (s);
    return (0);
  }

  if ((flags & tcp_FlagACK) && (NTOHL(tcp->acknum) == (s->seqnum + 1)))
  {
    s->window = NTOHS(tcp->window);
    if (s->window > MAX_WINDOW(s))
        s->window = MAX_WINDOW(s);
    s->seqnum++;
    s->flags   = tcp_FlagACK;
    s->state   = tcp_StateESTAB;
    s->unhappy = 0;
    return (0);
  }

  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * ESTABLISHED state
 */
static int __tcp_ESTAB_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                            const tcp_Header FAR *tcp, int flags)
{
  unsigned int len;
  long  ldiff;
  uint32_t ack;

  /* handle lost SYN
   */
  if ((flags & tcp_FlagSYN) && (flags & tcp_FlagACK))
  {
    TCP_SEND (s);
    return (0);
  }

  if (!(flags & tcp_FlagACK))   /* must ack something */
     return (0);

  if (!tcp_ProcessAck(s,tcp,&ldiff))
  {
    if (debug_lev)
      DBGPRINTF(("*tcp_ProcessAck() confused so set unacked "
                            "back to 0 from %u\n", s->unacked));
    s->unacked = 0;
  }
  if (s->unacked < 0)
      s->unacked = 0;

  s->flags = tcp_FlagACK;

  len = NTOHS(ip->length) - in_GetHdrLen (ip);
  ack = s->acknum;
  tcp_ProcessData (s, tcp, len, flags);

  if (s->state != tcp_StateCLOSWT &&
      (flags & tcp_FlagFIN)       &&
      s->missed_seg[0] == 0L      &&
      ack == NTOHL(tcp->seqnum))
  {
    s->acknum++;
    /*
    * if (s->err_msx == NULL)
    *    s->err_msx = "Connection closed";
    */

    TCP_SEND (s);

    s->locflags |= WSO_LF_GOT_FIN;
    s->flags    |= tcp_FlagFIN;    /* for the tcp_Retransmitter() */
    s->unhappy   = 1;
    s->state     = tcp_StateLASTACK;
  }

  /*
   * Eliminate the spurious ACK messages bug.
   * For the window update, the length should be the
   * data length only, so exclude the TCP header size
   *  -- Joe <jdhagen@itis.com>
   */
  len -= tcp->offset << 2;
  if ((ldiff > 0 && s->snd_buf.usedlen > 0) || len > 0)
  {
   /* Need to update window, but how urgent ??
    * Added new window update -GV 21-Nov-1998
    */
    unsigned int winfree   = s->rcv_buf.maxlen - s->rcv_buf.usedlen;
    unsigned int treshhold = s->max_seg >> 1;
    if (treshhold < winfree)
      treshhold = winfree;
    if (s->missed_seg[0] || ldiff > 0 || len > treshhold)
    {
      s->karn_count = 0;
      s->flags |= tcp_FlagPUSH;  /* !! added */
      TCP_SEND (s);
    }
    else
      TCP_SENDSOON (s);
  }
  return (0);
}

/* ----------------------------------------------------------------------- */

/*
 * ESTAB_CLOSE state
 */
static int __tcp_ESTCL_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                            const tcp_Header FAR *tcp, int flags)
{
  __tcp_ESTAB_state (s, ip, tcp, flags);
  __pcb_shutdown(s, 1);
  return (0);
}

/* ----------------------------------------------------------------------- */

/*
 * CLOSE_WAIT state
 */
static int __tcp_CLOSEWT_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                              const tcp_Header FAR *tcp, int flags)
{
  return __tcp_ESTAB_state (s, ip, tcp, flags);
}

/* ----------------------------------------------------------------------- */

/*
 * FIN_WAIT1 state -
 */
static int __tcp_FINWT1_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                             const tcp_Header FAR *tcp, int flags)
{
  long  ldiff;
  int   len = NTOHS(ip->length) - in_GetHdrLen (ip);
  uint32_t ack, seq;

  if (flags & tcp_FlagFIN)
     s->locflags |= WSO_LF_GOT_FIN;

 /* They have not necessarily read all the data
  * yet, we must still supply it as requested
  */
  if (tcp_ProcessAck(s,tcp,&ldiff))
  {
    if (ldiff == 0 || s->unacked < 0)
       s->unacked = 0;
  }

  /* they may still be transmitting data, we must read it
   */
  tcp_ProcessData (s, tcp, len, flags);

  ack = NTOHL(tcp->acknum);
  seq = NTOHL(tcp->seqnum);
 

  /* check if other tcp has ACK'ed all sent data and
   * is ready to change states.
   */
  if (s->missed_seg[0] == 0L && 
     (flags & (tcp_FlagSYN | tcp_FlagACK)) == (tcp_FlagSYN | tcp_FlagACK))
  {                         
    if (seq == s->acknum)
    {
      s->acknum++;               /* we must ACK their FIN! */

      if (ack >= s->seqnum + 1)
      {
        /* Not simultaneous close (they've ACKed our FIN)
         * We need to ACK their FIN and move to TIME_WAIT
         */
        s->seqnum++;
        s->state   = tcp_StateTIMEWT;
      }
      else
      {
        /* Simultaneous close (haven't ACKed our FIN yet)
         * We need to ACK their FIN and move to CLOSING
         */
        s->state = tcp_StateCLOSING;
      }
      s->flags   = tcp_FlagACK;
      s->unhappy = 0;
      TCP_SEND (s);
    }
  }
  else if (flags & tcp_FlagACK)
  {
    /* other side is legitimately acking our FIN
     */
    if (ack == s->seqnum + 1)
    {
      if (seq == s->acknum)
      {
        if (s->snd_buf.usedlen == 0)
        {
          s->seqnum++;
          s->state   = tcp_StateFINWT2;
          s->unhappy = 0;             /* we don't send anything */
        }
      }
      else if (seq == s->acknum + 1)
      {
        /* !! added 30-Aug 1999 GV
        * Try to stop that annoying retransmission bug/feature(?)
        * from FreeBSD 4.x which increments both SEQ and ACK.
        */
        s->seqnum++;
        s->acknum++;
        s->flags      = tcp_FlagRST;
        s->unhappy    = 0;
        s->karn_count = 0;
        s->snd_buf.usedlen    = 0;
        TCP_SEND (s);

        s->state = tcp_StateCLOSED;
        return (0);
      }
    }
  }
  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * FIN_WAIT2 state
 */
static int __tcp_FINWT2_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                             const tcp_Header FAR *tcp, int flags)
{
  int  len = NTOHS(ip->length) - in_GetHdrLen (ip);
  uint32_t ack, seq;

  /* They may still be transmitting data, we must read it
   */
  ack = NTOHL(tcp->acknum);
  seq = NTOHL(tcp->seqnum);

  tcp_ProcessData (s, tcp, len, flags);

  if (s->missed_seg[0])  /* must retransmit to get all data */
     return (1);

  if (flags & tcp_FlagFIN)
     s->locflags |= WSO_LF_GOT_FIN;

  if ((flags & (tcp_FlagSYN | tcp_FlagACK)) == (tcp_FlagSYN | tcp_FlagACK) &&
      ack == s->seqnum && seq == s->acknum)
  {
    s->acknum++;
    s->flags   = tcp_FlagACK;
    s->unhappy = 0;          /* don't send anything */
    s->state   = tcp_StateTIMEWT;
    TCP_SEND (s);
    return (0);
  }
  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * CLOSING state
 */
static int __tcp_CLOSING_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                              const tcp_Header FAR *tcp, int flags)
{
  ip = ip; /* unreferenced - shaddup compiler */

  if ((flags & (tcp_FlagSYN | tcp_FlagACK)) == (tcp_FlagSYN | tcp_FlagACK))  
  {                                                        /* ACK, no FIN */
    /* Per FINWT1 above, tcp->acknum should be s->seqnum+1,
     * which should cause us to bump s->seqnum to match
     */
    if (NTOHL(tcp->acknum) >= (s->seqnum + 1) &&
        NTOHL(tcp->seqnum) == s->acknum )
    {
      s->seqnum++;
      s->state   = tcp_StateTIMEWT;
      s->unhappy = 0;
    }
  }
  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * LASTACK state
 */
static int __tcp_LASTACK_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                              const tcp_Header FAR *tcp, int flags)
{
  ip = ip; /* unreferenced - shaddup compiler */

  if (flags & tcp_FlagFIN)
  {
    /* they lost our two packets, back up
     */
    s->locflags |= WSO_LF_GOT_FIN;
    s->flags     = (tcp_FlagSYN | tcp_FlagACK);
    TCP_SEND (s);
    s->unhappy = 1;
    return (0);
  }

  if (NTOHL(tcp->acknum) == (s->seqnum + 1) &&
      NTOHL(tcp->seqnum) == s->acknum)
  {
    s->state   = tcp_StateCLOSED;   /* no 2msl necessary */
    s->unhappy = 0;             /* we're done        */
    return (0);
  }
  return (1);
}

/* ----------------------------------------------------------------------- */

/*
 * TIMEWAIT state
 */
static int __tcp_TIMEWT_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                             const tcp_Header FAR *tcp, int flags)
{
  ip = ip; /* unreferenced - shaddup compiler */
  tcp = tcp; /* unreferenced - shaddup compiler */

  if ((flags & (tcp_FlagSYN | tcp_FlagACK)) == (tcp_FlagSYN | tcp_FlagACK))
  {
    /* our peer needs an ack
     */
    s->flags   = tcp_FlagACK;
    s->unhappy = 0;
    s->state   = tcp_StateCLOSED;  /* support 2 msl in RST code */
    TCP_SEND (s);
  }
  return (1);
}

/* ----------------------------------------------------------------------- */

static int __tcp_CLOSED_state (tcp_pcb FAR *s, const in_Header FAR *ip,
                             const tcp_Header FAR *tcp, int flags)
{
  s = s; /* unreferenced - shaddup compiler */

  if ((flags & tcp_FlagRST) == 0)
    TCP_SEND_RESET (NULL, ip, tcp);
  return (0);
}

/* ----------------------------------------------------------------------- */

/*
 *  __update_counters 
 *  Update idle timers, retransmission timer,  VJ algorithm and tcp windows.
*/
static void __update_counters(tcp_pcb FAR *s)
{
  if (s->karn_count != 2) /* no retransmission was needed */
  {
    if (s->wwindow != 255)
    {
      if (++s->wwindow >= s->cwindow)  /* A. Iljasov (iljasov@oduurl.ru) */
      {                                /* suggested this !!              */
        if (s->cwindow != 255)
          s->cwindow++;
        s->wwindow = 0;    /* mdurkin -- added 95.05.02 */
      }
    }

    /* (ACK is received for a segment which was not retransmitted) */
    if (s->vj_last)  /* Set when we expected an immediate response */
    {
      long deltaT = (timer_msec_count()+(timer_msec_tres()-1)) - s->vj_last;

      /* deltaT cannot be negative at this point. 
       * '(timer_msec_tres() - 1)' is a "fudge factor" to compensate for
       * the fact that timer_msec_count() probably does not have real 
       * one msec resolution (its 55ms on x86s, and 20ms on 68Ks).
      */

      deltaT   -= (s->vj_sa >> 3);
      s->vj_sa += deltaT;

      if (deltaT < 0)
        deltaT = -deltaT;

      deltaT   -= (s->vj_sd >> 2);
      s->vj_sd += deltaT;

      if (s->vj_sa > MAX_VJSA) /* 80000UL */
        s->vj_sa = MAX_VJSA;
      if (s->vj_sd > MAX_VJSD) /* 20000UL */
        s->vj_sd = MAX_VJSD;

      s->rexmit.intvl_ms = (TCPTV_RTO_BASE * 1000u) + 
                           (((s->vj_sa >> 2) + (s->vj_sd)) >> 1);

      if (debug_lev >= 1) /* >= 2 will do a full dump */
        DBGPRINTF(("*counters: rexmit.intvl_ms %u  sa %u  sd %u  cwindow %u"
         "  wwindow %u  unacked %u\n", 
         s->rexmit.intvl_ms, s->vj_sa, s->vj_sd, s->cwindow,
         s->wwindow, s->unacked ));
    }
  } /* if (s->karn_count != 2) */


  if (debug_lev && s->karn_count == 2 /* we re-transmitted */ && s->unacked)
    DBGPRINTF(("finally got karn_count safely zapped from %u\n",s->unacked));

  /* update our retransmission stuff (Karn algorithm) */
  s->karn_count = 0;

  /* Idle timers (Keepalive etc) depend on last_rx_time_s */
  s->last_rx_time_s = timer_sec_count(); 

  /* Restart retransmission timer based on calculated round-trip-time.
   * In *THIS* case, s->rexmit.timeout_s is changed only if the new
   * expiry time is later than the current expiry time.
  */
  SET_REXMIT_TIMEOUT(s, s->rexmit.intvl_ms, 1, __LINE__);
}

/* ----------------------------------------------------------------------- */

/*
 * __tcp_fsm - Our TCP-input state-machine.
*/
static int __tcp_fsm (tcp_pcb FAR *s, const in_Header FAR *ip )
{
  typedef int (*tcp_StateProc)(tcp_pcb FAR *, const in_Header FAR*, 
                               const tcp_Header FAR*, int flags );
  static tcp_StateProc tcp_state_tab[] = 
  {
    __tcp_LISTEN_state,   /* tcp_StateLISTEN  : listening for connection */
    __tcp_SYNSENT_state,  /* tcp_StateSYNSENT : SYN sent, active open */
    __tcp_SYNREC_state,   /* tcp_StateSYNREC  : SYN received, SYN+ACK sent. */
    __tcp_ESTAB_state,    /* tcp_StateESTAB   : established */
    __tcp_ESTCL_state,    /* tcp_StateESTCL   : established, but will FIN */
    __tcp_FINWT1_state,   /* tcp_StateFINWT1  : sent FIN */
    __tcp_FINWT2_state,   /* tcp_StateFINWT2  : sent FIN, received FIN+ACK */
    __tcp_CLOSEWT_state,  /* tcp_StateCLOSWT  : received FIN waiting for close */
    __tcp_CLOSING_state,  /* tcp_StateCLOSING : sent FIN, received FIN (waiting for FIN+ACK) */
    __tcp_LASTACK_state,  /* tcp_StateLASTACK : FIN received, ACK+FIN sent */
    __tcp_TIMEWT_state,   /* tcp_StateTIMEWT  : dally after sending final FIN+ACK */
    __tcp_CLOSED_state    /* tcp_StateCLOSED  : nixed */
  };

  int rc = 0;
  const tcp_Header FAR *tcp;
  uint16_t  len;
  uint8_t  flags;

  len   = (uint16_t)in_GetHdrLen (ip);                /* len of IP header   */
  tcp   = (const tcp_Header FAR*) ((const uint8_t FAR *)ip + len);  /* tcp frame pointer  */
  flags = (uint8_t)(tcp->flags & tcp_FlagMASK); /* get TCP-head flags */
  len   = (uint16_t)(NTOHS(ip->length) - len);     /* len of tcp+data   */

  /* update retrans timer, windows etc. */
  __update_counters(s);       

  if (debug_lev >= 2)
  {
    __dbg_dump_state( s );
  }

  /* Disallow IP-fragments if the peer disallows them.
  ** (to-do!! ensure that the packet size does not exceed _path_ MTU
  ** otherwise packets will get dropped by the routers (and we'll
  ** get ICMP_UNREACH/CANTFRAG)
  */
  if (NTOHS(ip->frag_ofs) & IP_DF)
    s->locflags |=  WSO_LF_NOFRAGMENT;
  else 
    s->locflags &= ~WSO_LF_NOFRAGMENT;

  if (flags & tcp_FlagRST)         /* got reset code */
  {
    rc = 0;
    __tcp_stop_conn(s, WSO_ECONNRESET, __LINE__);
    if (debug_lev)
      DBGPRINTF(("*TCP IN-%04x: got RST\n", NTOHS(ip->identification)));
  }
  else if (s->state > tcp_StateCLOSED)
  {
    rc = 0;
  }
  else
  {
    tcp_PseudoHeader ph;

    ph.src      = ip->source;
    ph.dst      = ip->destination;
    ph.mbz      = 0;
    ph.protocol = IPPROTO_TCP;
    ph.length   = HTONS((uint16_t)len);
    ph.checksum = checksum (tcp, len);

    if (checksum(&ph,sizeof(ph)) != 0xFFFF)
    {
      if (debug_lev)
        DBGPRINTF(("*TCP IN-%04x: DROP: bad checksum\n", 
                     NTOHS(ip->identification) ));
      rc = 2;
    }
    else
    {
      if (debug_lev)
        DBGPRINTF(("*TCP IN-%04x: calling tcp_%s_state()\n", 
          NTOHS(ip->identification), sock_strstate(s->state) ));

      /* State engine functions return !0 if an ACK is outstanding.
       * (retransmission may happen immediately or on RTO expiry)
      */
      if ((*tcp_state_tab[s->state]) (s, ip, tcp, flags))
        rc = (s->unhappy ? 2 : 1);

      if (debug_lev) 
        DBGPRINTF(("*TCP IN-%04x: end tcp_%s_state() => %d\n", 
         NTOHS(ip->identification), sock_strstate(s->state), rc ));

    } /* if (checksum) */
 
    if (rc == 2)
    {
      TCP_SENDSOON(s);
    }
  } /* if (!reset) */

  if (rc != 0 /* something worthwhile happened */ && debug_lev >= 2)
  {
    __dbg_dump_state( s );
  }

  return rc;
}

/* ==================================================================== */
/*                             tcp input                                */
/* ==================================================================== */

static void __tcp_ip_recv (const in_Header FAR *ip, int is_llbcast,
                           int raw_handled )
{
  tcp_pcb FAR *s = (tcp_pcb FAR *)0;
  const tcp_Header FAR *tcp;
  uint8_t  tcpflags;
  uint32_t srcIP = NTOHL(ip->source);
  uint32_t dstIP = NTOHL(ip->destination);
  uint16_t dstPort, srcPort, pktid;

  pktid = NTOHS(ip->identification);

  if (is_llbcast || !is_local_addr(dstIP) || is_multicast(srcIP))
  {
    DBGPRINTF(("*TCP IN-%04x: DROP (isbcast || ismcast || notforme)\n",pktid));
    return;
  }
  if ((NTOHS(ip->length) - in_GetHdrLen (ip)) < sizeof(*tcp))
  {
    DBGPRINTF(("*TCP IN-%04x: DROP (short header)\n",pktid));
    return;
  }

  tcp   = (const tcp_Header FAR *) (((const uint8_t FAR *)ip) + 
                                   in_GetHdrLen (ip)); /* tcp seg ptr */
  tcpflags = (uint8_t)(tcp->flags & tcp_FlagMASK); /* get TCP flags     */

  __validate_allsocs(NULL);

  if (debug_lev >= 2)
  {
    __dbg_dump_header(ip, tcp, NULL, 0 );
    DBGPRINTF(("*  *** src = %hI, dest = %hI\n", srcIP, dstIP ));
  }

  if (tcp->offset < (sizeof(*tcp)>>2))
  {
    DBGPRINTF(("*TCP IN-%04x: DROP (bad header)\n",pktid));
    return;
  }    

  if (NTOHS(ip->length) < (in_GetHdrLen(ip) + (((unsigned int)tcp->offset) << 2) ))
  {
    DBGPRINTF(("*TCP IN-%04x: DROP (short packet)\n",pktid));
    return;
  }    

  dstPort = NTOHS(tcp->dstPort);
  srcPort = NTOHS(tcp->srcPort);

  if ((tcpflags & tcp_FlagSYN) == 0  /* not new conn (in either direction) */
   || (tcpflags & (tcp_FlagSYN|tcp_FlagACK)) == (tcp_FlagSYN|tcp_FlagACK))
  {                                    /* we initiated new session */
    s = _tcp_allsocs;
    while (s)
    {
      if (s->state == tcp_StateCLOSED ||
          s->state == tcp_StateLISTEN ||  /* listen handled in next loop */
          dstPort != s->myport ||                  /* not my port? */
          (s->myaddr != 0 && dstIP != s->myaddr)) /* or not to my address? */  
      {
        ; /* nothing - fall through and loop to next */
      }
      else if (srcIP  == s->hisaddr &&   /* from my peer address */
               srcPort == s->hisport)     /* from correct remote port */
      {
        __tcp_fsm(s, ip); /* demux this (its for me) */
        return;         
      }
      else if ((tcpflags & (tcp_FlagRST|tcp_FlagACK)) 
                        == (tcp_FlagRST|tcp_FlagACK) ) /* got ACK,RST */
      {                                          
        if (NTOHL(tcp->acknum) == (s->seqnum + 1)) /* ACK = SEQ + 1 */
        {        
          /* e.g. a firewall is sending RST for server on inside */
          __tcp_stop_conn(s, WSO_ECONNRESET, __LINE__); 
        }
      }
      s = s->next;
    } /* while (s) */
  } /* if ((tcpflags & tcp_FlagSYN) == 0) */ /* not new conn */ 

  else if ((tcpflags & (tcp_FlagSYN|tcp_FlagRST)) == tcp_FlagSYN) 
  {                                       /* its a new session */
    int pass = 0;
    for (pass = 0; !s && pass < 2; pass++)
    {
      s = _tcp_allsocs;
      while (s)
      {
        if (s->state == tcp_StateLISTEN && 
            s->myport == dstPort &&
            (s->locflags & WSO_LF_CANTRECVMORE) == 0 && /* not shutdown(0) */
            ((pass == 0 && s->myaddr == dstIP) ||
             (pass == 1 && s->myaddr == 0)) )
        {
          if (s->locflags & WSO_LF_WATTCOMPAT)
          {
            if ((s->hisaddr == 0 || s->hisaddr == srcIP) &&
                (s->hisport == 0 || s->hisport == srcPort))
            {
              s->hisport = srcPort;   /* remember his IP-address */
              s->hisaddr = srcIP;     /*   and src-port */
              s->myaddr  = dstIP;     /* socket is now active */
              /* demux this (was listening, now active) */
              __tcp_fsm(s, ip); 
              return;         
            }
          }
          else /* uses tcp_accept() */
          {
            unsigned int i,n;

            n = s->bsd_accept.maxready;
            for (i = 0; i < n; i++)
            {
              if (!s->bsd_accept.list[i])
              {
                tcp_pcb FAR *newsock = __alloc_pcb();
                if (newsock)
                {
                  s->bsd_accept.list[i] = newsock;
                  newsock->parent       = s;
                  newsock->myaddr       = dstIP;
                  newsock->myport       = dstPort;
                  newsock->hisaddr      = srcIP;
                  newsock->hisport      = srcPort;

                  newsock->ttl          = s->ttl;
                  newsock->tos          = s->tos;
                  newsock->bc_idiocy.yield_fn = s->bc_idiocy.yield_fn;
                  newsock->xmit_timeout_s = s->xmit_timeout_s;
                  newsock->recv_timeout_s = s->recv_timeout_s;
                  newsock->locflags     = s->locflags;

                  __tcp_fsm(newsock, ip);
                  return;
                }
                break;
              }
            }

            break; /* no slot found of cloning failed - send a RST */

          } /* uses tcp_accept() */
        } /* listening and valid for this listener */

        s = s->next; 

      } /* while (s) */
    } /* for (pass = 0; !s && pass < 2; pass++) */
  } /* if (tcpflags & tcp_FlagSYN) */

  /* didn't find anything to demux to. */
  if (!raw_handled && /* wasn't already dealt with elsewhere */
      !(tcpflags & tcp_FlagRST))   /* don't answer RST */
  {
    TCP_SEND_RESET (NULL, ip, tcp);
  }

  pktid = pktid; /* not used unless debugging */
  return;
}

/* ----------------------------------------------------------------------- */

static void __tcp_icmp_recv (const in_Header FAR *ip, int soicmperr_code)
{
  tcp_pcb FAR *s;
  const tcp_Header FAR *tcp = (const tcp_Header FAR *)
                        (((const uint8_t FAR *)ip) + in_GetHdrLen (ip));
                                           
  uint32_t dstIP   = NTOHL(ip->destination);
  uint16_t srcPort = NTOHS(tcp->srcPort);
  uint16_t dstPort = NTOHS(tcp->dstPort);

  if (debug_lev)
        DBGPRINTF(("icmp msg type=%d src=%I:%d dest=%I:%d\n", 
         soicmperr_code, ip->source, srcPort, ip->destination, dstPort ));

  for (s = _tcp_allsocs; s; s = s->next)
  {
    if (s->state != tcp_StateLISTEN && s->state != tcp_StateCLOSED &&
        srcPort == s->myport && dstPort == s->hisport && dstIP == s->hisaddr)
    {
      switch (soicmperr_code)
      {
        case SOICMPERR_CANTFRAG:
             /* bogus - we don't do DF */ 
             break;

        case SOICMPERR_TIMXCEED_TTL:
             if (s->ttl < 255)
               s->ttl++;
             /* FALL-THROUGH */

        case SOICMPERR_TIMXCEED_FRAG:
             /* FALL-THROUGH */

        case SOICMPERR_UNREACH_HOST: /* includes unreachable net etc */
        case SOICMPERR_UNREACH_PORT:
             s->locflags |= WSO_LF_UPEND_QOS; 
             s->stress++; /* _pre_increment! */
             if (s->rigid < 100 &&       /* otherwise infinitely rigid */
                 s->stress > s->rigid)  
             {                           /* halt it */
               int oldstate = s->state;
               s->rcv_buf.usedlen = 0;
               s->snd_buf.usedlen  = 0;
               s->unhappy  = 0;
               __force_abort(s);
               if (oldstate < tcp_StateESTAB) 
                 s->so_errno = WSO_ECONNREFUSED;
               else
                 s->so_errno = WSO_ECONNRESET;
               s->locflags |= WSO_LF_ASYNC_ERROR;
               /* s->err_msx = "Unreachable"; */
               break;
             }
             /* FALL-THROUGH */

        case SOICMPERR_SOURCEQUENCH:    
             s->locflags |= WSO_LF_UPEND_QOS; 
             s->cwindow = 1;       /* slow-down tx-rate */
             s->wwindow = 1;
             s->vj_sa <<= 2;
             s->vj_sd <<= 2;
             s->rexmit.intvl_ms <<= 2;
             break;

        case SOICMPERR_PARAMPROB:
             /* should not happen since we don't do IP options */
             __pcb_shutdown(s, 1); /* was abort */
             s->locflags |= WSO_LF_ASYNC_ERROR;
             s->so_errno = WSO_EINVAL;
             break;

         default:
             if (debug_lev)
               DBGPRINTF(("huh! unknown tcp:soicmp_code %d\n",
                                          soicmperr_code ));
      }
    }
  }
  return;
}

/* ----------------------------------------------------------------------- */

/*
 * Handler for incoming udp packets and for ICMP messages.
 */
static void __tcp_handler (const in_Header FAR *ippkt, int protohndlr_flags)
{
  if (protohndlr_flags & PROTOHNDLR_FLAGS_ICMPCB)
  {
    int soicmp_code = (protohndlr_flags & 0xff);
    __tcp_icmp_recv( ippkt, soicmp_code );
  }
  else
  {
    __tcp_ip_recv(ippkt, (protohndlr_flags & PROTOHNDLR_FLAGS_LLBCAST),
                         (protohndlr_flags & PROTOHNDLR_FLAGS_RAWDONE) );
  }
  return;
}

/* ======================================================================= */
/*                         write to tcp sock                               */
/* ======================================================================= */

/*
 * tcp_send_to - enqueue/send 'buf'.
 * Returns the same results as BSD send, 
 * ie -1 Error (see s->so_errno), no data has been enqueued.
 *     0 No error, and no data could be enqueued (queue is full).
 *    >0 Total amount of data enqueued/sent (may be less than total
 *       even with blocking send if an error occurs).
 *
 * Argument subtleties:
 * - Zero byte writes are valid, and represent a flush. 
 *
 * For compatibility with old tcp_write()/sock_fastwrite() the 
 * function does not yield at _all_ unless WSO_MSG_WAITALL is specified.
*/

/* ------------------------------------------------------------------- */

static int __nb_enqueue_data( tcp_pcb FAR *s, 
                              const void *data, unsigned int *datalen, int flags )
{
  /* Returns the amount of data enqueued on the socket through *datalen;
   * Returns the result of TCP_SEND()/TCP_SENDSOON() (s->errno will be 
   * updated if those functions fail.)
  */
  const uint8_t *p;
  unsigned int room, rawlen, len = *datalen;
  
  room = (s->snd_buf.maxlen - 1) - s->snd_buf.usedlen;

  rawlen = 0; /* shaddup compiler */
  if (len && room && (s->locflags & WSO_LF_DOASCII)) /* ick, ick, ick */
  {
    unsigned int n = len;
    p = (const uint8_t *)data;
    rawlen = len = 0;
    while (n)
    {
      n--;
      if (*p++ == '\n')
      {
        if ((len + 2) > room)
          break;
        len++;
      }
      else if (len >= room)
        break;
      len++;
      rawlen++;
    }
  }

  if (len > room)
    len = room;
  else /* 'data' fits completely - there will be no more data coming */
    flags &= ~WSO_MSG_WAITALL;

  if (len > 0)
  {
    if ((s->locflags & WSO_LF_DOASCII) == 0) 
      FAR_MEMCPY(s->snd_buf.buf + s->snd_buf.usedlen, data, len);
    else /* gross, gross, gross */
    {
      uint8_t FAR *wpos = s->snd_buf.buf + s->snd_buf.usedlen;
      p = (const uint8_t *)data;
      while (rawlen)
      {
        rawlen--;
        if (*p == '\n')
          *wpos++ = '\r';
        *wpos++ = *p++;
      }
    }
    s->snd_buf.usedlen  += len;
    s->unhappy   = 1;    /* redundant because we have outstanding data */

    if (!(flags & WSO_MSG_WAITALL))
    {
      /* queue up data, flush on next write */
      if (s->locflags & WSO_LF_HOLDLOCAL) 
        s->locflags &= ~WSO_LF_HOLDLOCAL;
    }
  }
  *datalen = len;

  s->so_errno = 0;

  /* transmit if first data or reached MTU.
   * not true MTU, but better than nothing
  */
  if ((s->locflags & WSO_LF_NONAGLE) || 
    s->snd_buf.usedlen == len || s->snd_buf.usedlen > (s->max_seg >> 1))
  {
    return TCP_SEND (s);
  }
  return TCP_SENDSOON (s);
}

EXPORTED int tcp_send_to(struct sockio_info *__s, const void *data, 
                         unsigned int len, int flags,
                         const uint32_t *hisip, const uint16_t *hisport )
{
  tcp_pcb FAR *s = __validate_allsocs(__s);

  if (!s)
    return -1;

  if (s->locflags & WSO_LF_ASYNC_ERROR)
  {
    /* errno already set */
    s->locflags &= ~WSO_LF_ASYNC_ERROR; /* consume the error */
    return -1; /* errno is already set */
  }

  if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == 
               (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT))
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }
  else if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == 0)
  {
    if (s->locflags & WSO_LF_BLOCKING)
      flags |= WSO_MSG_WAITALL;
  }
  if ((flags & WSO_MSG_OOB) && (s->locflags & WSO_LF_OOBINLINE))
  {
    s->so_errno = WSO_EINVAL;
    return -1;
  }

  if (s->locflags & WSO_LF_CANTSENDMORE)
  {
    if (s->so_errno != WSO_ECONNABORTED)
      s->so_errno = WSO_ESHUTDOWN;
    return -1;
  }

  if (s->state != tcp_StateCLOSED)
  {
    if (s->state == tcp_StateLISTEN || s->state > tcp_StateESTAB)
    {
      s->so_errno = WSO_ENOTCONN;
      return -1; 
    }
    if ((hisip && *hisip != s->hisaddr) ||
        (hisport && *hisport != s->hisport))
    {
      s->so_errno = WSO_EINVAL;
      return -1;
    } 
  }

  if (s->state < tcp_StateESTAB)
  {
    uint32_t oldlocflags;
    const uint32_t FAR *addrP = &s->hisaddr;
    const uint16_t  FAR *portP = &s->hisport;
    if (s->state == tcp_StateCLOSED)
    {
      if (!hisip || !hisport)
      {
        s->so_errno = WSO_ENOTCONN;
        return -1;
      } 
      addrP = hisip;
      portP = hisport;
    }
    oldlocflags = s->locflags;
    s->locflags |= WSO_LF_BLOCKING;
    if (__pcb_connect(s, addrP, portP) != 0)
    {
      s->locflags = oldlocflags;
      return -1; /* s->so_errno is already set */
    }
    s->locflags = oldlocflags;
  }

  if (data == NULL) /* == flush */
  {
    s->locflags &= ~WSO_LF_HOLDLOCAL; /* undo any WSO_LF_HOLDLOCAL */
    if (s->snd_buf.usedlen > 0)
    {
      s->flags |= tcp_FlagPUSH;
      if (s->unacked == 0)  /* !! S. Lawson - only if data not moving */
         TCP_SEND (s);
    }
    len = 0;
    /* fallthrough */
  }

  if (len != 0)
  {
    unsigned int totalsent  = 0;
    uint32_t timeout   = 0ul;
    int oldbrkstate =-1;

    if (len == UINT_MAX)
      len--;

    if (s->xmit_timeout_s == 0) /* wait forever */
      flags |= WSO_MSG_WAITALL;

    if (len != 0 && (flags & WSO_MSG_WAITALL))
    {
      if (s->xmit_timeout_s)
        timeout = timer_sec_count() + s->xmit_timeout_s;
      blocker_yield(0);
    }

    while (len != 0)
    {
      unsigned int txadded;
      int rc;

      if (flags & WSO_MSG_OOB)
      {
        /* if we could do OOB this is where it would go */
        s->so_errno = WSO_EINVAL;
        break;
      }

      txadded = len; /* txadded is unsigned */
      rc = __nb_enqueue_data(s, data, &txadded, flags);
      /* __nb_enqueue_data() returns result of TCP_SEND() (-1=err, 0=ok) */

      data = (((const uint8_t *)data) + txadded);
      len -= txadded; 
      totalsent += txadded;

      if (rc == -1) /* TCP_SEND() error - s->errno is set */
        break;
      if (len == 0) /* all done */
        break;

      if (!(flags & WSO_MSG_WAITALL))
      {
        if (totalsent == 0) /* queue was already full */
          s->so_errno = WSO_EWOULDBLOCK;
        break;
      }

      if (txadded != 0) /* something got added, restart the timer */
      {
        if (timeout) /* not infinite */
          timeout = timer_sec_count() + s->xmit_timeout_s;    
      }
      else if (timeout && timer_sec_count() > timeout)
      {
        s->so_errno = WSO_ETIMEDOUT;
        break;
      }
      if (oldbrkstate == -1)
      {
        if ((oldbrkstate = set_cbreakstate(0))!=0) 
          oldbrkstate = 1;
      }
      
      if (blocker_yield(YIELDFLAG_CHKBRK|(!rc ? YIELDFLAG_HARD : 0)))
      {
        s->so_errno = WSO_EINTR;
        break;
      }
      if (s->bc_idiocy.yield_fn)
        (*(s->bc_idiocy.yield_fn))();
    }

    if (oldbrkstate != -1)
    {
      set_cbreakstate(oldbrkstate);
    }

    if (totalsent == 0 && len != 0)
      totalsent--; /* -1 */
    len = totalsent;
  }

  return len;
}

/* ======================================================================= */
/*                         read from tcp sock                              */
/* ======================================================================= */

/*
 * tcp_recv_from - read socket data to 'buf'.
 * Returns the same results as BSD recv, 
 * ie -1 == no data avail, (errno updated)
 *    0  == connection lost (errno is unmodified)
 *    >0 == number of bytes read. (errno is unmodified)
 *
 * Argument subtleties:
 * - Zero byte reads are valid. To differentiate between 'connection lost'
 *   and the returned result from the 'zero bytes to read', either check
 *   the s->so_errno or check if 'src_ip/src_port' were set.
 * - NULL buff reads are valid too. The data is simply discarded (if !peek).
 *   The NULL buff check is done AFTER the zero length write check.
 *
 * For compatibility with old tcp_read()/sock_fastread() the function does
 * not yield at _all_ unless WSO_MSG_WAITALL is specified.
*/
#undef  TRACE_RCVFROM

#ifdef  TRACE_RCVFROM
#undef  TRACE_RCVFROM
#define TRACE_RCVFROM(x) DBGPRINTF(x)
#else
#define TRACE_RCVFROM(x) NO_OPERATION()
#endif

static int __nb_dequeue_data( tcp_pcb FAR *s, 
                              void *buf, unsigned int maxlen, int flags,
                              uint32_t *src_ip, uint16_t *src_port )
{
  unsigned int len;

  if (s->locflags & WSO_LF_ASYNC_ERROR) /* s->so_errno already set */
  {
    s->locflags &= ~WSO_LF_ASYNC_ERROR; /* consume the error */
    if (debug_lev > 0)
      TRACE_RCVFROM(("return 0 (EDISCONNECT) due to WSO_LF_ASYNC_ERROR\n"));
    return 0; /* disconnect (errno is not modified) [is set] */
  }

  if (s->rcv_buf.usedlen <= 0)
  {
    if (s->state < tcp_StateESTAB)
    {
      s->so_errno = WSO_ENOTCONN;
      if (debug_lev > 0)
        TRACE_RCVFROM(("return -1 due to ENOTCONN\n"));
      return -1;
    }
    if (s->state > tcp_StateESTCL) /* closing/closed */
    {
      if (debug_lev > 0)
        TRACE_RCVFROM(("return 0 due to state > tcp_StateESTCL\n"));
      return 0; /* disconnect (errno is not modified) */
    }      
    s->so_errno = WSO_EWOULDBLOCK; 
    if (debug_lev > 0)
      TRACE_RCVFROM(("return -1 due to no data avail\n"));
    return -1; /* no data currently available */ 
  }

  /* if the pcb has been shutdown, simply purge the recv buf (return -1) */
  if (s->locflags & WSO_LF_CANTRECVMORE)
  {
    maxlen = s->rcv_buf.usedlen;
    buf = (void *)0;
    flags &= ~WSO_MSG_PEEK;
  }

  len = s->rcv_buf.usedlen;
  if (len > maxlen)
    len = maxlen;

  if (len > 0)
  {
    if (buf)
    {
      if ((s->locflags & WSO_LF_DOASCII) == 0)
        FAR_MEMCPY(buf, s->rcv_buf.buf, len);
      else /* ick, ick, ick */
      {
        register uint8_t FAR *p = s->rcv_buf.buf;
        register unsigned int rlen = s->rcv_buf.usedlen;
        register unsigned int clen = maxlen;
        uint8_t *b = (uint8_t *)buf;
        len = 0;
        while (len < rlen && clen > 0)
        {
          if (*p != '\r')
          {  
            *b++ = *p;
            clen--;
          }
          p++;
          len++;
        }
      }
    }
    if (debug_lev > 0)
      TRACE_RCVFROM(("copied %u bytes to %x\n", len, (uint32_t)buf )); 

    if ((flags & WSO_MSG_PEEK)==0)
    {
      s->rcv_buf.usedlen -= len;

      if (s->missed_seg[0] || s->rcv_buf.usedlen > 0)
      {
        int diff = 0;
        if (s->missed_seg[0])
          diff = abs((int)(s->missed_seg[1] - s->acknum));
        FAR_MEMMOVE( s->rcv_buf.buf, s->rcv_buf.buf + len, 
                     s->rcv_buf.usedlen + diff);
        TCP_SENDSOON (s);     /* update the window soon */
      }
      else if ((s->rcv_buf.maxlen - s->rcv_buf.usedlen) < (s->max_seg >> 1))
      {                       /* receive window needs an update. */
        TCP_SEND(s);  /* update window now */
      }
    }
  }

  if (s->locflags & WSO_LF_CANTRECVMORE) /* matches similar if() above */
  {
    if (debug_lev > 0)
      TRACE_RCVFROM(("return -1 due to WSO_LF_CANTRECVMORE\n"));
    s->so_errno = WSO_ESHUTDOWN;
    return -1;
  }

  if (src_ip)
    *src_ip = s->hisaddr;
  if (src_port)
    *src_port = s->hisport;

  if (debug_lev > 0)
    TRACE_RCVFROM(("return len = %u (s->rcv_buf.usedlen = %d)\n", len,
                           s->rcv_buf.usedlen )); 

  return (len);
}

EXPORTED int tcp_recv_from(struct sockio_info *__s, void *buf, unsigned int maxlen, 
                            int flags, uint32_t *src_ip, uint16_t *src_port )
{
  tcp_pcb FAR *s    = __validate_allsocs(__s);
  unsigned int read_len    = 0;
  uint32_t timeout    = 0ul;
  int oldbrkstate  = -1;
  int rc           = -1;

  if (!s)
  {
    if (debug_lev > 0)
      TRACE_RCVFROM(("returning -1 due !s\n" ));
    return -1;
  }

  if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT))
  {
    s->so_errno = WSO_EINVAL;
    if (debug_lev > 0)
      TRACE_RCVFROM(("returning -1 due to bad flags\n" ));
    return -1;
  }
  else if ((flags & (WSO_MSG_WAITALL|WSO_MSG_DONTWAIT)) == 0)
  {
    if (s->locflags & WSO_LF_BLOCKING) 
      flags |= WSO_MSG_WAITALL;
  }

  /* only handle the trivial conditions - 
  ** the rest is done in __nb_dequeue
  */
  if (s->state == tcp_StateLISTEN)
  {
    s->so_errno = WSO_ENOTCONN;
    if (debug_lev > 0)
      TRACE_RCVFROM(("returning -1 due to tcp_StateLISTEN\n" ));
    return -1;
  }

  if (flags & WSO_MSG_OOB)
  {
    if ((s->locflags & WSO_LF_OOBINLINE) || !(s->locflags & WSO_LF_URG_AVAIL))
    {
      s->so_errno = WSO_EINVAL;
      if (debug_lev >= 0)
        TRACE_RCVFROM(("returning -1 due to bad OOB flag\n" ));
      return -1;
    }
    s->locflags &= ~WSO_LF_URG_SIGPEND;
    if (maxlen > 0)
    {
      if (buf) 
      {
        *((uint8_t *)buf) = s->oobchar;
        buf = (((uint8_t *)buf) + 1);
      }
      maxlen = 0;
      read_len++;
      if ((flags & WSO_MSG_PEEK) == 0)
        s->locflags &= ~WSO_LF_URG_AVAIL;
    }
    return read_len; /* WSO_MSG_OOB always implies WSO_MSG_DONTWAIT */
  }

  if (s->locflags & WSO_LF_URG_AVAIL)
  {
    s->so_errno = WSO_EINVAL;
    if (debug_lev >= 0)
      TRACE_RCVFROM(("returning -1 due to urg avail but no oob flag\n" ));
    return -1;
  }

  tcp_send_to( __s, NULL, 0, 0, NULL, NULL ); /* flush */

  if (maxlen == UINT_MAX) /* can never return UINT_MAX (its also -1) */
  {
    maxlen--;
  }

  if (maxlen != 0 && (flags & WSO_MSG_WAITALL)!=0)
  {
    if (s->recv_timeout_s) /* otherwise infinite */
      timeout = timer_sec_count() + s->recv_timeout_s;
    blocker_yield(0); /* non-hardware fast yield */
  }

  for (;;) /* not 'while(maxlen)' - maxlen may be zero */
  {
    /* caution: 'rc' has meaning outside the loop */
    rc = __nb_dequeue_data( s, buf, maxlen, (flags & WSO_MSG_PEEK), 
                            src_ip, src_port );

    if (rc == 0) /* connection closed or 'maxlen' == 0 */
    {
      if (debug_lev > 0)
        TRACE_RCVFROM(("break due to rc == 0\n" ));
      break;     /* s->so_errno is unmodified (or is irrelevant) */
    }

    if (rc != -1) /* got data */
    {
      if (flags & WSO_MSG_PEEK)
      {
        read_len = rc; /* always the full length */
        if (read_len == maxlen || read_len == s->rcv_buf.maxlen) 
        {                           /* got as much as physically possible */
          maxlen = 0; /* break out below */
        }
      }
      else 
      {
        read_len += rc;
        if (buf)
          buf = (((uint8_t *)buf) + rc);
        maxlen -= rc;
      }

      if (maxlen == 0 || (flags & WSO_MSG_WAITALL) == 0) 
      {
        if (debug_lev > 0)
          TRACE_RCVFROM(("break due to done all || !WSO_MSG_WAITALL\n" ));
        break;
      }

      if (timeout) /* not infinite */
        timeout = timer_sec_count() + s->recv_timeout_s;
    }

    if (rc == -1) /* no data */
    {
      if (s->so_errno != WSO_EWOULDBLOCK)
      {
        if (debug_lev > 0)
          TRACE_RCVFROM(("break due to no (more) data + %s\n", 
                          sock_strerror(s->so_errno) ));
        break; /* errno already set */
      }

      if ((flags & WSO_MSG_WAITALL) == 0)
      {
        s->so_errno = WSO_EWOULDBLOCK;
        if (debug_lev > 0)
          TRACE_RCVFROM(("break due to no (more) data + !WAITALL\n" ));
        break;
      }

      if (timeout && timer_sec_count() > timeout)
      {
        s->so_errno = WSO_ETIMEDOUT;
        if (debug_lev > 0)
          TRACE_RCVFROM(("break due to no (more) data + TIMEDOUT\n" ));
        break;
      }
    }

    if (oldbrkstate == -1)
    {
      if ((oldbrkstate = set_cbreakstate(0))!=0)
        oldbrkstate = 1;
    }
    if (blocker_yield(YIELDFLAG_CHKBRK|YIELDFLAG_HARD))
    {
      s->so_errno = WSO_EINTR;
      break;
    }

    if (s->bc_idiocy.yield_fn)
      (*(s->bc_idiocy.yield_fn))();

  } /* for (;;) */

  if (read_len == 0 && /* got nothing and ... */
        rc != 0)         /* ... getting nothing was not due to a disconn */
  {
    read_len = (unsigned int)-1;
  }

  if (oldbrkstate != -1)
  {
    set_cbreakstate(oldbrkstate);
  }

  if (debug_lev > 0)
    TRACE_RCVFROM(("returning %d (%s)\n", (int)read_len,
            (read_len == UINT_MAX ? strerror(s->so_errno) : "no err") ));
  
  return read_len; /* -1 is == UINT_MAX here */
}

/* ======================================================================= */
/*                 socket options and condition wait                       */
/* ======================================================================= */

/* moved to a separate file for easier maintainance.
 * The code is mostly identical for udp and tcp
*/
#include "utcommon.ci"

/* ======================================================================= */
/*                          transport initialization                       */
/* ======================================================================= */

static const struct wso_usrreq_tab FAR *__get_dispatch_tab(void)
{
  /* please do not initialize this table at run-time. Initializing it
  ** at compile time makes it easy to catch errors.
  */
  static const struct wso_usrreq_tab FAR dispatch_tab = {
     IPPROTO_TCP,
     tcp_create,
     tcp_bind,
     tcp_connect,
     tcp_listen_bsd,
     tcp_accept,
     tcp_shutdown,
     tcp_destroy,
     tcp_recv_from,
     tcp_send_to,
     xxx_poll,
     xxx_gssockopt,
     tcp_listen_wcompat
  };
  return &dispatch_tab;
}

/* ---------------------------------------------------------------------- */

static void __ctor_pctcp(void)
{
  unsigned int sz;

  sz = sizeof(tcp_Header);
  if (sz != 20)
    __fatal_runtime_error( "PANIC: sizeof(tcp_Header) is wrong. (should be 20)\n" );

  __init_private_data();

  /* register "upwards" */
  _wsio_register_dispatch(__get_dispatch_tab());

  /* register "downwards" */
  _ip_proto_register(IPPROTO_TCP, __tcp_handler, __stack_shutdown, &debug_lev);

  /* register "port-in-use-scanner" */
  localport_register_pius(IPPROTO_TCP, &debug_lev, xxx_port_in_use_scanner );
  /* xxx_port_in_use_scanner is in utcommon.ci */

  return;
}

#define __SETUP_CONSTRUCTOR_FOR__ __ctor_pctcp
#include "ctordtor.h"

