/*
 * sock_mcast_*() userland function.
 *
 * Created by Cyrus Patel <cyp@fb14.uni-mainz.de>
 *
 * sock_mcast_*() are pseudo- externs for use by the user app
 * and are the programmatic way for the user to "request" multicast
 * support for his/her app. If the user doesn't do any group operations
 * this stuff will not be linked. (Using weak symbols would be nicer of 
 * course, but not all compilers support it.) 
 *
 * If the user doesn't use sock_mcast_*(), then this module doesn't
 * get linked. When this module doesn't get linked, igmp.c doesn't get
 * linked either (because this module is the only thing that references
 * stuff in igmp.c directly). When igmp.c doesn't get linked, the 
 * overhead (and to a lesser extent, code bloat) of igmp and multicast 
 * support is not added to the kernel.
 *
 * Written May 2001 by Cyrus Patel <cyp@fb14.uni-mainz.de>
 * 
*/
#define WANT_STACK_CHECKING

#include "target.h"   /* standard includes/types */
#include "ipmcast.h"  /* _ip_mcast_{leave|join|is_in}_group() */
#include "userland.h" /* ourselves (all sock_*() prototypes) */

/*
** int sock_mcast_group_op(in_addr_t ina, int op);
**
** Description
**    sock_mcast_group_op() joins or leaves a multicast group.
**
**    If 'op' is greater than zero, the group 'ina' is joined.
**    If 'op' is less than zero, the group 'ina' is left.
**    If 'op' is equal to zero, returns the current membership state.
**
** Returns
**    For non-zero 'op', sock_mcast_group_op() returns 1 on success or 
**    0 on failure. For a zero 'op', sock_mcast_group_op() returns 1 if
**    the current host is a member of group 'ina', or zero if it is not.
*/
extern const int __IPPROTO_IGMP__; 

int sock_mcast_group_op(in_addr_t ina, int op)
{
  if (!__IPPROTO_IGMP__) /* fake if() to force igmp.c to be linked */
    return 0;         /* __IPPROTO_IGMP__ is the only thing igmp.c exports */

  if (op > 0)
    return _ip_mcast_join_group(ina);
  if (op < 0)
    return _ip_mcast_leave_group(ina);
  return _ip_mcast_is_in_group(ina);
}

                            