/*
 * LIBTSO library exports (userland stuff) prototypes and manifest constants.
 * 
 * ======================================================================
 * This header file has two copies:
 * - as userland.h in the kernel source directory. This is the master copy.
 * - as libtso.h OR {tiny|watt|*}tcp.h in the userland include dir. 
 * When modifying one, remember to update the other.
 * ======================================================================
 *
 * -
 * Copyright (C) 2001 Cyrus Patel <cyp@fb14.uni-mainz.de>.  
 * Portions Copyright (C) 1986 Geof Cooper, IMAGEN Corporation
 * Portions Copyright (C) 1997 Geof Cooper and Rick Rodman (LGPL)
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notices, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must acknowledge providence.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHORS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * --Copyright--
 *
 *    $Id: userland.h,v 1.1 2001/05/07 11:17:21 cyp Exp $
*/
#ifndef __TSIO_VER__
#define __TSIO_VER__ 20011019 

#if defined(__WATCOMC__) && !defined(_ENABLE_AUTODEPEND)
#pragma read_only_file;
#endif

/* ================================================================= */
/*                things common to UDP and TCP  protocol layers      */
/* ================================================================= */
/*
 * TCP states (but also used by UDP), from tcp specification RFC-793.
 * The close-wait state is bypassed by automatically closing a connection
 *       when a FIN is received.
 */
#define tcp_StateLISTEN   0      /* listening for connection */
#define tcp_StateSYNSENT  1      /* SYN sent, active open */
#define tcp_StateSYNREC   2      /* SYN received, ACK+SYN sent. */
#define tcp_StateESTAB    3      /* established */
#define tcp_StateESTCL    4      /* established, but will FIN */
#define tcp_StateFINWT1   5      /* sent FIN */
#define tcp_StateFINWT2   6      /* sent FIN, received FINACK */
#define tcp_StateCLOSWT   7      /* received FIN waiting for close */
#define tcp_StateCLOSING  8      /* sent FIN, received FIN (waiting for FINACK) */
#define tcp_StateLASTACK  9      /* FIN received, FINACK+FIN sent */
#define tcp_StateTIMEWT   10     /* dally after sending final FINACK */
#define tcp_StateCLOSED   11     /* FINACK received */


/* Socket-states for sock_sselect().
 * Not used by BSD-socket API
 */
#define SOCKESTABLISHED  1
#define SOCKDATAREADY    2
#define SOCKCLOSED       4


/* XXX! TCP_MODE_BINARY  0 */    /* default mode */
#define TCP_MODE_ASCII   1
/* XXX! UDP_MODE_CHK     0 */    /* default to checksum */
#define UDP_MODE_NOCHK   2
/* XXX! TCP_MODE_NAGLE   0 */    /* Nagle algorithm */
#define TCP_MODE_NONAGLE 4

/* ----------------------------------------------------------------- */

/* Flags that xxx_recv_from() and xxx_send_to() understand
*/
#define WSO_MSG_OOB         0x1     /* process out-of-band data */
#define	WSO_MSG_PEEK	    0x2     /* peek at incoming message */
#define	WSO_MSG_WAITALL	    0x40    /* wait for full request or error */
#define WSO_MSG_DONTWAIT    0x80    /* this operation is non-blocking */
#define WSO_MSG_EOF         0x100   /* shutdown connection after write */

/* ----------------------------------------------------------------- */

/* socket error codes, originally from winsock.h
*/
#include <errno.h>

/* Codes not in use are commented out so that /when/ they are 
** used I don't forget to add support to sock_strerror().
*/
/*anything under this is a normal errno*/
#define WSO_BASEERR             10000
/* regular C library error constants */
#define WSO_EINTR                EINTR
/* fine WSO_EBADF                EBADF  */
#define WSO_EACCES               EACCES
/* fine WSO_EFAULT               EFAULT */
#define WSO_EINVAL               EINVAL 
/* fine WSO_EMFILE               EMFILE */
/* regular Berkeley error constants */
#define WSO_EWOULDBLOCK          (WSO_BASEERR+35)
#define WSO_EINPROGRESS          (WSO_BASEERR+36)
#define WSO_EALREADY             (WSO_BASEERR+37)
/* fine WSO_ENOTSOCK             (WSO_BASEERR+38) */
#define WSO_EDESTADDRREQ         (WSO_BASEERR+39)
#define WSO_EMSGSIZE             (WSO_BASEERR+40)
/* fine WSO_EPROTOTYPE           (WSO_BASEERR+41) */
#define WSO_ENOPROTOOPT          (WSO_BASEERR+42)
/* fine WSO_EPROTONOSUPPORT      (WSO_BASEERR+43) */
/* fine WSO_ESOCKTNOSUPPORT      (WSO_BASEERR+44) */
#define WSO_EOPNOTSUPP           (WSO_BASEERR+45)
/* fine WSO_EPFNOSUPPORT         (WSO_BASEERR+46) */
/* fine WSO_EAFNOSUPPORT         (WSO_BASEERR+47) */
#define WSO_EADDRINUSE           (WSO_BASEERR+48)
#define WSO_EADDRNOTAVAIL        (WSO_BASEERR+49)
#define WSO_ENETDOWN             (WSO_BASEERR+50)
/* fine WSO_ENETUNREACH          (WSO_BASEERR+51) */
/* fine WSO_ENETRESET            (WSO_BASEERR+52) */
#define WSO_ECONNABORTED         (WSO_BASEERR+53)
#define WSO_ECONNRESET           (WSO_BASEERR+54)
#define WSO_ENOBUFS              (WSO_BASEERR+55)
/* fine WSO_EISCONN              (WSO_BASEERR+56) */
#define WSO_ENOTCONN             (WSO_BASEERR+57)
#define WSO_ESHUTDOWN            (WSO_BASEERR+58)
/* fine WSO_ETOOMANYREFS         (WSO_BASEERR+59) */
#define WSO_ETIMEDOUT            (WSO_BASEERR+60)
#define WSO_ECONNREFUSED         (WSO_BASEERR+61)
/* fine WSO_ELOOP                (WSO_BASEERR+62) */
/* fine WSO_ENAMETOOLONG         (WSO_BASEERR+63) */
/* fine WSO_EHOSTDOWN            (WSO_BASEERR+64) */
#define WSO_EHOSTUNREACH         (WSO_BASEERR+65)
/* fine WSO_ENOTEMPTY            (WSO_BASEERR+66) */
/* fine WSO_EPROCLIM             (WSO_BASEERR+67) */
/* fine WSO_EUSERS               (WSO_BASEERR+68) */
/* fine WSO_EDQUOT               (WSO_BASEERR+69) */
/* fine WSO_ESTALE               (WSO_BASEERR+70) */
#define WSO_EREMOTE              (WSO_BASEERR+71)

/* ----------------------------------------------------------------- */

/*
 * UDP/TCP socket local flags (locflags) bits.
 * the names are such that the default state is zero 
 */
/* low 16bits are user options - persistant and only modifyable by user 
*/
#define WSO_LF_KEEPALIVE    (1ul<< 0) /* TCP only: send periodic keep alives */
#define WSO_LF_NONAGLE      (1ul<< 1) /* TCP only: don't use Nagle       */
#define WSO_LF_NOPUSH       (1ul<< 2) /* TCP only: don't push on write   */
#define WSO_LF_NOOPT        (1ul<< 3) /* TCP only: don't use tcp options */
#define WSO_LF_REUSEADDR    (1ul<< 4) /* TCP only: reuseaddr if other is TMWT */
#define WSO_LF_REUSEPORT    (1ul<< 5) /* both: reuseaddr if all are reuseport */
#define WSO_LF_OOBINLINE    (1ul<< 6) /* NYI: "urgent" data treated as regular data */
#define WSO_LF_DOASCII      (1ul<< 7) /* TCP only: ick, ick, ick */
#define WSO_LF_WATTCOMPAT   (1ul<< 8) /* old wattsock listen behaviour */
#define WSO_LF_BLOCKING     (1ul<< 9) /* socket is in blocking mode */
#define WSO_LF_NOCHECKSUM   (1ul<<10) /* UDP only: don't checksum outgoing data */
#define WSO_LF_MASK_USER_LF (0x0000fffful)
/* 
** high 16bits are per-connection settings and state 
*/
#define WSO_LF_NOFRAGMENT   (1ul<<16) /* both: peer (doesn't) use DF */
#define WSO_LF_ASYNC_ERROR  (1ul<<17) /* async close or shutdown (eg by ICMP msg)*/
#define WSO_LF_GOT_FIN      (1ul<<18) /* TCP only: */
#define WSO_LF_UPEND_QOS    (1ul<<19) /* tell poll about change of QOS */
#define WSO_LF_RCVD_TSTMP   (1ul<<20) /* TCP only: a timestamp was received in SYN */
#define WSO_LF_RCVD_SCALE   (1ul<<21) /* TCP only: a win-scale was received in SYN */
#define WSO_LF_CANTSENDMORE (1ul<<22) /* both: shutting down */
#define WSO_LF_CANTRECVMORE (1ul<<23) /* both: shutting down */
#define WSO_LF_HOLDLOCAL    (1ul<<24) /* TCP only: ignore NONAGLE for next */
#define WSO_LF_URG_AVAIL    (1ul<<25) /* TCP only: RCVATMARK */
#define WSO_LF_URG_SIGPEND  (1ul<<26) /* TCP only: SIGURG signal needs posting */
#define WSO_LF_USAWCONNECT  (1ul<<27) /* for internal so_async_poll() use */
/* 
** getset_sockopt() options that are not stored in locflags 
*/
#define WSO_LF_SNDTIMEO    0xffff0001ul
#define WSO_LF_RCVTIMEO    0xffff0002ul
#define WSO_LF_TTL         0xffff0003ul /* udp_SetTTL() (but also for TCP) */
#define WSO_LF_TOS         0xffff0004ul /* - */
#define WSO_LF_FIONREAD    0xffff0005ul /* sock_dataready()/sock_bytesready() */
#define WSO_LF_FIONWRITE   0xffff0006ul /* sorta like sock_tbavail() */
#define WSO_LF_SNDBUFSIZE  0xffff0007ul /* sock_tbsize() currently read-only */
#define WSO_LF_SNDBUFUSED  0xffff0008ul /* sock_tbused() currently read-only */
#define WSO_LF_RCVBUFSIZE  0xffff0009ul /* sock_rbsize() currently read-only */
#define WSO_LF_RCVBUFUSED  0xffff000aul /* sock_rbused() currently read-only */
#define WSO_LF_TXTLINELEN  0xffff000bul /* sock_bytesready() - TCP only */
#define WSO_LF_YIELDFUNC   0xffff000cul /* sock_yield(s, fn ) */
#define WSO_LF_PHASETIMEO  0xffff000dul /* sock_timeout(s, int sec) */
#define WSO_LF_IPPROTO     0xffff000eul
#define WSO_LF_ERRNO       0xffff000ful
#define WSO_LF_STATE       0xffff0010ul /* sockstate() et al */
#define WSO_LF_BOUNDPORT   0xffff0011ul /* getsockname */
#define WSO_LF_BOUNDADDR   0xffff0012ul /* getsockname */
#define WSO_LF_PEERPORT    0xffff0013ul /* getpeername */
#define WSO_LF_PEERADDR    0xffff0014ul /* getpeername */
#define WSO_LF_ASYNC_POLL  0xffff0015ul /* internal reserved */
#define WSO_LF_USERTIMER   0xffff0016ul /* USER THING: ip_timer_{init|expired}() */
#define WSO_LF_LINGERTIMEO 0xffff0017ul /* 0 = don't linger */
#define WSO_LF_IFCOUNT     0xffff0018ul

/* flags for xxx_poll() 
** this a superset of UNIX POLL* bits and possible select() fd types.
** so that we have flags with exactly one meaning and no overlap.
*/
#define WSO_ASYSEL_EBADF   (1u <<  0) /* not a valid socket */
#define WSO_ASYSEL_ERROR   (1u <<  1) /* Asynchronous error indication */
#define WSO_ASYSEL_HANGUP  (1u <<  2) /* Connection has been closed */
#define WSO_ASYSEL_CONNECT (1u <<  3) /* Connection initiated now complete */
#define WSO_ASYSEL_ACCEPT  (1u <<  4) /* Ready to accept() a new conn */
#define WSO_ASYSEL_QOS     (1u <<  5) /* Quality-of-Service changed */
#define WSO_ASYSEL_GRPQOS  (1u <<  6) /* Group QOS changed */
#define WSO_ASYSEL_WRITE   (1u <<  7) /* Ready for writing band-zero data */
#define WSO_ASYSEL_WROOB   (1u <<  8) /* Ready for writing band-!0 data */
#define WSO_ASYSEL_READ    (1u <<  9) /* Ready for reading band-zero data */
#define WSO_ASYSEL_RDOOB   (1u << 10) /* Ready for reading band-!0 data */


/* =============================================================== */
/*                  types needed for hostaddr/port                 */
/* =============================================================== */

#if !defined(INADDR_TYPES_DEFINED)
/* POSIX/SUS mandated types for IPV4 hostaddr and port */
/* See http://www.opengroup.org/onlinepubs/7908799/xns/arpainet.h.html */
typedef unsigned short in_port_t; /* An unsigned integral type of exactly 16 bits */
typedef unsigned long  in_addr_t; /* An unsigned integral type of exactly 32 bits */
#define INADDR_TYPES_DEFINED
#endif

/* ***************************************************************** */
/* *************** PART I: SOCKET FUNCTION PRIMITIVES ************** */
/* ***************************************************************** */

#ifdef __cplusplus
extern "C" {
#endif

extern int sock_create(void *s, int ip_proto);
extern int sock_bind(void *s, const in_addr_t *laddr, const in_port_t *lport);
extern int sock_connect(void *s, const in_addr_t *raddr, const in_port_t *rport);
extern int sock_accept(void *s, void *news, in_addr_t *raddr, in_port_t *rport);
extern int sock_listen(void *s, int qlen);
extern int sock_destroy(void *s);
extern int sock_shutdown( void * s, int how );
extern int sock_send_to( void *s, const void *dp, unsigned int len, int flags,
                   const in_addr_t *peer_addr, const in_port_t *peer_port );
extern int sock_recv_from( void *s, void *buf, unsigned int buflen, int flags,
                    in_addr_t *peer_addr, in_port_t *peer_port );
extern int sock_opt_getset(void *s, unsigned long opt, const int *newval, int *oldval);
extern int sock_poll(void *s, int events, unsigned long timeout);

#ifdef __cplusplus
}
#endif

/* ***************************************************************** */
/* *************** PART II: SOCKET FXN COMPATIBILITY *************** */
/* ***************************************************************** */

/* ================================================================= */
/*                 PART 2.1: skt function macro-izer helpers         */
/* ================================================================= */

#ifdef __cplusplus     /* these functions are all new (May 2001) and were */
extern "C" {           /* created to support backwards-compatibility sockio */
#endif

extern int __IPPROTO_TCP__; /* (6)  in pctcp.c - forces linkage */
extern int __SOCK_STREAM__; /* (1)  in pctcp.c - forces linkage */
extern int __IPPROTO_UDP__; /* (17) in pcudp.c - forces linkage */
extern int __SOCK_DGRAM__ ; /* (2)  in pcudp.c - forces linkage */

/* protohandler is totally obsolete - using it will cause failure */
typedef int (*ProtoHandler) (void *sock, const void *data, unsigned int len,
                             const void *tcp_phdr, const void *udp_hdr);

/* dummy xxx_Socket for backwards compatibility with TinyTCP (min size = 8) */
typedef struct { char undoc[16]; } tcp_Socket, udp_Socket, sock_type;

/* the tinyTCP-compatible open/listen functions. */
extern int sock_wattlisten(int ip_proto, void *s, in_port_t lport, in_addr_t raddr, 
                           in_port_t rport, ProtoHandler handler, 
                           unsigned int timeout, ... );
extern int sock_wattopen(int ip_proto, void *s, in_port_t lport, 
                          in_addr_t raddr, in_port_t rport, ProtoHandler handler, ... );

/* sock_set_yieldcb() is what sock_yield() used to be called */
extern int sock_set_yieldcb( void *s, void (*fn)(void) );      

/* sock_get{peer|sock}name are sane replacements for _get{peer|sock}name(). */
extern int sock_getpeername( void *s, in_addr_t *raddr, in_port_t *rport );     
extern int sock_getsockname( void *s, in_addr_t *laddr, in_port_t *lport );     

/* sock_condwait() is the base for _ip_delay_*() & sock_wait_*() macros */
extern int sock_condwait( void *s, int condition, int waitsecs,        
                          int (*blocking_hook)(void *), int *statusptr, 
                          int _ip_delayX_compatible );

/* sock_str*() are not really 's' fxns, but the macros that use them are */
extern const char *sock_strerror(int soerrno);
extern const char *sock_strstate(int sostate);

/* simplified access to socket options */
extern int sock_opt_get(void *s, unsigned long opt);
extern int sock_opt_set(void *s, unsigned long opt, int val);

#ifdef __cplusplus
}
#endif

/* ================================================================= */
/*                 PART 2.2: unmacroizable skt fxn WRAPPERS          */
/* ================================================================= */

#ifdef __cplusplus  /* these functions are all compatibility wrappers */
extern "C" {        /* around the primitives listed above, but are */
#endif              /* too complex to turn into macros. */

extern int sock_putc( void *s, int ch );
extern int sock_puts( void *s, const char *text );
extern int sock_fastread( void *s, void *dp, unsigned int len );
extern int sock_preread( void *s, void *dp, unsigned int len );
extern int sock_read( void *s, void *dp, unsigned int len );
extern int sock_getc( void *s );
extern int sock_gets( void *s, char *text, unsigned int len );

extern int sock_established(void *s);
extern int sock_sselect (void *s, int waitstate);

#ifdef __cplusplus
}
#endif

/* ================================================================= */
/*                  PART 2.3: macroizable skt-functions              */
/* ================================================================= */

#if !defined(__KERN_TARGET__)
/* This section defines macros for obsolete sock io calls (backwards-compat).
** The kernel (which includes sockio) doesn't use any of them. 
*/
#  define tcp_listen(s,lp, ra, rp, hdlr, to) sock_wattlisten(__IPPROTO_TCP__,s,lp,ra,rp,hdlr,to)
#  define tcp_open(s,lp, ra, rp, hdlr)       sock_wattopen(__IPPROTO_TCP__,s,lp,ra,rp,hdlr)
#  define udp_listen(s,lp, ra, rp, hdlr)     sock_wattlisten(__IPPROTO_UDP__,s,lp,ra,rp,hdlr,0)
#  define udp_open(s,lp, ra, rp, hdlr)       sock_wattopen(__IPPROTO_UDP__,s,lp,ra,rp,hdlr)

#  define sock_abort(__s)                    sock_destroy(__s)
#  define sock_close(__s)                    sock_shutdown(__s, 1)

#  define sock_write(s, dp, l)               sock_send_to(s,dp,l,WSO_MSG_WAITALL,0,0)
#  define sock_fastwrite(s, dp, l)           sock_send_to(s,dp,l,WSO_MSG_DONTWAIT,0,0)
#  define sock_enqueue(s, dp, l)             sock_send_to(s,dp,l,WSO_MSG_WAITALL,0,0)
#  define sock_flush(s)                      sock_send_to(s,0,0,WSO_MSG_WAITALL,0,0)
#  define sock_noflush(s)                    sock_opt_set(s, WSO_LF_HOLDLOCAL, 1)
#  define sock_flushnext(s)                  sock_opt_set(s, WSO_LF_HOLDLOCAL, 0)
#  define sock_tbused(s)                     sock_opt_get(s, WSO_LF_SNDBUFUSED )
#  define sock_tbsize(s)                     sock_opt_get(s, WSO_LF_SNDBUFSIZE )
#  define sock_tbleft(s)                     (sock_tbsize(s) - sock_tbused(s))

#  define sock_recv(s,dp,l)                  sock_fastread(s,dp,l)
#  define sock_recv_init(s,rb,rbl)           fprintf(stderr,"sock_recv_init(%p,%p,%u) is obsolete\n",s,dp,rbl)
#  define sock_rbsize(s)                     sock_opt_get(s, WSO_LF_RCVBUFSIZE )
#  define sock_rbused(s)                     sock_opt_get(s, WSO_LF_RCVBUFUSED )
#  define sock_rbleft(s)                     (sock_rbsize(s) - sock_rbused(s))
#  define sock_dataready(s)                  sock_opt_get(s, WSO_LF_FIONREAD)
#  define sock_bytesready(s)                 sock_opt_get(s, WSO_LF_TXTLINELEN)

#  define udp_SetTTL(s,ttl)                  sock_opt_set(s, WSO_LF_TTL, ttl )
#  define sock_get_errno(s)                  sock_opt_get(s, WSO_LF_ERRNO )
#  define sock_get_stateno(s)                sock_opt_get(s, WSO_LF_STATE)
#  define sockerr(s)                         sock_strerror(sock_get_errno(s))
#  define sockstate(s)                       sock_strstate(sock_get_stateno(s))
#  define sock_mode(s,__mode) do { int __m__ = (__mode); \
     /* ** NOTE ** sock_mode() is maintained for compatibility purposes only. \
      * It does not work correctly (and never did) in as far as reverting to  \
      * the original mode is concerned - Due to #define overlaps (all the     \
      * default states were/are defined to be zero) sock_mode() cannot revert \
      * to the default state since it cannot determine which default          \
      * state is to be reverted to. Use sock_opt_getset() instead.            \
     */                                                                       \
     if (__m__ == 0) { fprintf(stderr,"sock_mode() called with mode == 0\n"); exit(255); } \
     if (__m__ & TCP_MODE_ASCII)   sock_opt_set(s, WSO_LF_DOASCII, 1); \
     if (__m__ & TCP_MODE_NONAGLE) sock_opt_set(s, WSO_LF_NONAGLE, 1); \
     if (__m__ & UDP_MODE_NOCHK)   sock_opt_set(s, WSO_LF_NOCHECKSUM, 1); \
     break; } while (1)
#  define sock_yield(s,fn)                   sock_set_yieldcb(s,fn)
#  define ip_timer_init(s, secs)             sock_opt_set(s, WSO_LF_USERTIMER, secs)
#  define ip_timer_expired(s)                sock_opt_get(s, WSO_LF_USERTIMER)  
#  define tcp_tick(s) __sock_dummy_ifunc__((sock_poll(s, 0, 0UL)) ? (0) : ( s != NULL ))
#  define tcp_established(s)                 sock_established(s)
#  define _ip_delay0(s, secs, fn, statusptr) sock_condwait(s,SOCKESTABLISHED,secs,fn,statusptr,1)
#  define _ip_delay1(s, secs, fn, statusptr) sock_condwait(s,SOCKDATAREADY,secs,fn,statusptr,1)
#  define _ip_delay2(s, secs, fn, statusptr) sock_condwait(s,SOCKCLOSED,secs,fn,statusptr,1)

#endif /* !defined(__KERN_TARGET__) */

/* ***************************************************************** */
/* ************ PART III: NON-SOCKET FUNCTION PRIMITIVES *********** */
/* ***************************************************************** */

#ifdef __cplusplus
extern "C" {
#endif

const char *sock_ntoa( unsigned long ipaddr, char *buffer, unsigned int buflen );
int         sock_aton( const char *cp, in_addr_t *ipaddr );
in_addr_t   sock_addr( const char *cp );
in_addr_t   sock_gethostid(void);
int         sock_sethostid(in_addr_t id);
int         sock_gethostname( char *buffer, unsigned int buflength );
int         sock_sethostname( const char *name, unsigned int namelen );
int         sock_getdomainname( char *buffer, unsigned int buflength );
int         sock_setdomainname( const char *name, unsigned int namelen );
in_addr_t   sock_old_resolv( const char *hostname, char *ntoa_buffer );
int         sock_gethostbyname( const char *hname, in_addr_t *addrlist, int numaddrs);
int         sock_gethostbyaddr( in_addr_t ina, char *hnamebuf, unsigned int buflen);
int         sock_mcast_group_op(in_addr_t ina,int op);
int         sock_ping( in_addr_t *inaP, int *ttlP, int *tosP, 
                       unsigned int *bufsizeP, unsigned long *timeoutP, int df,
                       const void *pattern, unsigned int pattlen );

unsigned long  sock_htonl(unsigned long val);
unsigned long  sock_ntohl(unsigned long val);
unsigned short sock_htons(unsigned short val);
unsigned short sock_ntohs(unsigned short val);

extern unsigned long  *__sock_dummy_lpval__(unsigned long  what_to_preset);
extern unsigned short *__sock_dummy_spval__(unsigned short what_to_preset);
extern int            *__sock_dummy_ipval__(int what_to_preset);
extern int             __sock_dummy_ifunc__(int what_to_preset);

#ifdef __cplusplus
}
#endif

/* ***************************************************************** */
/* ************* PART IV: NON-SOCKET FXN COMPATIBILITY ************* */
/* ***************************************************************** */

#if !defined(__KERN_TARGET__)
#define intel(__val)           sock_htonl(__val)
#define intel16(__val)         sock_htons(__val)

#define tcp_cbreak(x)          __sock_dummy_ifunc__(0)
#define tcp_cbrk(x)            __sock_dummy_ifunc__(0)
#define wathndlcbrk            (*__sock_dummy_ipval__(0))
#define watcbroke              __sock_dummy_ifunc__(0)
#define set_cbreak(x)          __sock_dummy_ifunc__(x)

#define Random(a,b)            (((unsigned)rand())%((a)<(b))?((b)-(a)):((a)-(b)))
#define set_timeout(x)         (((unsigned long)clock()) + ((((unsigned long)(x))*CLK_TCK)/1000u))
#define chk_timeout(x)         ((x) && (((unsigned long)x) > ((unsigned long)clock())))
#define cmp_timeout(x,y)       (((x)==(y)) ? (0) : (((x)<(y)) ? -1 : +1))
#define set_ttimeout(t)        set_timeout(((((unsigned long)(t))*182ul)/10ul))

#define _gethostid()           sock_gethostid()
#define _sethostid(x)          sock_sethostid(x)
#if 0 /* collides with BSD name space (different handling if buf too small)*/
#define getdomainname(b,l)     sock_getdomainname(b,l)
#define setdomainname(b,l)     sock_getdomainname(b,l)
#define gethostname(b,l)       sock_gethostname(b,l)
#define sethostname(b,l)       sock_gethostname(b,l)
#endif

#define aton(str)              sock_addr(str)
#define isaddr(str)            sock_aton(str,((in_addr_t *)0))
#define _inet_ntoa(buf,ina)    sock_ntoa(dw,buf,sizeof("255.255.255.255"))
#define _inet_addr(str)        sock_addr(str)
#define _chk_socket(s)         (sock_poll(s,0,0)==0)
#define lookup_host(h,ap)      sock_old_resolv(h,ap)
#define resolve(h)             sock_old_resolv(h,NULL)

#define my_ip_addr             (*__sock_dummy_lpval__(sock_gethostid()))
#define sock_delay             (*__sock_dummy_ipval__(30))
#define multihomes /*sheesh!*/ (*__sock_dummy_ipval__(0))
#define mtu                    (*__sock_dummy_spval__(576))
#define mss                    (*__sock_dummy_spval__(576-(64+64)))

#define _survivebootp          (*__sock_dummy_ipval__(1))
#define _survivedhcp           (*__sock_dummy_ipval__(1))
#define _surviverarp           (*__sock_dummy_ipval__(1))

#define ctrace_on              (*__sock_dummy_ipval__(0))
#define debug_on               (*__sock_dummy_ipval__(0))
#define block_tcp              (*__sock_dummy_ipval__(0))
#define block_udp              (*__sock_dummy_ipval__(0))
#define block_ip               (*__sock_dummy_ipval__(0))
#define block_icmp             (*__sock_dummy_ipval__(0))
#define _watt_no_config        (*__sock_dummy_ipval__(0))
#define tcp_set_debug_state(x) debug_on = (x)

#define MAX_COOKIES            0
#define last_cookie            (*__sock_dummy_ipval__(0))
#define cookies                (__sock_dummy_dfunc(0))
#define MAX_NAMESERVERS        0
#define def_nameservers        (__sock_dummy_dfunc(0))
#define dns_timeout            (*__sock_dummy_ipval__(30))
#define last_nameserver        (*__sock_dummy_ipval__(0))

#define tcp_config(__fn)       __sock_dummy_ifunc__(1)
#define sock_init()            __sock_dummy_ifunc__(1)
#define sock_exit()            __sock_dummy_ifunc__(1)
#define dbug_init()            __sock_dummy_ifunc__(1)
#define dbug_exit()            __sock_dummy_ifunc__(1)
#define init_misc()            __sock_dummy_ifunc__(1)
#define sock_debugdump(s)      ((void)0) /* nothing - returned void */
#define inchksum               /* not supported in userland */

/* ------------------------------------------------------------------ */

#if defined(I_STOLE_TINYTCP_AND_CALLED_IT_WATTCP)

/* version numbers for what-a-poo-tcp compatibility */
#define WATTCP_MAJOR_VER  3u 
#define WATTCP_MINOR_VER  0u
#define WATTCP_DEVEL_REL  0u
#define WATTCP_VER \
        ((WATTCP_MAJOR_VER<<8)+(WATTCP_MINOR_VER<<4)+WATTCP_DEVEL_REL)

/* baloney anyone? */
/* extern const char *_watt_get_verstr(int which_str); */
#define wattcpCopyright       "" /* blah, blah, yadda, yadda, more lies */
#define wattcpVersion         "" /* version/date */
#define wattcpCapabilities()  "" /* what's built in */

#endif

/* ------------------------------------------------------------------ */

#define sock_wait_established(s,seconds,fn,statusptr) \
        {                                             \
           if (_ip_delay0 (s,seconds,fn,statusptr))   \
              goto sock_err;                          \
        }
#define sock_wait_input(s,seconds,fn,statusptr)       \
        {                                             \
           if (_ip_delay1 (s,seconds,fn,statusptr))   \
              goto sock_err;                          \
        }
#define sock_tick(s, statusptr)                       \
        {                                             \
           if (!tcp_tick(s)) {                        \
              if (statusptr) *statusptr = 1;          \
              goto sock_err;                          \
           }                                          \
        }
#define sock_wait_closed(s,seconds,fn,statusptr)      \
        {                                             \
           if (_ip_delay2(s,seconds,fn,statusptr))    \
              goto sock_err;                          \
        }

/* ------------------------------------------------------------------ */

#define outs(str)              fprintf(stderr,"%s",str)
#define outsnl(str)            fprintf(stderr,"%s\n",str)
#define outsn(str,n)           fprintf(stderr,"%*s",n,str)
#define outhex(c)              fprintf(stderr,"%02x",c);
#define outhexes(str,n)        { int __c = (n); const char *__s = (str);  \
                                 while (__c--) outhex(__s++); } ((void)0)
#define rip(__s)               { /* stupidity has no limits */            \
                                 char *__q0__ = __s;                      \
                                 char *__q1__ = strchr(__q0__,'\r');      \
                                 char *__q2__ = strchr(__q0__,'\n');      \
                                 if (__q2__ && __q2__ < __q1__)           \
                                  __q1__ = __q2__;                        \
                                 if (__q1__) *__q1__ = '\0';              \
                               } ((void)0)
#define psocket(s)             { /* just how dumb does it get? */         \
                                 in_addr_t _raddr; in_port_t _rport;      \
                                 if (sock_getpeername(__s,                \
                                             &_raddr, &_rport)==0)        \
                                   fprintf(stderr,"[%s:%u]",              \
                                      sock_ntoa(_raddr,NULL,0), _rport ); \
                               } ((void)0)

/* ------------------------------------------------------------------ */

#define multi_to_eth(__i,__e)  {                                          \
                                 in_addr_t _ina_ = (__i);                    \
                                 unsigned char *_eth_ =                   \
                                              (unsigned char *)(__e);     \
                                 _eth_[0] = (BYTE)0x01;                   \
                                 _eth_[1] = (BYTE)0x00;                   \
                                 _eth_[2] = (BYTE)0x5E;                   \
                                 _eth_[3] = (BYTE)((_ina_ >> 16) & 0x7F); \
                                 _eth_[4] = (BYTE)((_ina_ >>  8) & 0xFF); \
                                 _eth_[5] = (BYTE)((_ina_ >>  0) & 0xFF); \
                               } ((void)0)
#define is_multicast(ina)      ((ina & 0xF0000000ul) == 0xE0000000ul)
#define _multicast_on          (*__sock_dummy_ipval__(1))
#define join_mcast_group(ina)  sock_mcast_group_op(ina, +1);
#define leave_mcast_group(ina) sock_mcast_group_op(ina, -1);

/* ------------------------------------------------------------------ */
/* Old sick types. DO NOT TYPEDEF (user may want to '#undef' them). */

/* WatTCP sick types */
#define WORD     unsigned short        /* 16 bits   */
#define DWORD    unsigned long         /* 32 bits   */

/* TinyTCP sick types */
#define word     unsigned short        /* 16 bits   */
#define longword unsigned long         /* 32 bits   */

#endif /* !defined(__KERN_TARGET__) */

/* ================================================================= */
/*             Oh, Toto, I don't think I'm in Kansas anymore.        */
/* ================================================================= */

#endif /* __TSIO_VER__ */
