#include <stdio.h>  /* fprintf() */
#include <conio.h>  /* kbhit() */
#include <dos.h>    /* delay() */
#include <io.h>     /* write() */
#include <stdlib.h> /* _exit(), isatty(), write() */
#include <fcntl.h>  /* O_BINARY */
#include <signal.h> /* SIGBREAK/SIGINT */
#include <time.h>   /* clock() */
#include <string.h> /* str*(), mem*() */
#include <limits.h> /* ULONG_MAX */
#include <errno.h>  /* strerror() */
#include "userland.h"

#if defined(__TURBOC__)
unsigned _Cdecl _stklen = 4464;
#elif defined(__DJGPP__)
#include <unistd.h>
#endif

/* TEST_LEVELs:
** 0 = base only, no msgs (simply way to determine minimal lib size)
** 1 = base only
** 2 = base+udp, 
** 3 = base+udp+dns, 
** 4 = base+udp+dns+tcp 
*/
#define TEST_LEVEL 4

/* ====================================================================== */

#if (TEST_LEVEL >= 2)
static char buffer[512];
#endif

#if (TEST_LEVEL >= 1)
static int got_brk_signal = 0;
static void sig_handler(int sig) { sig = sig; got_brk_signal = 1; }
#endif

/* ====================================================================== */
/*             stack boot (conf load/dhcp/bootp/arp etc) test             */
/* ====================================================================== */

#if (TEST_LEVEL >= 1)
static int __hostid_test(const char *appname, const char *testname, 
                         const char *arg)
{
  if (arg)
  {
    fprintf(stderr, "Syntax: %s %s without further arguments to\n"
    "get the host IP address. This tests confload/dhcp/bootp/arp etc\n"
    "To observe changes to the host ID use '%s -bgloop 1'\n"
    "", appname, testname, appname );
  }
  else
  {
    fprintf(stderr,"Initializing stack with sock_gethostid()...\n");
    fprintf(stderr,"Host ID is %s\n", sock_ntoa(sock_gethostid(),NULL,0) );
  }
  return 0;
}
#endif

/* ====================================================================== */
/*                   basic background/idle processing tests               */
/* ====================================================================== */

#if (TEST_LEVEL >= 1)
#include "dpmi.h"

static void __clr_kbd(void) /* clear keybd buffer without using interrupts */
{
#if defined(__DPMI32_PEEKPOKE_DEFINED__)
  _farpokew( _dos_ds, 0x41a, _farpeekw(_dos_ds, 0x41c) );
#elif defined(__DPMI32__)
  unsigned short headtail[2];
  dosmemget( ((0x0040ul<<4)|0x001Aul), sizeof(headtail), &headtail[0] );
  headtail[0] = headtail[1];
  dosmemput( &headtail[0], sizeof(headtail), ((0x0040ul<<4)|0x001Aul) );
#else
  *((unsigned short _far *)0x41c) = *((unsigned short _far *)0x41a);
#endif
}

static int __chk_kbd(void) /* check for keypress without using interrupts */
{
#if defined(__DPMI32_PEEKPOKE_DEFINED__)
  return (_farpeekw(_dos_ds, 0x41a) != _farpeekw(_dos_ds, 0x41c));
#elif defined(__DPMI32__)
  unsigned short w[2];
  dosmemget(0x41a, sizeof(w), &w[0]);
  return (w[0] != w[1]);
#else
  return (*((unsigned short _far *)0x41a) != *((unsigned short _far *)0x41c));
#endif
}
#endif

/* background processing test */
#if (TEST_LEVEL >= 1)
static int __bgloop_test(const char *appname, const char *testname, 
                     const char *numstr)
{
  int num = (numstr ? *numstr : 0);

  __clr_kbd();
  if (num == '1')
  {
    in_addr_t ina;
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning delay() test... (press ^C to stop)\n");
    ina = sock_gethostid();
    while (!got_brk_signal && !__chk_kbd()) 
    {
      in_addr_t newina = sock_gethostid();
      if (newina != ina)
      {
        fprintf(stderr,"Host ID changed. New host ID is %s\n",
                                            sock_ntoa(newina,NULL,0) );
        ina = newina;
      }
      delay(1000);
    }      
    fprintf(stderr,"End delay() test...\n");
  }
  else if (num == '2')
  {
    #if defined(__TURBOC__) || defined(__WATCOMC__)
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning time() test... (press ^C to stop)\n");
    while (!got_brk_signal && !__chk_kbd())
      time(NULL);
    #else
    fprintf(stderr, "Test 2 is only meaningful for TurboC/WatcomC\n");
    #endif
  }
  else if (num == '3')
  {
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning kbhit() test... (press any key to stop)\n");
    while (!got_brk_signal)
    {
      if (!kbhit())
        num = 0;
      else if (!num)
        break;
      else if (!getch())
        getch();
    }
  }
  else if (num == '4')
  {
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning Int21/ah=0Bh test... (press any key to stop)\n");
    while (!got_brk_signal)
    {
      char gotkbhit = 0;
#if defined(__DPMI32__)
      __dpmi_regs r;
      r.x.ax = 0x0B00;
      __dpmi_int(0x21, &r);
      gotkbhit = r.h.al;
#else
      __asm mov ah,0bh
      __asm int 21h
      __asm mov gotkbhit,al
#endif
      if (!gotkbhit)
        num = 0;
      else if (!num)
        break;
      else if (!getch())
        getch();
    }
  }
  else if (num == '5')
  {  
#if defined(__DPMI32__)
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning video bios test... (press any key to stop)\n");
    while (!got_brk_signal && !__chk_kbd())        
    {
      __dpmi_regs r;
      r.h.ah = 0x03;
      r.h.bh = 0;
      __dpmi_int(0x10, &r);
    }
#else
    fprintf(stderr,"Int 10h is not handled by real mode targets\n");
#endif
  }
  else if (num == '6')
  {
#if defined(__DPMI32__)
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning __dpmi_yield() test... (press any key to stop)\n");
    while (!got_brk_signal && !__chk_kbd())
      __dpmi_yield();
#else
    fprintf(stderr,"Int 2Fh/ax=1680h is not handled by real mode targets\n");
#endif
  }
  else if (num == '7')
  {
    __hostid_test(appname,testname,0);
    fprintf(stderr,"Beginning file I/O yield test... (press any key to stop)\n");
    while (!got_brk_signal && !__chk_kbd())
    {
      const char *spin = "|\r/\r-\r\\\r";
      write(2,&spin[((++num)&3)<<1],2);
    }
  }
  else
  {
    fprintf(stderr,"Syntax: %s %s <num>\n" 
    "  This test determines the efficiency of the various interrupt-time\n"
    "  and/or idle function background handlers. The test initializes\n"
    "  the stack and then goes into an endless loop that does not call any\n"
    "  any network library functions. PINGs etc should STILL respond ok,\n"
    "  but responsiveness will depend on which <num> is tested.\n"
    ""
    "\n"
    "<num> may be one one of the following:\n"
    "\n"
    "  1   libc delay() (tests fxn redirection)\n"
    "  2   libc time()  (tests fxn redirection) - WatcomC and TurboC only.\n"
    "  3   libc kbhit() (tests fxn redirection)\n"
    "  4   raw  kbhit [Int 21h/ah=0Bh] (tests Int28h and Int21h handling)\n"
    "  5   Int 10h/ah=03h (tests Int10h handling) - 32bit targets only\n"
    "  6   Int 2Fh/ax=1680h (DPMI yield) - 32bit targets only\n"
    "  7   libc write() (tests Int 28h handling)\n"
    "\n"
    "  Note 1: All DPMI targets trap Int31h/ax=0300h/bl=10h,21h,15h,2Fh\n"
    "  Note 2: All targets trap Int28h (DOS Idle)\n"
    "  Note 3: Many targets redirect typical yield functions. eg delay()\n"
    "          sleep(), usleep(), kbhit(), __dpmi_yield() etc.\n"
    "  Note 4: WatcomC and TurboC real-mode targets also yield when the\n"
    "          application makes a libc call that (indirectly) retrieves\n"
    "          the time-of-day from DOS.\n"
    "", appname, testname );
  }
  __clr_kbd();
  return 0;
}
#endif /* (TEST_LEVEL >= 1) */

/* ====================================================================== */
/*                                 DNS tests                              */
/* ====================================================================== */

#if (TEST_LEVEL >= 3)
static int __dns_test(const char *appname, const char *testname, 
                      const char *name)
{
  in_addr_t addrlist[10];
  int i, numaddrs = 0;
  if (!name)
  {
    fprintf(stderr,"No argument specified. Defaulting to use of local IP address.\n");
    addrlist[0] = sock_gethostid();
    numaddrs = 1;
    fprintf(stderr,"local IP address is %s\n", sock_ntoa(addrlist[0],NULL,0) );
  }
  else if (*name == '?')
  {
     fprintf(stderr,"Syntax: %s %s [<hostname|hostaddr>]\n"
     "\tDoes a backwards and forwards resolve for <hostname|address>.\n"
     "\tUses this host's IP address if no argument is specified.\n"
     "", appname, testname );
  }
  else if (sock_aton( name, &addrlist[0] ))
  {
    numaddrs = 1;
  }
  else
  {
    fprintf(stderr,"Attempting to resolve '%s'...\n", name);
    numaddrs = sock_gethostbyname( name, &addrlist[0],
                    sizeof(addrlist)/sizeof(addrlist[0]));
    if (numaddrs < 0)
      fprintf(stderr,"Resolve failed: error %d\n", numaddrs );
    else if (numaddrs == 0)
      fprintf(stderr,"Resolve failed: '%s' does not exist.\n", name );
    else
    {
      for (i = 0; i < numaddrs; i++)
        fprintf(stderr,"'%s' => %s\n", name, sock_ntoa(addrlist[i],NULL,0) );
    }
  }
  if (numaddrs > 0)
  {
    for (i = 0; i < numaddrs; i++)
    {
      int rc;
      fprintf(stderr,"Attempting to reverse-resolve %s...\n", 
              sock_ntoa(addrlist[i],NULL,0) );
      rc = sock_gethostbyaddr(addrlist[i], buffer, sizeof(buffer));
      if (rc > 0)
        fprintf(stderr,"%s: %s\n", sock_ntoa(addrlist[i],NULL,0), buffer );
      else if (rc == 0)
        fprintf(stderr,"%s: (no name found)\n", sock_ntoa(addrlist[i],NULL,0) );
      else 
        fprintf(stderr,"%s: (error %d)\n", sock_ntoa(addrlist[i],NULL,0), rc );
    }
  }
  appname = appname;
  return 0;
}
#endif

/* ====================================================================== */
/*                          more DNS tests                                */
/* ====================================================================== */

#if (TEST_LEVEL >= 3)
static int __name_test(const char *appname, const char *testname, 
                       const char *arg)
{
  if (arg)
  {
    fprintf(stderr,"Syntax: %s %s without further arguments to\n"
    "\tGets the hostname and domainname.\n"
    "\tThis test is meaningful only if no host/domain name was\n"
    "\tspecified in the .cfg, ie the hostname/domain name has to\n"
    "\tbe determined from the network.\n"
    "\tThe test is run twice, and each pass is timed.\n"
    "\tThe second pass should retrieve the result from a cache.\n"
    "", appname, testname );
  }
  else
  {
    int i;

    for (i = 0; !got_brk_signal && i < 2; i++)
    {
      unsigned long elapsed, start = (unsigned long)clock();

      if (sock_gethostname( buffer, sizeof(buffer) ) != 0)
      {
        fprintf(stderr,"Unable to obtain host name\n");
        break;
      }
      else
        fprintf(stderr,"Host name: '%s'\n",buffer);
      if (sock_getdomainname( buffer, sizeof(buffer) ) != 0)
        fprintf(stderr,"Unable to obtain domain name\n");
      else
        fprintf(stderr,"Domain name: '%s'\n",buffer);
   
      elapsed = ((clock()-start)*10000u) / (((unsigned)CLOCKS_PER_SEC) * 10u);
      fprintf(stderr,"Lookup time: %lu.%03lu secs\n", 
                       elapsed/1000u, elapsed%1000u );
    }
  }
  return 0;
}
#endif

/* ====================================================================== */
/*                 function backwards-compatibility test                  */
/* ====================================================================== */

#if (TEST_LEVEL >= 4)
static const char *__parse_url( const char *url, 
                                char *hostnamebuf, int hostnamebuflen,
                                in_port_t *portP )
{
  if (!url)
    return NULL;

  if (*url == '?')
    return NULL;

  if (strncmp(url,"http://",7)!=0)
  {
    fprintf(stderr,"The url '%s' does not begin with 'http://'\n",url);
    return NULL;
  }

  url += 7;
  while (*url && *url != '/' && *url != ':')
  {
    if (hostnamebuflen < 2)
    {
      fprintf(stderr,"Hostname is too long for internal buffer\n");
      return NULL;
    }
    *hostnamebuf++ = *url++;
    hostnamebuflen--;
  }
  *hostnamebuf = '\0';

  *portP = 80;
  if (*url == ':')
  {
    url++;
    if (*url >= '0' && *url <= '9')
    {
      unsigned int port = 0;  
      while (*url >= '0' && *url <= '9')
      {
        if (port > (0xffff/10) || (port * 10) < port)
        {
          port = 0;
          break;
        }
        port = (port * 10) + (*url++ - '0');
      }
      if (port == 0)
      {
        fprintf(stderr,"Bad port number\n");
        return NULL;
      }
      *portP = (in_port_t)port;
    }
  }
  if (*url && *url != '/')
  {
    fprintf(stderr,"Malformed URL\n");
    return NULL;
  }
  return url;
}
#endif /* (TEST_LEVEL >= 4) */

#if (TEST_LEVEL >= 4)
static int __htget_test(const char *appname, const char *testname, 
                        const char *url)
{
  tcp_Socket sock;
  const char *filename;
  int len, status = 0;
  in_port_t port;
  in_addr_t host;

  filename = __parse_url(url, buffer, sizeof(buffer), &port );
  if (!filename)
  {
    fprintf(stderr,"Syntax: %s %s <url>\n"
         "\teg %s %s http://www.de.freebsd.org/handbook/history.html\n"
         "\tor %s %s http://www.de.freebsd.org\n"
         "\t   If no filename is specified (not even a single slash)\n"
         "\t   HTTPGET will make a 'HEAD' request\n"
         "\tThe contents of <url> are dumped to stdout.\n"
         "%s %s is a test of:\n"
         "\ta) the TCP stack\n"
         "\tb) IP packet defragmentation and\n"
         "\tc) backwards-compatibility API\n"
         "", appname, testname, appname, testname, 
             appname, testname, appname, testname );
    return 3;
  }

  fprintf(stderr,"Attempting to resolve '%s'...\n", buffer );
  host = lookup_host(buffer,NULL);
  if (!host)
  {
    fprintf(stderr,"Unable to resolve '%s'\n", buffer );
    return 3;
  }

  fprintf(stderr,"Creating TCP socket to %s:%u...\n", sock_ntoa(host,NULL,0), port );
  if (!tcp_open(&sock,0,host,port,NULL))
  {
    fprintf(stderr,"tcp_open() failed\n");
    return (3);
  }

  fprintf(stderr,"Waiting for connect establishment to %s:%u...\n", buffer, port);
  fflush (stdout);
  sock_wait_established (&sock,sock_delay,NULL,&status);

  if (*filename)
    len = sprintf(buffer,"GET %s HTTP/0.9\r\n", filename );
  else
    len = sprintf(buffer,"HEAD / /HTTP/0.9\r\n" );
                  
  fprintf(stderr,"Connected. Sending request '%*.*s'...\n", len-2, len-2, buffer );
  
  sock_write( &sock, buffer, len );
  /* sock_close (&sock);  */ /* close sending side.... */

  fprintf(stderr,"Beginning read/wait loop...\n" );
  setmode(1,O_BINARY);
  while (!got_brk_signal)
  {
    sock_wait_input (&sock,sock_delay,NULL,&status);
    len = sock_fastread (&sock,buffer,sizeof(buffer));
    /* fprintf( stderr, "sock_fastread() => %d\n", len ); */
    if (len > 0)
      write(1,buffer,len);
  }         
  setmode(1,O_TEXT);

sock_err:
  if (!got_brk_signal)
  {
    switch (status)
    {
      case 1 : fprintf(stderr,"\nRemote host closed connection\n");
               break;
      case -1: fprintf(stderr,"\ntimeout/abort\n");
               break;
    }
  }

  fprintf(stderr,"Destroying socket...\n" );
  sock_destroy(&sock);
  fprintf(stderr,"Socket destroyed.\n" );
  return (status);
}
#endif /* (TEST_LEVEL >= 4) */

/* ====================================================================== */
/*                              FINGER clients                            */
/* ====================================================================== */

#if (TEST_LEVEL >= 4)
static in_port_t __parse_finger_arg(const char *appname, const char *testname,
                                  const char *arg, const char *descrip,
                                  char *userbuf, unsigned int userbuflen,
                                  char *servbuf, unsigned int servbuflen )
{
  in_port_t port = 0;

  if (arg && strchr(arg,'@'))
  {
    unsigned int pos;
    const char *oldarg = arg;

    pos = 0;
    userbuflen--;
    while (*arg && *arg != '@' && pos < userbuflen)
      userbuf[pos++] = *arg++;
    userbuf[pos] = '\0';

    pos = 0;
    servbuflen--;
    if (*arg != '@')
    {
      fprintf(stderr,"Username in %s is too long\n", oldarg );
      return 0;
    }
    else
    {
      arg++;
      while (*arg && *arg != ':' && pos < servbuflen)
        servbuf[pos++] = *arg++;
      if (!*arg)
        port = 79;
      else if (*arg == ':')
      {
        arg++;
        while (*arg >= '0' && *arg <= '9')
        {
          if (port > (0xffffu/10) || ((port * 10) + (*arg - '0')) < port)
            break;
          port = (in_port_t)((port * 10) + (*arg++ - '0'));
        }
        if (*arg)
          port = 0;
      }
      else
      {
        fprintf(stderr,"Servername in %s is too long\n", oldarg );
        return 0;
      }
    }
    servbuf[pos] = '\0';
  }

  if (!port)
  {
    fprintf(stderr, "This test is a finger client using %s\n"
                    "Syntax: %s %s [user]@host[:port]\n"
                    "\teg %s %s cyp@distributed.net\n"
                    "\tThe default for 'user' is \"all users\"\n"
                    "\tThe default for 'port' is 79\n"
                    "", descrip, appname, testname, appname, testname );
  }
  return port;
}
#endif /* (TEST_LEVEL >= 4) */

/* ------------------------------------------------------------------- */

#if (TEST_LEVEL >= 4)
static int __oldfinger_test(const char *appname, const char *testname,
                            const char *name)
{
  char  user[32];
  char  server[64];
  static tcp_Socket sock;
  int    len;
  int    status = 0;
  in_addr_t  host;

  in_port_t port = __parse_finger_arg(appname, testname, name,
       "old style APIs (ergo implicitely non-blocking)",
       user, sizeof(user), server, sizeof(server) );
  if (!port) /* message already printed */
    return (3); 
  
  fprintf(stderr,"Attempting to resolve '%s'...\n", server );
  host = lookup_host( server, NULL );
  if (!host)
  {
    fprintf(stderr,"Could not resolve host '%s'\n",server);
    return (3);
  }

  fprintf(stderr,"Creating TCP socket to %s:%u...\n", sock_ntoa(host,NULL,0), port );
  if (!tcp_open(&sock,0,host, port,NULL))
  {
    fprintf(stderr,"tcp_open() failed\n");
    return (3);
  }

  fprintf(stderr,"Waiting for connect establishment...\n");
  sock_wait_established (&sock,sock_delay,NULL,&status);
  fprintf(stderr,"Connect completed...\n");

  strcpy (buffer,user);
  rip (buffer);                      /* kill all \n and \r's */
  strcat (buffer,"\r\n");
  fprintf(stderr, "Sending request...\n" );  
  sock_puts (&sock,buffer);
  //sock_close (&sock);               /* close sending side.... */

  fprintf(stderr,"Beginning wait/read loop...\n");
  setmode(1,O_BINARY);
  while (!got_brk_signal)
  {
    sock_wait_input (&sock,sock_delay,NULL,&status);
    len = sock_fastread (&sock,buffer,sizeof(buffer));
    if (len > 0)
      write(1,buffer,(size_t)len);
  }         
  setmode(1,O_TEXT);

sock_err:
  if (!got_brk_signal)
  {
    switch (status)
    {
      case 1 : fprintf(stderr,"\nRemote closed connection\n");
               break;
      case -1: fprintf(stderr,"\ntimeout/abort (Error: %s)\n",sockerr(&sock));
               break;
    }
  }

  sock_destroy(&sock);
  return (status);
}
#endif /* (TEST_LEVEL >= 4) */

/* ------------------------------------------------------------------- */

#if (TEST_LEVEL >= 4)
static int __newfinger_test(const char *appname, const char *testname,
                            const char *name)
{
  char  user[32];
  char  server[64];
  static tcp_Socket sock;
  in_addr_t  host;

  in_port_t port = __parse_finger_arg(appname, testname, name,
       "new style APIs with non-blocking mode enabled",
       user, sizeof(user), server, sizeof(server) );
  if (!port) /* message already printed */
    return (3); 
  
  fprintf(stderr,"Attempting to resolve '%s'...\n", server );
  if (sock_gethostbyname( server, &host, 1 ) < 1)
  {
    fprintf(stderr,"Could not resolve host '%s'\n",server);
    return (3);
  }
  fprintf(stderr,"'%s' resolved to %s\n", server, sock_ntoa(host,NULL,0) );

  fprintf(stderr,"Creating socket...\n");
  if (sock_create(&sock, __IPPROTO_TCP__) != 0)
  {
    fprintf(stderr,"sock_create() failed: err %d (%s)\n", errno, strerror(errno) );
  }
  else
  {
    fprintf(stderr,"Socket created ok, connecting to '%s:%u'...\n", server, port);
    if (sock_connect(&sock, &host, &port ) != 0)
    {
      fprintf(stderr,"sock_connect() failed: err %d (%s)\n", 
          sock_get_errno(&sock), sock_strerror(sock_get_errno(&sock)) );
    }
    else
    {
      int rc;
      strcat(strcpy(buffer, user),"\r\n");

      fprintf(stderr,"Connected to '%s', sending request...\n", server);
      rc = sock_write( &sock, buffer, strlen(buffer) );
      if (rc < ((int)strlen(buffer)))
      {
        fprintf(stderr,"sock_write() returned %d (expected %u). err = %d (%s)\n",
             rc, (unsigned)strlen(buffer), 
             sock_get_errno(&sock), sock_strerror(sock_get_errno(&sock)) );
      }
      else
      {
        unsigned int total = 0;

        fprintf(stderr,"Sent request ok, beginning read loop...\n");
        fprintf(stderr,"---------------------------------------\n");
        rc = 1;
        setmode(1,O_BINARY);
        while (rc > 0 && !got_brk_signal)
        {
          rc = sock_read( &sock, buffer, sizeof(buffer) );
          if (rc > 0)
          {
            write(1, buffer, rc );
            total += rc;
          }
        }
        setmode(1,O_TEXT);
        fprintf(stderr,"\n---------------------------------------\n");
        fprintf(stderr,"Total read = %u\n", total );
        if (!got_brk_signal)
          fprintf(stderr,"Last sock_read() returned %d. err = %d (%s)\n", rc,
               sock_get_errno(&sock), sock_strerror(sock_get_errno(&sock)) );
      } 
    }
    sock_destroy(&sock);
  }  
  return 0;
}
#endif /* (TEST_LEVEL >= 4) */

/* ====================================================================== */

#if (TEST_LEVEL >= 1)
static int __exit_test(const char *appname, const char *testname,
                            const char *arg)
{
  if (arg)
  {
    fprintf(stderr,
      "Syntax: %s %s\n"
      "This test initializes the stack, and then _exit()s.\n"
      "This test therefore determines if the atterm() handler is ok,\n"
      "but is meaningless unless the library was compiled for debugging\n"
      "since messages are not otherwise printed.\n"
      "", appname, testname );
    return 3;
  }
  __hostid_test(appname, testname, 0 );
  if (!arg) /* dummy if */
    _exit(0);
  return 0;
}
#endif

/* ====================================================================== */
/*                               PING                                     */
/* ====================================================================== */

#if (TEST_LEVEL >= 1)
static int __ping(int argc, char *argv[])
{
  struct
  {
    int rresolve;
    unsigned long num_requests;
    unsigned int bufsize;
    int df;
    int ttl;
    int tos;
    unsigned long timeout;
    char pattern[16];
    unsigned int pattlen;
    const char *hostname;
    in_addr_t hostaddr;
  } opts;

  struct 
  {
    int argi;
    unsigned int minv;
    unsigned long maxv;
  } badarg;

  struct
  {
    int hit_limits;
    unsigned long num_sent, num_rcvd;
    unsigned long total_time, max_time, min_time;
  } count;  

  int i;
  char rres_buf[256];

  opts.rresolve     = 0;
  opts.num_requests = 4;
  opts.bufsize      = 32;
  opts.df           = 0;
  opts.ttl          = 0xff;
  opts.tos          = 0;
  opts.timeout      = 1000ul;
  opts.pattern[0]   = '\0';
  opts.pattlen      = 0;
  opts.hostname     = (const char *)0;
  opts.hostaddr     = 0;

  badarg.argi = -1;
  i = 1;

  while (i < argc && !opts.hostname)
  {
    if (*argv[i] != '-' && *argv[i] != '/')
    {
      opts.hostname = argv[i];
      i++;
      continue;
    }
    else if (*(argv[i]+2) == '\0')
    {
      /* All the arg checking done here is purely for prettyness.
       * sock_ping() does its own checking as well.
      */
      long tmp;
      if (*(argv[i]+1) == 't') /* "Ping until stopped with ^C */
      {
        opts.num_requests = 0; /* continuous */
        i++;
        continue;
      }
      else if (*(argv[i]+1) == 'a') /* Resolve addresses to hostnames */
      {
        opts.rresolve = 1;
        i++;
        continue;
      }
      else if (*(argv[i]+1) == 'f') /* Set Don't Fragment flag in packet */
      {
        opts.df = 1;
        i++;
        continue;
      }
      else if ((i+1) >= argc) /* all other options require a value */ 
      {
        ; /* fallthrough */
      }
      else if (*(argv[i]+1) == 'p') /* Pattern */
      {
        const char *q = argv[i+1];
        opts.pattlen = 0;
        if (strlen(q) & 1) /* odd length implies leading zero */
        {
          opts.pattern[0] = 0;
          opts.pattlen = 1;
        }
        while (*q)
        {
          char c = 0;
          if (*q >= 'a' && *q <= 'f')
            c = (char)(10 + (*q++ - 'a'));
          else if (*q >= 'A' && *q <= 'F')
            c = (char)(10 + (*q++ - 'F'));
          else if (*q >= '0' && *q <= '9')
            c = (char)(*q++ - '0');
          else
          {
            fprintf(stderr, "Patterns must be specified as hex digits.\n");
            break;
          }
          if ((opts.pattlen & 1) == 0)
            opts.pattern[opts.pattlen>>1] = (char)(c << 4);
          else
            opts.pattern[opts.pattlen>>1] |= c;

          if (((++opts.pattlen)>>1) > sizeof(opts.pattern))
          {
            fprintf(stderr, 
              "Patterns may only be upto %u octets (%u hex digits) long.\n",
                                     (unsigned)(sizeof(opts.pattern)), 
                                     (unsigned)(sizeof(opts.pattern)<<1) );
            break;
          }
        }
        if (!*q)
        {
          opts.pattlen >>= 1;
          i+=2;
          continue;
        }
        badarg.argi = i;
        badarg.minv = 0; /* max == min means "message already printed" */
        badarg.maxv = 0;
      }
      else if (*argv[i+1] < '0' || *argv[i+1] > '9')
      {
        ; /* fallthrough */
      }
      else if (*(argv[i]+1) == 'n') /* Number of requests to send */
      {
        tmp = atol(argv[i+1]);
        if (tmp != 0)
        {
          opts.num_requests = tmp;
          i+=2;
          continue;
        } 
        badarg.argi = i;
        badarg.minv = 1;
        badarg.maxv = ULONG_MAX;
      }
      else if (*(argv[i]+1) == 'l') /* Send buffer size */
      {
        /* 65500u == max ip pkt - (iphdr_size+icmphdr_size+echohdr_size+4) */
        tmp = atol(argv[i+1]);
        if (tmp >= 0L && tmp <= 65500L)
        {
          opts.bufsize = (unsigned int)tmp;
          i+=2;
          continue;
        }
        badarg.argi = i;
        badarg.minv = 0;
        badarg.maxv = 65500u;
      } 
      else if (*(argv[i]+1) == 'w') /* Timeout in milliseconds */
      { 
        opts.timeout = atol(argv[i+1]);
        i+=2;
        continue;
      }
      else if (*(argv[i]+1) == 'i') /* TTL */
      { 
        tmp = atol(argv[i+1]);
        if (tmp >= 1 && tmp <= 0xff)
        {
          opts.ttl = (int)tmp;
          i+=2;
          continue;
        }
        badarg.argi = i;
        badarg.minv = 1;
        badarg.maxv = 0xff;
      }
      else if (*(argv[i]+1) == 'v') /* TOS */
      { 
        tmp = atol(argv[i+1]);
        if (tmp >= 0 && tmp <= 0xff)
        {
          opts.tos = (int)tmp;
          i+=2;
          continue;
        }
        badarg.argi = i;
        badarg.minv = 0;
        badarg.maxv = 0xff;
      }
    }   
    opts.hostname = (const char *)0;
    break;
  }

  if (!opts.hostname)
  {
    if (badarg.argi != -1)
    {
      if (badarg.maxv != badarg.minv) /* else message already printed */
      {
        fprintf(stderr, 
          "Bad value for option %s, valid range is from %u to %lu.\n",
             argv[badarg.argi], badarg.minv, badarg.maxv );
      }
    }
    else
    {
      const char *appname = argv[0] + strlen(argv[0]);
      while (appname > argv[0] && 
        *(appname-1) != ':' && *(appname-1) != '\\' && *(appname-1) != '/')
      {
        appname--;
      }
      fprintf(stderr, "\n"
        "Usage: %s [-t] [-a] [-n count] [-l size] [-f] [-i TTL] [-v TOS]\n"
        "          [-w timeout] [-p pattern] destination\n"
        "\n"
        "Options:\n"
        "    -t             Ping the specified host until stopped with ^C.\n"
        "    -a             Resolve addresses to hostnames.\n"
        "    -n count       Number of echo requests to send.\n"
        "    -l size        Send buffer size.\n"
        "    -f             Set Don't Fragment flag in packet.\n"
        "    -i TTL         Time To Live.\n"
        "    -v TOS         Type Of Service.\n"
        "    -w timeout     Timeout in milliseconds to wait for each reply.\n"
        "    -p pattern     Pattern to pad buffer with.\n"
        "", appname );
    }
    return 0;
  }  

  if (!sock_aton(opts.hostname, &opts.hostaddr))
  {
#if (TEST_LEVEL >= 2)
    i = sock_gethostbyname(opts.hostname, &opts.hostaddr, 1);
    if (i < 1)
    {
      fprintf(stderr, "Unable to resolve '%s' (DNS error %d)\n",  
                         opts.hostname, i );
      return 0;
    }
#else
    fprintf(stderr, "Unable to do DNS resolution at TEST_LEVEL < 2\n");
    return 0;
#endif
  }

  rres_buf[0] = '\0';
#if (TEST_LEVEL >= 2)
  if (opts.rresolve)
  {
    if (sock_gethostbyaddr( opts.hostaddr, rres_buf, 
        (sizeof(rres_buf) - sizeof("[255.255.255.255]")) ) < 1)
      rres_buf[0] = '\0';
    else
    {
      i = strlen( strcat(rres_buf," [" ) );
      sock_ntoa(opts.hostaddr, &rres_buf[i], sizeof("255.255.255.255"));
      strcat( rres_buf, "]" );
    }
  }
#endif
  if (!rres_buf[0])
    sock_ntoa( opts.hostaddr, rres_buf, sizeof(rres_buf) );

  printf("\nPinging %s with %u bytes of data:\n\n", rres_buf, opts.bufsize );

  memset( &count, 0, sizeof(count));
  for (;;)
  {
    int ttl = opts.ttl, tos = opts.tos;
    unsigned int bufsize = opts.bufsize;
    unsigned long timeout = opts.timeout;
    in_addr_t hostaddr = opts.hostaddr;
    int so_errno;

    so_errno = sock_ping( &hostaddr, &ttl, &tos, &bufsize, &timeout, opts.df,
                          opts.pattern, opts.pattlen );

    if (so_errno == 0)
    {
      printf("\rReply from %s: bytes=%u time=%s%lums TTL=%u\n", 
             sock_ntoa( hostaddr, rres_buf, sizeof(rres_buf) ), bufsize, 
             ((!timeout) ? "<" : ""), ((!timeout)? 55ul : timeout), ttl );

      count.num_sent++;
      count.num_rcvd++;

      if (timeout > count.max_time)
        count.max_time = timeout;
      if (count.num_rcvd == 1 || timeout < count.min_time)
        count.min_time = timeout;
      if (count.total_time >= (ULONG_MAX - timeout))
      {
        count.hit_limits = 1;
        break;
      }
      count.total_time += timeout;
    }
    else if (so_errno == WSO_EREMOTE)
    {
      const char *msg = (const char *)0;
      switch (bufsize)
      {
        case -1: msg = "?? (Unrecognized ICMP error)"; break; /* -1 */
        default: /* fallthrough */
        case  0: msg = "huh?"; break;
        case  1: msg = "Destination host/net unreachable."; break;
        case  2: msg = "Destination protocol unreachable."; break; /* never */
        case  3: msg = "Destination port unreachable."; break;     /* never */
        case  4: msg = "Packet needs to be fragmented but DF set."; break;  
        case  5: msg = "TTL expired during reassembly."; break;
        case  6: msg = "TTL expired in transit."; break;
        case  7: msg = "Source quench received."; break;
        case  8: msg = "IP parameter problem."; break;
      }
      printf("\rReply from %s: %s\n", 
             sock_ntoa( hostaddr, rres_buf, sizeof(rres_buf) ), msg );

      count.num_sent++;
      count.num_rcvd++;
    }
    else if (so_errno == WSO_ETIMEDOUT)
    {
      printf("\rRequest timed out.\n" );
      count.num_sent++;
    }
    else if (so_errno == WSO_EINTR)
    {
      count.num_sent++;
      break;
    }
    else 
    {
      printf("Permanent fatal (local) error %u (%s)\n", 
                    so_errno, sock_strerror(so_errno) );
      break;
    }

    if (got_brk_signal)
      break;

    if (opts.num_requests && count.num_sent >= opts.num_requests)
      break;
    if (count.num_sent == ULONG_MAX)
    {
      count.hit_limits = 1;
      break; /* can't do statistics otherwise */
    }

    /* don't ping more than once per second (regardless of opts.timeout) */
    if (so_errno == 0 && timeout < 1000ul)
      delay(1000 - ((unsigned)timeout));

  } /* for (;;) */

  if (count.hit_limits)
    printf("Cancelled ping due to numerical limits\n");

  {
    unsigned int perc = 10000;
    unsigned long avgtm = 0;
    unsigned long lost = count.num_sent - count.num_rcvd;
    if (count.num_rcvd)
    {
      if (lost == 0)
        perc = 0;
      else if (lost < (ULONG_MAX / 10000ul))
        perc = (unsigned)((lost * 10000ul)/count.num_sent);
      else if (lost < (ULONG_MAX / 1000ul))
        perc = (unsigned)((lost * 1000ul)/(count.num_sent/10ul));
      else if (lost < (ULONG_MAX / 100ul))
        perc = (unsigned)((lost * 100ul)/(count.num_sent/100ul));
      else if (lost < (ULONG_MAX / 10ul))
        perc = (unsigned)((lost * 10ul)/(count.num_sent/1000ul));
      else
        perc = (unsigned)((lost)/(count.num_sent/10000ul));
      if (count.total_time < (ULONG_MAX>>1))
        avgtm = ((count.total_time<<1) / count.num_rcvd)>>1;
      else
        avgtm = count.total_time / count.num_rcvd;
    }

    printf("\nPing statistics for %s:\n", opts.hostname );
    printf("   Packets: Sent = %lu, Received = %lu, Lost = %lu (%u.%02u%% loss),\n",
               count.num_sent, count.num_rcvd, lost, perc / 100, perc % 100 );
    printf("Approximate round trip times in milli-seconds:\n" );
    printf("   Minimum = %lums, Maximum = %lums, Average = %lums\n",
              count.min_time, count.max_time, avgtm );
  }
  return 0;
}
#endif

/* ====================================================================== */
/*                                                                        */
/* ====================================================================== */

static int __sigcheck(const char *appname, const char *testname,
                      const char *arg)  
{
  if (arg)
  {
    fprintf(stderr, "Syntax: %s %s without further arguments to\n"
    "enter a signal waiter loop.\n", appname, testname );
  }
  else
  {
    fprintf(stderr, "Begin sigcheck\n");
    while (!got_brk_signal)
      delay(1000); 
    fprintf(stderr, "End sigcheck\n");
  }
  return 0;
}

/* ====================================================================== */
/*                                                                        */
/* ====================================================================== */

int main(int argc, char *argv[])
{
#if (TEST_LEVEL >= 1)
  static const struct {
    const char *key,*arg; int (*fxn)(const char *,const char *,const char *);
  } opts[] = {
    { "-hostid",     "",                           __hostid_test     },
    { "-exit",       "",                           __exit_test       },
    { "-bgloop",     "<type>",                     __bgloop_test     },
#if (TEST_LEVEL >= 3)
    { "-dns",        "[<hostname>|<address>]",     __dns_test        },
    { "-name",       "",                           __name_test       },
#endif
#if (TEST_LEVEL >= 4)
    { "-htget",      "<url>",                      __htget_test      },
    { "-oldfinger",  "<user@host>",                __oldfinger_test  },
    { "-newfinger",  "<user@host>",                __newfinger_test  },
#endif
    { "-sigchk",     "",                           __sigcheck        },
    { NULL, NULL, NULL }
  };

  unsigned int i;
  const char *appname = argv[0] + strlen(argv[0]);

  while (appname > argv[0] && 
        *(appname-1) != ':' && *(appname-1) != '\\' && *(appname-1) != '/')
  {
    appname--;
  }

  #ifdef SIGBREAK
  signal(SIGBREAK, sig_handler );
  #endif
  signal(SIGINT, sig_handler );

  if (argc > 1)
  {
    #if (TEST_LEVEL >= 1)
    if (strcmp(argv[1],"-ping") == 0)
    {
      return __ping( argc-1, &argv[1] );
    }
    #endif
    for (i = 0; opts[i].fxn && i < (sizeof(opts)/sizeof(opts[0])); i++)
    {
      const char *testname = argv[1];
      if (strcmp(testname, opts[i].key) == 0)
      {
        const char *arg = (argc > 2 ? argv[2] : NULL);
        if (arg && (*arg == '-' || *arg == '/'))
        {
          if (arg[1] == 'h' || arg[1] == 'H' || arg[1] == '?')
            arg = "?";
        }
        return (*(opts[i].fxn))(appname, testname, arg);
      }
    } 
  }

  fprintf(stderr,"Syntax: %s <testname> [<testargs>]\n"
                 "where <testname> is one of...\n\n", appname );
  for (i = 0; opts[i].fxn && i < (sizeof(opts)/sizeof(opts[0])); i++)
    fprintf(stderr, "   %s %s\n", opts[i].key, opts[i].arg );
  #if (TEST_LEVEL >= 1)
  fprintf(stderr, "   -ping [-t] [-a] [-n count] [-l size] [-i TTL] [-v TOS]\n"
                  "         [-w timeout] destination\n" );
  #endif
  fprintf(stderr,"\nUse '%s <testname> -?' for for information about a test.\n"
          "", appname );

#else
  argc = argc;
  argv = argv;
  sock_gethostid();
#endif
  return 0; 
}

